/*
 * Copyright (c) 2003-2004 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: boardlist_ui.c,v 1.92.2.9 2005/02/18 17:33:16 fuyu Exp $
 */

#define _BOARDLIST_UI_C_ 1
#include "config.h"

#include "ochusha_private.h"
#include "ochusha.h"
#include "ochusha_utils_2ch.h"
#include "worker.h"

#include "ochusha_ui.h"
#include "bbs_thread_ui.h"
#include "board_properties.h"
#include "boardlist_ui.h"
#include "boardlist_view.h"
#include "bulletin_board_ui.h"
#include "paned_notebook.h"
#include "response_editor.h"
#include "text_search_window.h"
#include "ugly_gtk2utils.h"
#include "virtual_board.h"

#include <glib.h>
#include <gtk/gtk.h>

#include <errno.h>

#if HAVE_ONIG_ONIGPOSIX_H
# include <onig/onigposix.h>
#else
# include <onigposix.h>
#endif

#include <fcntl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

#include <zlib.h>


#if DEBUG_WIDGET_MOST
static void select_category_cb(BoardlistView *view,
			       OchushaBoardCategory *category,
			       gpointer user_data);
#endif
static void category_expanded_cb(BoardlistView *view,
				 OchushaBoardCategory *category,
				 OchushaApplication *application);
static void category_collapsed_cb(BoardlistView *view,
				  OchushaBoardCategory *category,
				  OchushaApplication *application);
static void toggle_hide_category_cb(BoardlistView *view,
				    OchushaBoardCategory *category,
				    OchushaApplication *application);
static void toggle_hide_board_cb(BoardlistView *view,
				 OchushaBulletinBoard *board,
				 OchushaApplication *application);
static gboolean kill_category_cb(BoardlistView *view,
				 OchushaBoardCategory *category,
				 OchushaApplication *application);
static gboolean kill_board_cb(BoardlistView *view,
			      OchushaBoardCategory *category,
			      OchushaBulletinBoard *board,
			      OchushaApplication *application);
static void copy_board_url_cb(BoardlistView *view, OchushaBulletinBoard *board,
			      OchushaApplication *application);
static void open_properties_cb(BoardlistView *view, GObject *object,
			       OchushaApplication *application);
static void uri_dropped_cb(BoardlistView *view, OchushaBoardCategory *category,
			   OchushaBulletinBoard *board, const char *uri,
			   GtkTreeViewDropPosition pos,
			   OchushaApplication *application);

static OchushaBulletinBoard *board_new_cb(OchushaBBSTable *table,
					  int bbs_type, const gchar *name,
					  const char *base_url,
					  OchushaApplication *application);
static void boardlist_read_category_element_cb(OchushaBBSTable *table,
					       OchushaBoardCategory *category,
					       GHashTable *cat_attributes,
					       OchushaApplication *app);
static void boardlist_read_board_element_cb(OchushaBBSTable *table,
					    OchushaBulletinBoard *board,
					    GHashTable *cat_attributes,
					    OchushaApplication *app);

static void boardlist_write_category_element_cb(OchushaBBSTable *table,
						OchushaBoardCategory *category,
						gzFile boardlist_xml,
						OchushaApplication *app);
static void boardlist_write_board_element_cb(OchushaBBSTable *table,
					     OchushaBulletinBoard *board,
					     gzFile boardlist_xml,
					     OchushaApplication *app);

static void write_threadlist(gpointer data, gpointer user_data);

static void build_boardlist(WorkerThread *employee, gpointer args);

static void setup_boardlist(OchushaApplication *application);

static void filter_toggled_cb(GObject *unused,
			      OchushaApplication *application);
static void item_view_required_cb(PanedNotebook *paned_notebook,
				  gpointer item, GtkWidget **item_view,
				  const gchar **title, GtkWidget **tab_label,
				  OchushaApplication *application);
static void item_view_being_closed_cb(PanedNotebook *paned_notebook,
				      OchushaBulletinBoard *board,
				      GtkWidget *item_view,
				      OchushaApplication *application);
static void item_view_closed_cb(PanedNotebook *paned_notebook,
				OchushaBulletinBoard *board, GtkWidget *item_view,
				OchushaApplication *application);
static void page_switched_cb(PanedNotebook *paned_notebook,
			     GtkWidget *previous_page, GtkWidget *new_page,
			     OchushaApplication *application);
static void page_double_clicked_cb(PanedNotebook *paned_notebook,
				   OchushaBulletinBoard *board,
				   GtkWidget *item_view,
				   OchushaApplication *application);

static void make_new_thread_button_cb(GObject *unused,
				      OchushaApplication *application);
static void open_board_properties_button_cb(GObject *unused,
					    OchushaApplication *application);
static void refresh_board_button_cb(GObject *unused,
				    OchushaApplication *application);
static void start_threadlist_search_button_cb(GObject *unused,
					      OchushaApplication *application);
static void new_thread_dialog_response_cb(GtkWidget *dialog, int response_id,
					  OchushaApplication *application);
static GtkWidget *create_new_thread_dialog(OchushaApplication *application,
					   OchushaBulletinBoard *board);


/* Υå˰٤ǤⳫ(OchushaBulletinBoard *)Υꥹȡ*/
static GSList *board_list;

static pthread_mutex_t board_list_lock;

#define BOARD_LIST_LOCK					\
  if (pthread_mutex_lock(&board_list_lock) != 0)	\
    {							\
      fprintf(stderr, "Couldn't lock a mutex.\n");	\
      abort();						\
    }

#define BOARD_LIST_UNLOCK				\
  if (pthread_mutex_unlock(&board_list_lock) != 0)	\
    {							\
      fprintf(stderr, "Couldn't unlock a mutex.\n");	\
      abort();						\
    }


static GQuark category_gui_info_id;


void
prepare_boardlist_ui_initialization(OchushaApplication *application)
{
  if (pthread_mutex_init(&board_list_lock, NULL) != 0)
    {
      fprintf(stderr, "Couldn't init a mutex.\n");
      abort();
    }

  category_gui_info_id
    = g_quark_from_static_string("BoardlistUI::CategoryInfo");
}


static void
open_threadlist_view_cb(GtkWidget *selector, OchushaBulletinBoard *board,
			gboolean in_tab, gboolean with_browser,
			PanedNotebook *paned_notebook)
{
  OchushaApplication *application = g_object_get_qdata(G_OBJECT(selector),
						       application_id);
  if (!with_browser)
    paned_notebook_open_item_view(paned_notebook, board, in_tab,
				  application->select_tab_when_opened);
  else
    {
      if (application != NULL)
	ochusha_open_url(application,
			 ochusha_bulletin_board_get_base_url(board),
			 in_tab, with_browser);
    }
}


static guint new_thread_netstat_id;


static void
setup_initial_favorite_thread_list(OchushaBulletinBoard *board,
				   OchushaApplication *application)
{
  if (board->thread_list == NULL)
    ochusha_bulletin_board_read_threadlist_xml(board,
					       &application->config,
					       application->session_subdir,
					       FALSE);
}


static void
ensure_my_folder(OchushaApplication *application)
{
  BoardCategoryGUIInfo *info;
  if (application->my_folder == NULL)
    {
      application->my_folder = ochusha_board_category_new(_("My Folder"));
      info = ensure_board_category_info(application->my_folder);
      info->category_type = CATEGORY_TYPE_OCHUSHA_MY_FOLDER;
      ochusha_bbs_table_prepend_category(application->table,
					 application->my_folder);
    }

  if (application->favorites == NULL)
    {
      /* äURLǤܤ */
      application->favorites = virtual_board_new(_("Favorites"),
						 "http:///favorites",
						 application);
      ochusha_board_category_add_board(application->my_folder,
				       application->favorites);
      application->favorites_list
	= g_slist_prepend(application->favorites_list, application->favorites);
    }

  if (application->all_threads == NULL)
    {
      BulletinBoardGUIInfo *board_info;
      application->all_threads = virtual_board_new(_("All Threads Read"),
						   "http:///all-threads",
						   application);
      board_info = ensure_bulletin_board_info(application->all_threads,
					      application);
      board_info->use_static_threadlist = TRUE;
      ochusha_board_category_add_board(application->my_folder,
				       application->all_threads);
    }

  if (application->thread_list_already_read)
    return;

  application->thread_list_already_read = TRUE;

#if 1	/* ɲäȤ硢ΰѰդƤɬפ롣
	 * ΰ١λɤ߹Ǥɬפ˱ɤ߹褦ˤ
	 * ȤԲǽǤϤʤޥɥ
	 */
  g_slist_foreach(application->favorites_list,
		  (GFunc)setup_initial_favorite_thread_list, application);
#else
  if (application->favorites->thread_list == NULL)
    ochusha_bulletin_board_read_threadlist_xml(application->favorites,
					       &application->config,
					       application->session_subdir,
					       FALSE);
#endif

#if 1
  if (application->all_threads->thread_list == NULL)
    ochusha_bulletin_board_read_threadlist_xml(application->all_threads,
					       &application->config,
					       application->session_subdir,
					       FALSE);
#endif
}


void
initialize_boardlist(OchushaApplication *application)
{
  GtkWidget *scrolled_window;
  GtkWidget *filter_button;
#if GTK_MINOR_VERSION <= 2
  GtkToolbar *toolbar;
#else
  GtkActionGroup *action_group;
  GtkUIManager *ui_manager;
  static GtkActionEntry toolbar_entries[] =
    {
      {
	"CreateNewThread", GTK_STOCK_NEW, NULL, NULL,
	N_("Make a New Thread at the Current Board"),
	(GCallback)make_new_thread_button_cb
      },
      {
	"OpenBoardProperties", GTK_STOCK_PROPERTIES, NULL, NULL,
	N_("Open the Current Board's Properties"),
	(GCallback)open_board_properties_button_cb
      },
      {
	"SearchOnThreadlist", GTK_STOCK_FIND, NULL, NULL,
	N_("Start Text Search on the Current Threadlist"),
	(GCallback)start_threadlist_search_button_cb
      },
      {
	"RefreshThreadlist", GTK_STOCK_REFRESH, NULL, NULL,
	N_("Refresh the Current Threadlist"),
	(GCallback)refresh_board_button_cb
      }
    };

  static GtkToggleActionEntry toolbar_toggle_entries[] =
    {
      {
	"ApplyFilter", NULL, N_("Filter"), NULL,
	N_("Apply Filter to the Current Threadlist"),
	(GCallback)filter_toggled_cb,
	FALSE
      }
    };

  static const char *toolbar_ui =
    "<ui>"
    "  <toolbar name=\"PanedNotebookLeftToolbar\">"
    "    <toolitem name=\"Refresh\" action=\"RefreshThreadlist\"/>"
    "    <toolitem name=\"Search\" action=\"SearchOnThreadlist\"/>"
    "    <placeholder name=\"BoardToolItems\">"
    "      <toolitem name=\"Filter\" action=\"ApplyFilter\"/>"
    "      <toolitem name=\"Properties\" action=\"OpenBoardProperties\"/>"
    "    </placeholder>"
    "    <toolitem name=\"NewThread\" action=\"CreateNewThread\"/>"
    "  </toolbar>"
    "</ui>";

  GError *error;
#endif

  board_list = NULL;

  new_thread_netstat_id = gtk_statusbar_get_context_id(application->statusbar,
						       "new-thread-netstat");

  application->boardlist_view = boardlist_view_new();
  g_object_set_qdata(G_OBJECT(application->boardlist_view), application_id,
		     application);
  g_signal_connect(G_OBJECT(application->boardlist_view), "category_expanded",
		   G_CALLBACK(category_expanded_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view), "category_collapsed",
		   G_CALLBACK(category_collapsed_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view),
		   "toggle_hide_category",
		   G_CALLBACK(toggle_hide_category_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view),
		   "toggle_hide_board",
		   G_CALLBACK(toggle_hide_board_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view), "kill_category",
		   G_CALLBACK(kill_category_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view), "kill_board",
		   G_CALLBACK(kill_board_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view), "copy_board_url",
		   G_CALLBACK(copy_board_url_cb), application);
  g_signal_connect(G_OBJECT(application->boardlist_view), "open_properties",
		   G_CALLBACK(open_properties_cb), application);

  g_signal_connect(G_OBJECT(application->boardlist_view), "uri_dropped",
		   G_CALLBACK(uri_dropped_cb), application);

#if TRACE_MEMORY_USAGE
  g_signal_connect(G_OBJECT(application->boardlist_view), "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &application->boardlist_view);
#endif
  boardlist_view_set_default_open_in_tab(
				BOARDLIST_VIEW(application->boardlist_view),
				application->default_open_in_tab);
  gtk_widget_show(application->boardlist_view);

  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_add(GTK_CONTAINER(scrolled_window),
		    application->boardlist_view);

  application->contents_window
    = paned_notebook_new_with_selector(application->boardlist_pane_style,
				       scrolled_window);
  g_signal_connect(G_OBJECT(application->boardlist_view),
		   "open_threadlist_view", G_CALLBACK(open_threadlist_view_cb),
		   application->contents_window);
  gtk_paned_set_position(GTK_PANED(application->contents_window),
			 application->boardlist_width);
  paned_notebook_set_always_show_contents_pane(PANED_NOTEBOOK(application->contents_window),
					       TRUE);
  paned_notebook_set_tab_shrinkable(PANED_NOTEBOOK(application->contents_window), application->board_tab_shrinkable);
  paned_notebook_set_minimum_tab_label_size(PANED_NOTEBOOK(application->contents_window), application->board_tab_minimum_size);
  paned_notebook_set_tooltips(PANED_NOTEBOOK(application->contents_window),
			      application->board_tab_enable_tooltips);
  paned_notebook_set_tab_policy(PANED_NOTEBOOK(application->contents_window),
				application->board_tab_always_show
				? GTK_POLICY_ALWAYS : GTK_POLICY_AUTOMATIC);

#if GTK_MINOR_VERSION <= 2
  toolbar = paned_notebook_get_toolbar(PANED_NOTEBOOK(application->contents_window));

  gtk_toolbar_prepend_space(toolbar);
  gtk_toolbar_insert_stock(toolbar, GTK_STOCK_NEW,
			   _("Make a New Thread at the Current Board"),
			   "make_new_thread_at_current_board",
			   GTK_SIGNAL_FUNC(make_new_thread_button_cb),
			   application,
			   0);

  gtk_toolbar_prepend_space(toolbar);
  gtk_toolbar_insert_stock(toolbar, GTK_STOCK_PROPERTIES,
			   _("Open the Current Board's Properties"),
			   "open_the_current_board_properties",
			   GTK_SIGNAL_FUNC(open_board_properties_button_cb),
			   application,
			   0);
  filter_button = gtk_toggle_button_new_with_label(_("Filter"));
  GTK_WIDGET_UNSET_FLAGS(filter_button, GTK_CAN_FOCUS);
  gtk_widget_show(filter_button);
  application->filter_button = GTK_TOGGLE_BUTTON(filter_button);
  gtk_toolbar_prepend_widget(toolbar, filter_button,
			     _("Apply Filter to the Current Threadlist"),
			     "apply_filter_to_the_current_threadlist");
  g_signal_connect(G_OBJECT(filter_button), "toggled",
		   G_CALLBACK(filter_toggled_cb), application);
# if TRACE_MEMORY_USAGE
  g_signal_connect(G_OBJECT(filter_button), "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &application->filter_button);
# endif

  gtk_toolbar_prepend_space(toolbar);
  gtk_toolbar_insert_stock(toolbar, GTK_STOCK_FIND,
			   _("Start Text Search on the Current Threadlist"),
			   "start_text_search_on_threadlist",
			   GTK_SIGNAL_FUNC(start_threadlist_search_button_cb),
			   application,
			   0);

  gtk_toolbar_prepend_space(toolbar);
  gtk_toolbar_insert_stock(toolbar, GTK_STOCK_REFRESH,
			   _("Refresh the Current Threadlist"),
			   "refresh_current_threadlist",
			   GTK_SIGNAL_FUNC(refresh_board_button_cb),
			   application,
			   0);
#else
  action_group = gtk_action_group_new("ThreadlistToolbarActions");
  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group,
			       toolbar_entries,
			       G_N_ELEMENTS(toolbar_entries),
			       application);
  gtk_action_group_add_toggle_actions(action_group,
				      toolbar_toggle_entries,
				      G_N_ELEMENTS(toolbar_toggle_entries),
				      application);
  ui_manager = PANED_NOTEBOOK(application->contents_window)->ui_manager;
  gtk_ui_manager_insert_action_group(ui_manager, action_group, 0);
  g_object_unref(action_group);

  error = NULL;
  if (!gtk_ui_manager_add_ui_from_string(ui_manager,
					 toolbar_ui, -1, &error))
    {
      g_message("building toolbar failed: %s\n", error->message);
      g_error_free(error);
      exit(EXIT_FAILURE);
    }

  filter_button
    = gtk_ui_manager_get_widget(ui_manager,
				"/PanedNotebookLeftToolbar/BoardToolItems/Filter");
  GTK_WIDGET_UNSET_FLAGS(filter_button, GTK_CAN_FOCUS);
  application->filter_button = GTK_TOGGLE_TOOL_BUTTON(filter_button);

# if 1
  {
    /* XXX: ξ줷Τκ */
    GtkToolButton *tool_button = GTK_TOOL_BUTTON(application->filter_button);
    GtkWidget *icon_widget = gtk_tool_button_get_icon_widget(tool_button);
    GtkWidget *label_widget = gtk_tool_button_get_label_widget(tool_button);

    if (icon_widget == NULL && label_widget == NULL)
      {
	label_widget = gtk_label_new(gtk_tool_button_get_label(tool_button));
	gtk_widget_show(label_widget);
	gtk_tool_button_set_icon_widget(tool_button, label_widget);
      }

    gtk_tool_item_set_homogeneous(GTK_TOOL_ITEM(tool_button), FALSE);
  }
# endif
#endif

#if DEBUG_WIDGET_MOST
  g_signal_connect(G_OBJECT(application->boardlist_view), "select_category",
		   G_CALLBACK(select_category_cb), application);
#endif
  g_signal_connect(G_OBJECT(application->contents_window),
		   "item_view_required", G_CALLBACK(item_view_required_cb),
		   application);
  g_signal_connect(G_OBJECT(application->contents_window),
		   "item_view_being_closed",
		   G_CALLBACK(item_view_being_closed_cb), application);
  g_signal_connect(G_OBJECT(application->contents_window),
		   "item_view_closed", G_CALLBACK(item_view_closed_cb),
		   application);
  g_signal_connect(G_OBJECT(application->contents_window),
		   "page_switched", G_CALLBACK(page_switched_cb),
		   application);
  g_signal_connect(G_OBJECT(application->contents_window),
		   "page_double_clicked", G_CALLBACK(page_double_clicked_cb),
		   application);

  application->threadlist_search_window = NULL;

  application->table = ochusha_bbs_table_new();

  g_signal_connect(G_OBJECT(application->table),
		   "board_new",
		   G_CALLBACK(board_new_cb), application);
  g_signal_connect(G_OBJECT(application->table),
		   "boardlist_read_category_element",
		   G_CALLBACK(boardlist_read_category_element_cb),
		   application);
  g_signal_connect(G_OBJECT(application->table),
		   "boardlist_read_board_element",
		   G_CALLBACK(boardlist_read_board_element_cb),
		   application);
  g_signal_connect(G_OBJECT(application->table),
		   "boardlist_write_category_element",
		   G_CALLBACK(boardlist_write_category_element_cb),
		   application);
  g_signal_connect(G_OBJECT(application->table),
		   "boardlist_write_board_element",
		   G_CALLBACK(boardlist_write_board_element_cb),
		   application);

  if (ochusha_bbs_table_read_boardlist_xml(application->table,
					   &application->config,
					   application->session_subdir))
    {
      setup_boardlist(application);
      return;
    }

  refresh_boardlist(application);
}


#if TRACE_MEMORY_USAGE
static void
trace_free(gpointer pointer)
{
  G_FREE(pointer);
}
#define TRACE_FREE	trace_free
#else
#define TRACE_FREE	g_free
#endif


BoardCategoryGUIInfo *
ensure_board_category_info(OchushaBoardCategory *category)
{
  BoardCategoryGUIInfo *info = g_object_get_qdata(G_OBJECT(category),
						  category_gui_info_id);
  if (info == NULL)
    {
      info = G_NEW0(BoardCategoryGUIInfo, 1);
      g_object_set_qdata_full(G_OBJECT(category), category_gui_info_id,
			      info, (GDestroyNotify)TRACE_FREE);
    }

  return info;
}


void
refresh_boardlist(OchushaApplication *application)
{
  WorkerJob *job = G_NEW0(WorkerJob, 1);

  job->canceled = FALSE;
  job->job = build_boardlist;
  job->args = application;

  commit_job(job);
}


static void
update_board_list(gpointer list_data, gpointer user_data)
{
  if (g_slist_find(board_list, list_data) == NULL)
    {
      OchushaBulletinBoard *board = OCHUSHA_BULLETIN_BOARD(list_data);
      if (board->thread_list != NULL)
	board_list = g_slist_append(board_list, board);
    }
}


static void
save_favorites(VirtualBoard *favorites, OchushaApplication *application)
{
  g_return_if_fail(IS_VIRTUAL_BOARD(favorites));

  g_slist_foreach(favorites->board_list, update_board_list, NULL);
  virtual_board_clean_up_threadlist(favorites, FALSE);
  write_threadlist(favorites, application);
}


void
finalize_boardlist(OchushaApplication *application)
{
#if 0
  fprintf(stderr, "finalize_boardlist\n");
#endif
  if (application->contents_window != NULL)
    {
      PanedNotebook *boardlist = PANED_NOTEBOOK(application->contents_window);
      GtkWidget *current_board;
      application->boardlist_width
	= gtk_paned_get_position(GTK_PANED(boardlist));
      current_board = paned_notebook_get_current_item_view(boardlist);

      if (current_board != NULL)
	application->threadlist_height
	  = gtk_paned_get_position(GTK_PANED(current_board));
    }

  if (application->table != NULL)
    {
      if (!ochusha_bbs_table_write_boardlist_xml(application->table,
						 &application->config,
						 application->session_subdir))
	fprintf(stderr, "Couldn't write boardlist.xml.\n");

      if (application->session_subdir != NULL
	  && !ochusha_bbs_table_write_boardlist_xml(application->table,
						    &application->config,
						    NULL))
	fprintf(stderr, "Couldn't write boardlist.xml.\n");
    }

  /* XXX: Ĥɲä򥵥ݡȤ顢ͤɬפ */
#if 1
  g_slist_foreach(application->favorites_list,
		  (GFunc)save_favorites, application);
#else
  if (IS_VIRTUAL_BOARD(application->favorites))
    {
      VirtualBoard *favorites = VIRTUAL_BOARD(application->favorites);
      g_slist_foreach(favorites->board_list, update_board_list, NULL);
      virtual_board_clean_up_threadlist(favorites, FALSE);
      write_threadlist(favorites, application);
    }
#endif

  if (IS_VIRTUAL_BOARD(application->all_threads))
    {
      VirtualBoard *all_threads = VIRTUAL_BOARD(application->all_threads);
      g_slist_foreach(all_threads->board_list, update_board_list, NULL);
      virtual_board_clean_up_threadlist(all_threads, TRUE);
      write_threadlist(all_threads, application);
    }

  g_slist_foreach(board_list, write_threadlist, application);
}


static char *tree_view_font_name = NULL;
static GtkWidget *font_selection_dialog = NULL;


static gboolean
write_gtkrc_for_tree_view_font(OchushaApplication *application,
			       const char *font_name)
{
  int fd = ochusha_config_open_file(&application->config,
				    OCHUSHA_PREFS_GTKRC, NULL,
				    O_WRONLY | O_TRUNC | O_CREAT);
  if (fd >= 0)
    {
      FILE *file = fdopen(fd, "w");
      if (file != NULL)
	{
	  fprintf(file, "# " OCHUSHA_PREFS_GTKRC " generated by " PACKAGE_STRING ".\n");
	  fprintf(file, "# Do NOT Edit!\n\n");
	  fprintf(file, "style \"ochusha-tree-view-style\"\n{\n");
	  if (font_name != NULL)
	    fprintf(file, "  font_name=\"%s\"\n", font_name);
	  else
	    fprintf(file, "  font_name=\"\"\n");
	  fprintf(file, "}\n\n");
	  fprintf(file, "class \"ThreadlistView\" style \"ochusha-tree-view-style\"\n");
	  fprintf(file, "class \"BoardlistView\" style \"ochusha-tree-view-style\"\n");
	  if (application->thread_view_font_name != NULL)
	    {
	      fprintf(file, "\n\nstyle \"ochusha-bookmark-style\"\n{\n");
	      fprintf(file, "  font_name=\"%s\"\n",
		      application->thread_view_font_name);
	      fprintf(file, "}\n\n");
	      fprintf(file, "widget \"*.mybookmark.GtkLabel\" style \"ochusha-bookmark-style\"\n");
	    }
	  fclose(file);
	  return TRUE;
	}
      else
	close(fd);
    }
  return FALSE;
}


static void
reload_gtkrc(OchushaApplication *application)
{
  char *resource_file = ochusha_config_find_file(&application->config,
						 OCHUSHA_PREFS_GTKRC, NULL);
  if (resource_file != NULL)
    {
      gtk_rc_parse(resource_file);
      gtk_rc_add_default_file(resource_file);
      G_FREE(resource_file);
    }

  gtk_widget_reset_rc_styles(GTK_WIDGET(application->top_level));
}


static void
preview_tree_view_font(OchushaApplication *application, const char *font_name)
{
  if (write_gtkrc_for_tree_view_font(application, font_name))
    reload_gtkrc(application);
}


static void
set_tree_view_font(OchushaApplication *application, const char *font_name)
{
  if (write_gtkrc_for_tree_view_font(application, font_name))
    reload_gtkrc(application);
}


static void
reset_tree_view_font(OchushaApplication *application)
{
  if (write_gtkrc_for_tree_view_font(application, tree_view_font_name))
    reload_gtkrc(application);
}


static void
font_selection_dialog_response_cb(GtkWidget *dialog, int response_id,
				  OchushaApplication *application)
{
  GtkFontSelectionDialog *font_selection;
  gchar *font_name = NULL;

  g_return_if_fail(GTK_IS_FONT_SELECTION_DIALOG(dialog));

  font_selection = GTK_FONT_SELECTION_DIALOG(dialog);

  switch (response_id)
    {
    case GTK_RESPONSE_APPLY:
      font_name = gtk_font_selection_dialog_get_font_name(font_selection);
      preview_tree_view_font(application, font_name);
      g_free(font_name);
      return;

    case GTK_RESPONSE_OK:
      font_name	= gtk_font_selection_dialog_get_font_name(font_selection);
      set_tree_view_font(application, font_name);
      g_free(font_name);
      break;

    case GTK_RESPONSE_CANCEL:
      reset_tree_view_font(application);
      break;
    }

  gtk_widget_hide(dialog);
  gtk_widget_unrealize(dialog);
  gtk_widget_destroy(dialog);

  if (tree_view_font_name != NULL)
    {
      g_free(tree_view_font_name);
      tree_view_font_name = NULL;
    }
}


void
select_tree_view_font(OchushaApplication *application)
{
  PangoContext *context;
  PangoFontDescription *font_desc;
  if (font_selection_dialog != NULL)
    return;

  font_selection_dialog
    = gtk_font_selection_dialog_new(_("Boardlist and Threadlist View Font Selection"));

  context = gtk_widget_get_pango_context(application->boardlist_view);
  font_desc = pango_context_get_font_description(context);

  if (tree_view_font_name != NULL)
    g_free(tree_view_font_name);
  tree_view_font_name = pango_font_description_to_string(font_desc);

  if (tree_view_font_name != NULL && *tree_view_font_name != '\0')
    gtk_font_selection_dialog_set_font_name(GTK_FONT_SELECTION_DIALOG(font_selection_dialog),
					    tree_view_font_name);

  g_signal_connect(G_OBJECT(font_selection_dialog), "response",
		   G_CALLBACK(font_selection_dialog_response_cb), application);
  g_signal_connect(G_OBJECT(font_selection_dialog), "destroy",
		   G_CALLBACK(gtk_widget_destroyed), &font_selection_dialog);

  gtk_widget_show(GTK_FONT_SELECTION_DIALOG(font_selection_dialog)->apply_button);
  gtk_widget_show_all(font_selection_dialog);
}


static void
make_new_thread(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  OchushaBulletinBoard *board;
  BulletinBoardGUIInfo *info;
  GtkWidget *new_thread_dialog;

  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);

  if (application->config.offline)
    return;

  contents_window = PANED_NOTEBOOK(application->contents_window);

  board = (OchushaBulletinBoard *)paned_notebook_get_current_item(contents_window);
  if (board == NULL || IS_VIRTUAL_BOARD(board))
    return;

  if (!ochusha_bulletin_board_is_new_thread_supported(board))
    {
      ochusha_open_url(application,
		       ochusha_bulletin_board_get_base_url(board),
		       FALSE, TRUE);
      return;
    }

  info = ensure_bulletin_board_info(board, application);

  if (info->new_thread_dialog == NULL)
    {
      new_thread_dialog = create_new_thread_dialog(application, board);
      info->new_thread_dialog = new_thread_dialog;
    }
  else
    return;

  g_signal_connect(G_OBJECT(new_thread_dialog), "response",
		   G_CALLBACK(new_thread_dialog_response_cb), application);
  g_signal_connect(G_OBJECT(new_thread_dialog), "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &info->new_thread_dialog);

  gtk_widget_show_all(new_thread_dialog);
}


void
open_current_board_properties(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  OchushaBulletinBoard *board;

  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);

  contents_window = PANED_NOTEBOOK(application->contents_window);

  board = (OchushaBulletinBoard *)paned_notebook_get_current_item(contents_window);
  if (board == NULL)
    return;

  open_board_properties(application, board);
}


void
refresh_current_threadlist(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  OchushaBulletinBoard *board;

  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);

  contents_window = PANED_NOTEBOOK(application->contents_window);

  board = (OchushaBulletinBoard *)paned_notebook_get_current_item(contents_window);
  if (board == NULL)
    return;

  refresh_threadlist_view(application, board);
}


static void
refresh_threadlist_helper(gpointer data, gpointer user_data)
{
  OchushaBulletinBoard *board = (OchushaBulletinBoard *)data;
  OchushaApplication *application = (OchushaApplication *)user_data;
  refresh_threadlist_view(application, board);
  sleep(1);
}


void
refresh_current_threadlist_all(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);
  contents_window = PANED_NOTEBOOK(application->contents_window);
  paned_notebook_foreach_item(contents_window, refresh_threadlist_helper,
			      application);
}


static void
redraw_threadlist_helper(gpointer data, gpointer user_data)
{
  OchushaBulletinBoard *board = (OchushaBulletinBoard *)data;
  OchushaApplication *application = (OchushaApplication *)user_data;
  redraw_threadlist_view(application, board);
}


void
redraw_current_threadlist_all(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);
  contents_window = PANED_NOTEBOOK(application->contents_window);
  paned_notebook_foreach_item(contents_window, redraw_threadlist_helper,
			      application);
  fix_filter_button_state(application);
}


static ThreadlistSearchQueryData search_query_data =
{
  NULL,		/* key */
  TEXT_SEARCH_DIRECTION_DOWN,
  NULL,		/* last_view */
  {},		/* regexp */
  FALSE,	/* regexp_available */
  FALSE,	/* enable_wrap */
  FALSE,	/* match_case */
  FALSE,	/* use_regexp */
};


static void
text_search_window_response_cb(TextSearchWindow *window, int response_id,
			       OchushaApplication *application)
{
  switch (response_id)
    {
    case GTK_RESPONSE_DELETE_EVENT:
      application->threadlist_search_window = NULL;
      /* fall through */

    case GTK_RESPONSE_CANCEL:
      /* XXX: ɥξȥϿ٤ */
      if (search_query_data.last_view != NULL)
	threadlist_view_set_being_searched(search_query_data.last_view, FALSE);
      search_query_data.last_view = NULL;
      gtk_widget_hide(GTK_WIDGET(window));
      return;
    }
}


static gboolean
do_search_on_current_threadlist(OchushaApplication *application,
				gboolean include_current_pos)
{
  GtkWidget *widget;
  ThreadlistView *view;

  g_return_val_if_fail(application != NULL
		       && application->contents_window != NULL, FALSE);

  widget = paned_notebook_get_current_item_view(
				PANED_NOTEBOOK(application->contents_window));
  if (widget == NULL)
    return FALSE;
  g_return_val_if_fail(IS_PANED_NOTEBOOK(widget), FALSE);

  widget = paned_notebook_get_selector(PANED_NOTEBOOK(widget));
  if (widget == NULL)
    return FALSE;

  g_return_val_if_fail(GTK_IS_SCROLLED_WINDOW(widget), FALSE);
  widget = gtk_bin_get_child(GTK_BIN(widget));

  g_return_val_if_fail(IS_THREADLIST_VIEW(widget), FALSE);
  view = THREADLIST_VIEW(widget);

  search_query_data.last_view = view;
  threadlist_view_set_being_searched(view, TRUE);

  return threadlist_view_find_next(view,
				   search_query_data.direction,
				   search_query_data.enable_wrap,
				   include_current_pos,
				   &search_query_data);
}


static gboolean
text_search_window_query_changed_cb(TextSearchWindow *window, const gchar *key,
				    TextSearchDirection direction,
				    gboolean enable_wrap, gboolean match_case,
				    gboolean use_regexp,
				    OchushaApplication *application)
{
  gboolean result;
  gchar *normalized_key;

  application->threadlist_search_direction = direction;
  application->threadlist_search_enable_wrap = enable_wrap;
  application->threadlist_search_match_case = match_case;
  application->threadlist_search_use_regexp = use_regexp;

  if (search_query_data.key != NULL)
    G_FREE(search_query_data.key);

  normalized_key = g_utf8_normalize(key, -1, G_NORMALIZE_ALL);
#if TRACE_MEMORY_USAGE
  {
    gchar *tmp_key = normalized_key;
    normalized_key = G_STRDUP(tmp_key);
    g_free(tmp_key);
  }
#endif
  if (!match_case)
    {
      gchar *case_normalized_key = g_utf8_casefold(normalized_key, -1);
      G_FREE(normalized_key);
#if TRACE_MEMORY_USAGE
      normalized_key = G_STRDUP(case_normalized_key);
      g_free(case_normalized_key);
#else
      normalized_key = case_normalized_key;
#endif
    }
  search_query_data.key = normalized_key;
  search_query_data.direction = direction;
  search_query_data.enable_wrap = enable_wrap;
  search_query_data.match_case = match_case;
  search_query_data.use_regexp = use_regexp;
  if (use_regexp)
    {
      if (search_query_data.regexp_available)
	regfree(&search_query_data.regexp);

      if (regcomp(&search_query_data.regexp,
		  normalized_key, REG_EXTENDED | REG_NEWLINE | REG_NOSUB) == 0)
	{
	  search_query_data.regexp_available = TRUE;
	  result = do_search_on_current_threadlist(application, TRUE);
	}
      else
	{
	  regfree(&search_query_data.regexp);
	  search_query_data.regexp_available = FALSE;
	  result = FALSE;
	}
    }
  else
    result = do_search_on_current_threadlist(application, TRUE);

  return result;
}


static gboolean
text_search_window_find_next_cb(TextSearchWindow *window, const gchar *key,
				TextSearchDirection direction,
				gboolean enable_wrap, gboolean match_case,
				gboolean use_regexp,
				OchushaApplication *application)
{
  return do_search_on_current_threadlist(application, FALSE);
}


static void
setup_threadlist_search_window(OchushaApplication *application)
{
  TextSearchWindow *window;
  if (search_query_data.key != NULL)
    {
      G_FREE(search_query_data.key);
      search_query_data.key = NULL;
    }

  application->threadlist_search_window = text_search_window_new();

  window = TEXT_SEARCH_WINDOW(application->threadlist_search_window);
  text_search_window_set_enable_incremental_search(window, TRUE);
  text_search_window_set_direction(window,
				   application->threadlist_search_direction);
  text_search_window_set_enable_wrap(window,
				     application->threadlist_search_enable_wrap);
  text_search_window_set_match_case(window,
				    application->threadlist_search_match_case);
  text_search_window_set_use_regexp(window,
				    application->threadlist_search_use_regexp);

  g_signal_connect(application->threadlist_search_window, "response",
		   G_CALLBACK(text_search_window_response_cb),
		   application);
  g_signal_connect(application->threadlist_search_window, "query_changed",
		   G_CALLBACK(text_search_window_query_changed_cb),
		   application);
  g_signal_connect(application->threadlist_search_window, "find_next",
		   G_CALLBACK(text_search_window_find_next_cb),
		   application);
#if TRACE_MEMORY_USAGE
  g_signal_connect(application->threadlist_search_window, "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &application->threadlist_search_window);
#endif
  gtk_window_set_title(GTK_WINDOW(application->threadlist_search_window),
		       _("Find in the Current Threadlist"));
  gtk_window_set_transient_for(
			GTK_WINDOW(application->threadlist_search_window),
			application->top_level);
}


void
advance_threadlist_search(OchushaApplication *application,
			  ThreadlistView *view,
			  ThreadlistSearchAction search_action)
{
  switch (search_action)
    {
    case THREADLIST_SEARCH_ACTION_FORWARD:
    case THREADLIST_SEARCH_ACTION_REGEXP_FORWARD:
      search_query_data.direction = THREADLIST_SEARCH_DIRECTION_FORWARD;
      break;

    case THREADLIST_SEARCH_ACTION_BACKWARD:
    case THREADLIST_SEARCH_ACTION_REGEXP_BACKWARD:
      search_query_data.direction = THREADLIST_SEARCH_DIRECTION_BACKWARD;
      break;

    default:
      /* SHOULD NOT reached here */
      /* THREADLIST_SEARCH_ACTION_TERMINATEcaller¦ǽ */
      break;
    }

  if (application->threadlist_search_window == NULL
      || !GTK_WIDGET_VISIBLE(view)
      || search_query_data.last_view != view)
    {
      if (search_query_data.last_view != NULL)
	threadlist_view_set_being_searched(search_query_data.last_view,
					   FALSE);
      start_current_threadlist_search(application);
      search_query_data.last_view = view;
    }
  else
    do_search_on_current_threadlist(application,
				    search_query_data.last_view != view);
}


void
terminate_threadlist_search(OchushaApplication *application,
			    ThreadlistView *view)
{
  TextSearchWindow *window;
  window = TEXT_SEARCH_WINDOW(application->threadlist_search_window);
  if (window != NULL)
    OCHU_OBJECT_REF(window);

  if (window != NULL)
    {
      text_search_window_response(window, GTK_RESPONSE_CANCEL);
      OCHU_OBJECT_UNREF(window);
    }
}


void
start_current_threadlist_search(OchushaApplication *application)
{
  if (application->threadlist_search_window == NULL)
    setup_threadlist_search_window(application);

  gtk_widget_show(application->threadlist_search_window);
  text_search_window_set_key_selected(
		TEXT_SEARCH_WINDOW(application->threadlist_search_window));
}


/*
 * ϤʤΤѰդषBulletinBoardGUIInfoѲưŪ
 * ˻ΤƼΧŪб٤ʤΤʡ
 */
void
fix_filter_button_state(OchushaApplication *application)
{
  PanedNotebook *contents_window;
  OchushaBulletinBoard *board;
  g_return_if_fail(application != NULL
		   && application->contents_window != NULL);
  contents_window = PANED_NOTEBOOK(application->contents_window);
  board = (OchushaBulletinBoard *)paned_notebook_get_current_item(contents_window);
  if (board == NULL)
    return;

#if GTK_MINOR_VERSION <= 2
  gtk_toggle_button_set_active(application->filter_button,
		ensure_bulletin_board_info(board, application)->enable_filter);
#else
  gtk_toggle_tool_button_set_active(application->filter_button,
				    ensure_bulletin_board_info(board,
						application)->enable_filter);
#endif
}


static void
filter_toggled_cb(GObject *unused,
		  OchushaApplication *application)
{
  PanedNotebook *board_pane = PANED_NOTEBOOK(application->contents_window);
  OchushaBulletinBoard *board
    = (OchushaBulletinBoard *)paned_notebook_get_current_item(board_pane);
  BulletinBoardGUIInfo *info;

  if (board == NULL)
    return;

  info = ensure_bulletin_board_info(board, application);
#if GTK_MINOR_VERSION <= 2
  if (info->enable_filter
      == gtk_toggle_button_get_active(application->filter_button))
    return;
#else
  if (info->enable_filter
      == gtk_toggle_tool_button_get_active(application->filter_button))
    return;
#endif

  info->enable_filter = !info->enable_filter;

#if GTK_MINOR_VERSION <= 2
  gtk_toggle_button_set_active(application->filter_button,
			       info->enable_filter);
#else
  gtk_toggle_tool_button_set_active(application->filter_button,
				    info->enable_filter);
#endif

  gdk_threads_leave();
  redraw_threadlist_view(application, board);
  gdk_threads_enter();
}


#if DEBUG_WIDGET_MOST
static void
select_category_cb(BoardlistView *view, OchushaBoardCategory *category,
		   gpointer user_data)
{
  fprintf(stderr, "select_category_cb: %p, %p, %p\n",
	  view, category, user_data);
}
#endif


static void
item_view_required_cb(PanedNotebook *paned_notebook, gpointer item,
		      GtkWidget **item_view, const gchar **title,
		      GtkWidget **tab_label,
		      OchushaApplication *application)
{
  OchushaBulletinBoard *board;

  if (!OCHUSHA_IS_BULLETIN_BOARD(item))
    {
      *tab_label = NULL;
      *item_view = NULL;
      return;
    }

  board = OCHUSHA_BULLETIN_BOARD(item);
  *tab_label = get_tab_label(board->name);
  *item_view = open_bulletin_board(application, board, ICON_LABEL(*tab_label));

  if (*item_view == NULL)
    {
      gtk_object_sink(GTK_OBJECT(*tab_label));
      *tab_label = NULL;
      return;
    }

  if (!IS_VIRTUAL_BOARD(board))
    {
      BOARD_LIST_LOCK
      {
	if (g_slist_find(board_list, board) == NULL)
	  board_list = g_slist_append(board_list, board);
      }
      BOARD_LIST_UNLOCK;
    }

  gtk_widget_show(*item_view);
  *title = board->name;
}


extern guint threadlist_netstat_id;


static void
item_view_being_closed_cb(PanedNotebook *paned_notebook,
			  OchushaBulletinBoard *board,
			  GtkWidget *item_view,
			  OchushaApplication *application)
{
  OchushaAsyncBuffer *buffer
    = snatch_threadlist_source_buffer_for_board(board, application);
  BulletinBoardGUIInfo *info
    = ensure_bulletin_board_info(board, application);

  /*
   * MEMO: ɽƤ륦åȤºݤ˼ΤƤ륿ߥ
   *       򤺤餹
   * 
   * GtkTreeViewŪidleؿϿƤꡢ줬Ƥ֤˥
   * åȤȲ줿åȤidleؿưƤޤ
   * ͡ˡʤΤǡǤϥåȤref_count1äƤ
   * Υץ饤ƥidleؿˤäơˡref_count򸺤餹
   * Υץ饤ƥidleؿưȤȤϡGTK+Ū
   * ȤäƤidleؿλŻλƤȤȤˤʤΤǡ
   * פ
   */
  OCHU_OBJECT_REF(item_view);

  /*
   * MEMO: Ĥ褦ȤƤ륹Ρsubject.txtɤ߹ߡפ
   *       ֥󥰡פԤäƤthreadŪ˽λ롣
   */
  if (buffer != NULL)
    {
      gdk_threads_leave();
      ochusha_async_buffer_suspend(buffer);
      ochusha_async_buffer_terminate(buffer);
      /* ochusha_async_buffer_resume(buffer); */
      gdk_threads_enter();
    }

  if (item_view != NULL)
    {
      if (paned_notebook->number_of_items_shown == 1)
	{
	  PanedNotebook *board_notebook = PANED_NOTEBOOK(item_view);
	  GtkWidget *threadlist_view
	    = gtk_bin_get_child(GTK_BIN(board_notebook->selector));
	  if (gtk_widget_is_focus(threadlist_view))
	    {
	      gtk_widget_grab_focus(GTK_WIDGET(application->boardlist_view));
	    }
	}
    }

  if (buffer != NULL)
    {
      OCHU_OBJECT_UNREF(buffer);

#if 1
      if (IS_VIRTUAL_BOARD(board))
	virtual_board_clean_up_threadlist(VIRTUAL_BOARD(board),
					  board == application->all_threads);
#else
      if (board == application->favorites)
	virtual_board_clean_up_threadlist(VIRTUAL_BOARD(board), FALSE);
      else if (board == application->all_threads)
	virtual_board_clean_up_threadlist(VIRTUAL_BOARD(board), TRUE);
#endif
    }

  if (info->message_id != 0)
    {
      gtk_statusbar_remove(application->statusbar, threadlist_netstat_id,
			   info->message_id);
      info->message_id = 0;
    }
}


static void
item_view_closed_cb(PanedNotebook *paned_notebook, OchushaBulletinBoard *board,
		    GtkWidget *item_view, OchushaApplication *application)
{
  if (paned_notebook_get_current_item_view(paned_notebook) == NULL)
    {
      /* ɽˤǤƤڡĤ줿 */
      if (item_view != NULL && IS_PANED_NOTEBOOK(item_view))
	{
	  application->threadlist_height
	    = gtk_paned_get_position(GTK_PANED(item_view));
	}
    }

  finalize_bulletin_board_view(board, item_view);
  gtk_widget_hide(item_view);
  gtk_widget_unrealize(item_view);
  delayed_g_object_unref(G_OBJECT(item_view));

  if (ochusha_bulletin_board_trylock_thread_list(board))
    {
      write_threadlist(board, application);
      ochusha_bulletin_board_unlock_thread_list(board);
    }
}


static void
page_switched_cb(PanedNotebook *paned_notebook,
		 GtkWidget *previous_page, GtkWidget *new_page,
		 OchushaApplication *application)
{
  g_return_if_fail((previous_page != NULL && IS_PANED_NOTEBOOK(previous_page))
		   || (new_page != NULL && IS_PANED_NOTEBOOK(new_page)));

  if (previous_page != NULL)
    {
      GtkWidget *scrolled_window;

      application->threadlist_height
	= gtk_paned_get_position(GTK_PANED(previous_page));
      scrolled_window
	= paned_notebook_get_selector(PANED_NOTEBOOK(previous_page));

      if (scrolled_window != NULL)
	{
	  ThreadlistView *threadlist_view
	    = THREADLIST_VIEW(gtk_bin_get_child(GTK_BIN(scrolled_window)));
	  if (search_query_data.last_view == threadlist_view)
	    {
	      threadlist_view_set_being_searched(threadlist_view, FALSE);
	      search_query_data.last_view = NULL;
	    }
	}
    }

  if (new_page != NULL)
    {
      OchushaBulletinBoard *board
	= (OchushaBulletinBoard *)paned_notebook_get_item(paned_notebook, new_page);
      BulletinBoardGUIInfo *info;
      GtkWidget *tmp_widget;

      g_return_if_fail(board != NULL);
      info = ensure_bulletin_board_info(board, application);
      paned_notebook_set_tooltips(PANED_NOTEBOOK(new_page),
				  application->thread_tab_enable_tooltips);
      paned_notebook_set_paned_style(PANED_NOTEBOOK(new_page),
				     application->threadlist_pane_style);
      gtk_paned_set_position(GTK_PANED(new_page),
			     application->threadlist_height);
#if GTK_MINOR_VERSION <= 2
      gtk_toggle_button_set_active(application->filter_button,
				   info->enable_filter);
#else
      gtk_toggle_tool_button_set_active(application->filter_button,
					info->enable_filter);
#endif
      tmp_widget = paned_notebook_get_selector(PANED_NOTEBOOK(new_page));
      tmp_widget = gtk_bin_get_child(GTK_BIN(tmp_widget));
      threadlist_view_set_default_open_in_tab(THREADLIST_VIEW(tmp_widget),
					application->default_open_in_tab);
      threadlist_view_set_remove_hidden_thread(THREADLIST_VIEW(tmp_widget),
					       application->hide_hidden_threads);
    }
}


static void
page_double_clicked_cb(PanedNotebook *paned_notebook,
		       OchushaBulletinBoard *board, GtkWidget *item_view,
		       OchushaApplication *application)
{
  refresh_threadlist_view(application, board);
}


static OchushaBulletinBoard *
board_new_cb(OchushaBBSTable *table, int bbs_type, const gchar *name,
	     const char *base_url, OchushaApplication *application)
{
  if (bbs_type == OCHUSHA_BBS_TYPE_VIRTUAL_BOARD)
    {
      OchushaBulletinBoard *board
	= virtual_board_new(name, base_url, application);
#if 1
      if (strcmp(base_url, "http:///all-threads") == 0)
	application->all_threads = board;
      else
	{
	  if (strcmp(base_url, "http:///favorites") == 0)
	    application->favorites = board;
	  application->favorites_list
	    = g_slist_append(application->favorites_list, board);
	}
#else
      if (strcmp(base_url, "http:///favorites") == 0)
	application->favorites = board;
      else if (strcmp(base_url, "http:///all-threads") == 0)
	application->all_threads = board;
#endif
      return board;
    }
  return NULL;	/* ¾ΤϤǤ */
}


static void
boardlist_read_category_element_cb(OchushaBBSTable *table,
				   OchushaBoardCategory *category,
				   GHashTable *category_attributes,
				   OchushaApplication *application)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);

  info->expanded = ochusha_utils_get_attribute_boolean(category_attributes,
						       "expanded");
  info->category_type = ochusha_utils_get_attribute_int(category_attributes,
							"category_type");
  if (info->category_type == CATEGORY_TYPE_OCHUSHA_MY_FOLDER)
    {
      application->my_folder = category;
      ochusha_board_category_set_name(category, _("My Folder"));
    }

  return;
}


static void
boardlist_read_board_element_cb(OchushaBBSTable *table,
				OchushaBulletinBoard *board,
				GHashTable *board_attributes,
				OchushaApplication *application)
{
  BulletinBoardGUIInfo *info = ensure_bulletin_board_info(board, application);
  info->show_mailto_mode
    = (ThreadViewMailtoMode)ochusha_utils_get_attribute_int(board_attributes,
							   "show_mailto_mode");
  if (info->show_mailto_mode < THREAD_VIEW_MAILTO_MODE_DEFAULT
      || info->show_mailto_mode > THREAD_VIEW_MAILTO_MODE_HIDE)
    info->show_mailto_mode = THREAD_VIEW_MAILTO_MODE_DEFAULT;

  info->use_static_threadlist
    = ochusha_utils_get_attribute_boolean(board_attributes,
					 "use_static_threadlist");

  info->enable_filter = ochusha_utils_get_attribute_boolean(board_attributes,
							    "enable_filter");
  info->filter.rule = ochusha_utils_get_attribute_string(board_attributes,
							 "filter.rule");
  info->filter.ignore_threshold
    = ochusha_utils_get_attribute_int(board_attributes,
				      "filter.ignore_threshold");
  info->last_name = ochusha_utils_get_attribute_string(board_attributes,
						       "last_name");
  info->last_mail = ochusha_utils_get_attribute_string(board_attributes,
						       "last_mail");
  if (!info->recover_mode)
    info->recover_mode = ochusha_utils_get_attribute_boolean(board_attributes,
							     "recover_mode");

  info->board_local_a_bone
    = ochusha_utils_get_attribute_boolean(board_attributes,
					  "board_local_a_bone");
  info->a_bone_by_name
    = ochusha_utils_get_attribute_boolean(board_attributes, "a_bone_by_name");
  info->a_bone_by_name_pattern
    = ochusha_utils_get_attribute_string(board_attributes,
					 "a_bone_by_name_pattern");
  info->a_bone_by_id
    = ochusha_utils_get_attribute_boolean(board_attributes,
					  "a_bone_by_id");
  info->a_bone_by_id_pattern
    = ochusha_utils_get_attribute_string(board_attributes,
					 "a_bone_by_id_pattern");
  info->a_bone_by_content
    = ochusha_utils_get_attribute_boolean(board_attributes,
					  "a_bone_by_content");
  info->a_bone_by_content_pattern
    = ochusha_utils_get_attribute_string(board_attributes,
					 "a_bone_by_content_pattern");
  return;
}


#define OUTPUT_CATEGORY_ATTRIBUTE_BOOLEAN(gzfile, category, attribute)	\
  do									\
    {									\
      if ((category)->attribute)					\
	{								\
	  gzprintf(gzfile,						\
		   "      <attribute name=\"" #attribute	"\">\n"	\
		   "        <boolean val=\"%s\"/>\n"			\
		   "      </attribute>\n",				\
		   (category)->attribute ? "true" : "false");		\
	}								\
    } while (0)

#define OUTPUT_CATEGORY_ATTRIBUTE_INT(gzfile, category, attribute)	\
  do									\
    {									\
      if ((category)->attribute)					\
	{								\
	  gzprintf(gzfile,						\
		   "      <attribute name=\"" #attribute	"\">\n"	\
		   "        <int val=\"%d\"/>\n"			\
		   "      </attribute>\n", (category)->attribute);	\
	}								\
    } while (0)


static void
boardlist_write_category_element_cb(OchushaBBSTable *table,
				    OchushaBoardCategory *category,
				    gzFile boardlist_xml,
				    OchushaApplication *application)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);

  OUTPUT_CATEGORY_ATTRIBUTE_BOOLEAN(boardlist_xml, info, expanded);
  OUTPUT_CATEGORY_ATTRIBUTE_INT(boardlist_xml, info, category_type);
}


#define OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(gzfile, board, attribute)	\
  do									\
    {									\
      if ((board)->attribute)						\
	{								\
	  gzprintf(gzfile,						\
		  "        <attribute name=\"" #attribute "\">\n"	\
		  "          <boolean val=\"%s\"/>\n"			\
		  "        </attribute>\n",				\
		  (board)->attribute ? "true" : "false");		\
	}								\
    } while (0)


#define OUTPUT_BOARD_ATTRIBUTE_INT(gzfile, board, attribute)		\
  do									\
    {									\
      if ((board)->attribute)						\
	{								\
	  gzprintf(gzfile,						\
		  "        <attribute name=\"" #attribute "\">\n"	\
		  "          <int val=\"%d\"/>\n"			\
		  "        </attribute>\n", (board)->attribute);	\
	}								\
    } while (0)


#define OUTPUT_BOARD_ATTRIBUTE_STRING(gzfile, board, attribute)		\
  do									\
    {									\
      if ((board)->attribute != NULL)					\
	{								\
	  gchar *text = g_markup_escape_text((board)->attribute, -1);	\
	  gzprintf(gzfile,						\
		   "        <attribute name=\"" #attribute "\">\n"	\
		   "          <string>%s</string>\n"			\
		   "        </attribute>\n", text);			\
	  g_free(text);							\
	}								\
    } while (0)


static void
boardlist_write_board_element_cb(OchushaBBSTable *table,
				 OchushaBulletinBoard *board,
				 gzFile boardlist_xml,
				 OchushaApplication *application)
{
  BulletinBoardGUIInfo *info = ensure_bulletin_board_info(board, application);
  OUTPUT_BOARD_ATTRIBUTE_INT(boardlist_xml, info, show_mailto_mode);
  OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info, last_name);
  OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info, last_mail);
  if (info->recover_mode)
    OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, recover_mode);

  OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, enable_filter);
  if (info->filter.rule != NULL)
    {
      OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info, filter.rule);
      OUTPUT_BOARD_ATTRIBUTE_INT(boardlist_xml, info, filter.ignore_threshold);
    }

  OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, use_static_threadlist);

  if (info->board_local_a_bone)
    {
      OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, board_local_a_bone);
      OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, a_bone_by_name);
      OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info,
				    a_bone_by_name_pattern);
      OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, a_bone_by_id);
      OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info, a_bone_by_id_pattern);
      OUTPUT_BOARD_ATTRIBUTE_BOOLEAN(boardlist_xml, info, a_bone_by_content);
      OUTPUT_BOARD_ATTRIBUTE_STRING(boardlist_xml, info,
				    a_bone_by_content_pattern);
    }

  return;
}


static void
write_threadlist(gpointer data, gpointer user_data)
{
  OchushaBulletinBoard *board = (OchushaBulletinBoard *)data;
  OchushaApplication *application = (OchushaApplication *)user_data;

#if 0
  fprintf(stderr, "write_threadlist: base_url=%s\n", board->base_url);
#endif
  /* XXX: 顼 */
  if (!ochusha_bulletin_board_write_threadlist_xml(board, &application->config,
						application->session_subdir))
    fprintf(stderr, "Couldn't write threadlist.xml.\n");

  if (application->session_subdir != NULL
      && !ochusha_bulletin_board_write_threadlist_xml(board,
						      &application->config,
						      NULL))
    fprintf(stderr, "Couldn't write threadlist.xml.\n");
}


static OchushaBoardCategory *last_expanded_category = NULL;


static void
category_expanded_cb(BoardlistView *view, OchushaBoardCategory *category,
		     OchushaApplication *application)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);

  info->expanded = TRUE;
  if (application->expand_only_one_category)
    {
      if (last_expanded_category != NULL && last_expanded_category != category)
	boardlist_view_collapse_category(view, last_expanded_category);

      last_expanded_category = category;
    }
}


static void
category_collapsed_cb(BoardlistView *view, OchushaBoardCategory *category,
		      OchushaApplication *application)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);
  info->expanded = FALSE;
}


static void
toggle_hide_category_cb(BoardlistView *view, OchushaBoardCategory *category,
			OchushaApplication *application)
{
  category->hidden = !(category->hidden);
}


static void
toggle_hide_board_cb(BoardlistView *view, OchushaBulletinBoard *board,
		     OchushaApplication *application)
{
  board->hidden = !(board->hidden);
}


static gboolean
kill_category_cb(BoardlistView *view, OchushaBoardCategory *category,
		 OchushaApplication *application)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);
  if (info->category_type != CATEGORY_TYPE_OCHUSHA_MY_FOLDER)
    {
      category->killed = TRUE;
      return TRUE;
    }
  return FALSE;
}


static gboolean
kill_board_cb(BoardlistView *view, OchushaBoardCategory *category,
	      OchushaBulletinBoard *board, OchushaApplication *application)
{
  if (category != NULL)
    {
      if (IS_VIRTUAL_BOARD(board))
	{
	  BoardCategoryGUIInfo *info = ensure_board_category_info(category);
	  const char *base_url = ochusha_bulletin_board_get_base_url(board);
	  if (info->category_type == CATEGORY_TYPE_OCHUSHA_MY_FOLDER
	      && (strcmp(base_url, "http:///favorites") == 0
		  || strcmp(base_url, "http:///all-threads") == 0))
	    return FALSE;
	}
      ochusha_board_category_remove_board(category, board);
    }
  return TRUE;
}


static void
copy_board_url_cb(BoardlistView *view, OchushaBulletinBoard *board,
		  OchushaApplication *application)
{
  ochusha_clipboard_set_text(application,
			     ochusha_bulletin_board_get_base_url(board));
}


static void
open_properties_cb(BoardlistView *view, GObject *object,
		   OchushaApplication *application)
{
  if (!OCHUSHA_IS_BULLETIN_BOARD(object))
    return;

  open_board_properties(application, OCHUSHA_BULLETIN_BOARD(object));
}


static void
uri_dropped_cb(BoardlistView *view, OchushaBoardCategory *category,
	       OchushaBulletinBoard *board, const char *uri,
	       GtkTreeViewDropPosition pos,
	       OchushaApplication *application)
{
  OchushaBulletinBoard *dropped_board = NULL;
  char *dropped_thread_id = NULL;
  gboolean is_kako_html = FALSE;
  BoardCategoryGUIInfo *category_info;
#if 0
  fprintf(stderr, "uri_dropped_cb: category(%s)/board(%s) uri=\"%s\"\n",
	  ochusha_board_category_get_name(category),
	  ochusha_bulletin_board_get_name(board), uri);
#endif
  if (!OCHUSHA_IS_BOARD_CATEGORY(category))
    return;

  category_info = ensure_board_category_info(category);
  if (category_info->category_type == CATEGORY_TYPE_BBS_MENU_CATEGORY)
    return;

  if (ochusha_bbs_table_check_url(application->table, uri,
				  &dropped_board, NULL, NULL,
				  &dropped_thread_id, NULL, NULL,
				  &is_kako_html))
    {
      if (dropped_board == NULL)
	return;

      if (dropped_thread_id == NULL)
	{
	  /* Ĥɥåפ줿 */
	  /* 褺categoryΥĤͤù */
	  OchushaBulletinBoard *tmp_board
	    = ochusha_board_category_lookup_board_by_url(category,
							 dropped_board->base_url);
	  if (tmp_board != NULL)
	    {
	      OCHU_OBJECT_REF(tmp_board);
	      ochusha_board_category_remove_board(category, tmp_board);
	      boardlist_view_remove_board(view, category, tmp_board);
	    }
	  
	  ochusha_board_category_add_board(category, dropped_board);

	  if (tmp_board != NULL)
	    OCHU_OBJECT_UNREF(tmp_board);
	  
	  boardlist_view_add_board(view, dropped_board,
				   category, board);
	}
      else
	{
	  /* 줬ɥåפ줿 */
	  OchushaBBSThread *dropped_thread
	    = ochusha_bulletin_board_lookup_bbs_thread_by_id(dropped_board,
							     dropped_thread_id);
	  if (dropped_thread != NULL)
	    bookmark_thread(board, dropped_thread, application);
	  G_FREE(dropped_thread_id);
	}
    }
}


static gboolean
prepare_category(OchushaBoardCategory *category, gpointer user_data)
{
  ensure_board_category_info(category);
  return TRUE;
}


static gboolean
prepare_board(OchushaBulletinBoard *board, gpointer user_data)
{
  ensure_bulletin_board_info(board, (OchushaApplication *)user_data);
  return TRUE;
}


static OchushaBulletinBoard *
process_board_move(OchushaBulletinBoard *old_board,
		   OchushaBulletinBoard *new_board,
		   gpointer user_data)
{
  OchushaApplication *application = (OchushaApplication *)user_data;
  const char *new_base_url = ochusha_bulletin_board_get_base_url(new_board);
  ochusha_bulletin_board_process_board_move(old_board, new_base_url,
					    &application->config);

  return old_board;
}


void
redraw_boardlist(OchushaApplication *application)
{
  int status;

  if ((status = pthread_mutex_trylock(&board_list_lock)) != 0)
    {
      if (status == EBUSY)
	return;		/* İι1åɤ */

      fprintf(stderr, "Internal error: invalid mutex.\n");
      abort();
    }

  setup_boardlist(application);

  if (pthread_mutex_unlock(&board_list_lock) != 0)
    {
      fprintf(stderr, "Couldn't unlock a mutex.\n");
      abort();
    }
}


static void
build_boardlist(WorkerThread *employee, gpointer args)
{
  int status;
  OchushaAsyncBuffer *buffer;

  OchushaApplication *application = (OchushaApplication *)args;
  OchushaBBSTable *table = application->table;

  if ((status = pthread_mutex_trylock(&board_list_lock)) != 0)
    {
      if (status == EBUSY)
	return;		/* İι1åɤ */

      fprintf(stderr, "Internal error: invalid mutex.\n");
      abort();
    }

  buffer = ochusha_utils_2ch_get_bbsmenu(application->broker);

  if (buffer == NULL)
    {
      /* XXX: 顼 */
      fprintf(stderr, "Couldn't get %s\n", application->config.bbsmenu_url);
      goto cleanup_before_return;
      return;
    }

  if (ochusha_utils_2ch_analyze_bbsmenu(table, buffer,
					prepare_category,
					prepare_board,
					process_board_move,
					application))
    {
      gdk_threads_enter();
      setup_boardlist(application);
      gdk_threads_leave();
      ochusha_bbs_table_write_boardlist_xml(table, &application->config,
					    application->session_subdir);
      if (application->session_subdir != NULL)
	ochusha_bbs_table_write_boardlist_xml(table, &application->config,
					      NULL);
    }

 cleanup_before_return:
  if (buffer != NULL)
    OCHU_OBJECT_UNREF(buffer);

  if (pthread_mutex_unlock(&board_list_lock) != 0)
    {
      fprintf(stderr, "Couldn't unlock a mutex.\n");
      abort();
    }
}


static void
restore_expanded_state(OchushaBoardCategory *category, BoardlistView *view)
{
  BoardCategoryGUIInfo *info = ensure_board_category_info(category);
  if (info->expanded)
    boardlist_view_expand_category(view, category);
}


static void
setup_boardlist(OchushaApplication *application)
{
  ensure_my_folder(application);
  boardlist_view_open(BOARDLIST_VIEW(application->boardlist_view),
		      application->table->category_list,
		      application->hide_hidden_boardlist_items);
  if (!application->expand_only_one_category)
    g_slist_foreach(application->table->category_list,
		    (GFunc)restore_expanded_state,
		    application->boardlist_view);

  gtk_widget_show_all(application->boardlist_view);
  gtk_widget_show_all(application->contents_window);
}


static void
make_new_thread_button_cb(GObject *unused, OchushaApplication *application)
{
  make_new_thread(application);
}


static void
open_board_properties_button_cb(GObject *unused,
				OchushaApplication *application)
{
  open_current_board_properties(application);
}


static void
refresh_board_button_cb(GObject *unused, OchushaApplication *application)
{
  refresh_current_threadlist(application);
}


static void
start_threadlist_search_button_cb(GObject *unused,
				  OchushaApplication *application)
{
  start_current_threadlist_search(application);
}


static void
new_thread_dialog_response_cb(GtkWidget *dialog, int response_id,
			      OchushaApplication *application)
{
  OchushaBulletinBoard *board = g_object_get_data(G_OBJECT(dialog), "board");
  gboolean result;
  guint message_id = (guint)(long)g_object_get_data(G_OBJECT(dialog),
						    "statusbar_message");

  if (response_id == GTK_RESPONSE_REJECT)
    {
      ResponseEditor *editor = g_object_get_data(G_OBJECT(dialog), "editor");
      response_editor_clear_response(editor);
      return;
    }

  if (message_id != 0)
    {
      gtk_statusbar_remove(application->statusbar,
			   new_thread_netstat_id, message_id);
      g_object_set_data(G_OBJECT(dialog), "statusbar_message", NULL);
    }

  if (response_id == GTK_RESPONSE_OK)
    {
      ResponseEditor *editor = g_object_get_data(G_OBJECT(dialog), "editor");
      GtkEntry *title_entry = g_object_get_data(G_OBJECT(dialog), "title");
      const gchar *title = gtk_entry_get_text(title_entry);
      gchar *message;
      OchushaBBSResponse response;

      response.name = response_editor_get_name(editor);
      response.mailto = response_editor_get_mail(editor);
      message = response_editor_get_response(editor);
      response.content = message;

      if (title != NULL && *title != '\0'
	  && message != NULL && *message != '\0')
	{
	  gchar statusbar_message[4096];

	  snprintf(statusbar_message, 4096,
		   _("Creating new thread at %s board."), board->name);
	  message_id = gtk_statusbar_push(application->statusbar,
					  new_thread_netstat_id,
					  statusbar_message);

	  gdk_threads_leave();
	  result = ochusha_bulletin_board_create_new_thread(board,
						application->broker,
						title, &response,
						application->dolib_use_id_for_posting);
	  gdk_threads_enter();
	  g_free(message);
	  if (result)
	    {
	      BulletinBoardGUIInfo *board_info
		= ensure_bulletin_board_info(board, application);

	      gtk_statusbar_remove(application->statusbar,
				   new_thread_netstat_id, message_id);

	      if (board_info->last_name == NULL
		  || strcmp(board_info->last_name, response.name) != 0)
		{
		  if (board_info->last_name != NULL)
		    G_FREE(board_info->last_name);
		  board_info->last_name = G_STRDUP(response.name);
		}

	      if (board_info->last_mail == NULL
		  || strcmp(board_info->last_mail, response.mailto) != 0)
		{
		  if (board_info->last_mail != NULL)
		    G_FREE(board_info->last_mail);
		  board_info->last_mail = G_STRDUP(response.mailto);
		}

	      if (application->last_name == NULL
		  || strcmp(application->last_name, response.name) != 0)
		{
		  if (application->last_name != NULL)
		    G_FREE(application->last_name);
		  application->last_name = G_STRDUP(response.name);
		}

	      if (application->last_mail == NULL
		  || strcmp(application->last_mail, response.mailto) != 0)
		{
		  if (application->last_mail != NULL)
		    G_FREE(application->last_mail);
		  application->last_mail = G_STRDUP(response.mailto);
		}
	    }
	  else
	    {
	      gtk_statusbar_remove(application->statusbar,
				   new_thread_netstat_id, message_id);
	      snprintf(statusbar_message, 4096,
		       _("Creation of new thread at %s board failed."),
		       board->name);
	      message_id = gtk_statusbar_push(application->statusbar,
					      new_thread_netstat_id,
					      statusbar_message);
	      g_object_set_data(G_OBJECT(dialog), "statusbar_message",
				(gpointer)(long)message_id);
	    }
	}
      else
	{
	  /* ʸʤϥ顼ˤʤ餷 */
	  result = TRUE;
	  response_id = GTK_RESPONSE_CANCEL;
	}
    }
  else
    result = TRUE;

  if (result)
    {
      if (response_id == GTK_RESPONSE_OK)
	refresh_threadlist_view(application, board);
      gtk_widget_hide(dialog);
      gtk_widget_unrealize(dialog);
      gtk_widget_destroy(dialog);
    }
  else
    ochusha_open_url(application,
		     ochusha_bulletin_board_get_base_url(board),
		     FALSE, TRUE);
}


static GtkWidget *
create_new_thread_dialog(OchushaApplication *application,
			 OchushaBulletinBoard *board)
{
  GtkWidget *new_thread_dialog;
  GtkWidget *dialog_vbox;
#if 0
  GtkWidget *url_hbox;
  GtkWidget *url_entry;
#endif
  GtkWidget *title_hbox;
  GtkWidget *title_entry;
  GtkWidget *board_label;
  GtkWidget *editor;
  ResponseEditor *response_editor;
  BulletinBoardGUIInfo *board_info = ensure_bulletin_board_info(board,
								application);
  gboolean has_name = FALSE;
  gboolean has_mail = FALSE;
  gchar board_name_buffer[2048];

  new_thread_dialog = gtk_dialog_new_with_buttons(_("Create a New Thread"),
					 application->top_level,
					 GTK_DIALOG_DESTROY_WITH_PARENT,
					 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					 GTK_STOCK_CLEAR, GTK_RESPONSE_REJECT,
					 GTK_STOCK_OK, GTK_RESPONSE_OK,
					 NULL);
  gtk_window_set_position(GTK_WINDOW(new_thread_dialog), GTK_WIN_POS_CENTER);
  gtk_window_set_default_size(GTK_WINDOW(new_thread_dialog), -1, 250);

  dialog_vbox = GTK_DIALOG(new_thread_dialog)->vbox;

  snprintf(board_name_buffer, 2048, _("Create a New Thread on %s board"),
	   ochusha_bulletin_board_get_name(board));
  board_label = gtk_label_new(board_name_buffer);
  gtk_label_set_justify(GTK_LABEL(board_label), GTK_JUSTIFY_LEFT);
  gtk_widget_show(board_label);
  gtk_box_pack_start(GTK_BOX(dialog_vbox), board_label, FALSE, FALSE, 10);

#if 0
  url_hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(url_hbox), gtk_label_new("URL: "),
		     FALSE, FALSE, 0);
  url_entry = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(url_entry),
		     ochusha_bulletin_board_get_base_url(board));
  gtk_editable_set_editable(GTK_EDITABLE(url_entry), FALSE);
  GTK_WIDGET_UNSET_FLAGS(url_entry, GTK_CAN_FOCUS);
  gtk_box_pack_start(GTK_BOX(url_hbox), url_entry, TRUE, TRUE, 0);
  gtk_widget_show_all(url_hbox);
  gtk_box_pack_start(GTK_BOX(dialog_vbox), url_hbox, FALSE, FALSE, 0);
#endif

  title_hbox = gtk_hbox_new(FALSE, 5);
  gtk_box_pack_start(GTK_BOX(title_hbox), gtk_label_new(_("Title: ")),
		     FALSE, FALSE, 0);
  title_entry = gtk_entry_new();
  gtk_editable_set_editable(GTK_EDITABLE(title_entry), TRUE);
  GTK_WIDGET_SET_FLAGS(title_entry, GTK_CAN_FOCUS);
  gtk_box_pack_start(GTK_BOX(title_hbox), title_entry, TRUE, TRUE, 0);
  gtk_widget_show_all(title_hbox);
  gtk_box_pack_start(GTK_BOX(dialog_vbox), title_hbox, FALSE, FALSE, 0);

  editor = response_editor_new();
  response_editor = RESPONSE_EDITOR(editor);

  if (board_info->last_name != NULL)
    {
      response_editor_add_name(response_editor, board_info->last_name);
      if (!has_name)
	{
	  response_editor_set_name(response_editor, board_info->last_name);
	  has_name = TRUE;
	}
    }

  if (board_info->last_mail != NULL)
    {
      response_editor_add_mail(response_editor, board_info->last_mail);
      if (!has_mail)
	{
	  response_editor_set_mail(response_editor, board_info->last_mail);
	  has_mail = TRUE;
	}
    }

  if (application->last_name != NULL)
    {
      response_editor_add_name(response_editor, application->last_name);
      if (!has_name)
	response_editor_set_name(response_editor, application->last_name);
    }

  if (application->last_mail != NULL)
    {
      response_editor_add_mail(response_editor, application->last_mail);
      if (!has_mail)
	response_editor_set_mail(response_editor, application->last_mail);
    }

  gtk_widget_show(editor);

#if GTK_MINOR_VERSION > 2
  setup_aalist(application, response_editor);
#endif

  gtk_box_pack_start(GTK_BOX(dialog_vbox), editor, TRUE, TRUE, 0);

  g_object_set_data(G_OBJECT(new_thread_dialog), "editor", editor);
  g_object_set_data(G_OBJECT(new_thread_dialog), "title", title_entry);
  g_object_set_data(G_OBJECT(new_thread_dialog), "board", board);

  return new_thread_dialog;
}

