/*
 * Copyright (c) 2003-2004 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: boardlist_view.h,v 1.18.2.1 2004/06/11 12:02:05 fuyu Exp $
 */

#ifndef _BOARDLIST_VIEW_H_
#define _BOARDLIST_VIEW_H_

#include "ochusha.h"

#include <gtk/gtk.h>

#ifdef __cplusplus
extern "C" {
#endif


#define BOARDLIST_VIEW_TYPE			(boardlist_view_get_type())
#define BOARDLIST_VIEW(obj)			(G_TYPE_CHECK_INSTANCE_CAST((obj), BOARDLIST_VIEW_TYPE, BoardlistView))
#define BOARDLIST_VIEW_CLASS(klass)		(G_TYPE_CHECK_CLASS_CAST((klass), BOARDLIST_VIEW_TYPE, BoardlistViewClass))
#define IS_BOARDLIST_VIEW(obj)			(G_TYPE_CHECK_INSTANCE_TYPE((obj), BOARDLIST_VIEW_TYPE))
#define IS_BOARDLIST_VIEW_CLASS(klass)		(G_TYPE_CHECK_CLASS_TYPE((klass), BOARDLIST_VIEW_TYPE))
#define BOARDLIST_VIEW_GET_CLASS(obj)		(G_TYPE_INSTANCE_GET_CLASS((obj), BOARDLIST_VIEW_TYPE, BoardlistViewClass))

#define BOARDLIST_VIEW_ACTION_TYPE		(boardlist_view_action_get_type())

typedef enum
{
  TOGGLE_HIDDEN_STATE,
  KILL_ITEM,
  ADVANCE_VIEW,
  BACK_VIEW,
} BoardlistViewAction;


GType boardlist_view_action_get_type(void) G_GNUC_CONST;


typedef struct _BoardlistView BoardlistView;
typedef struct _BoardlistViewClass BoardlistViewClass;

struct _BoardlistView
{
  GtkTreeView view;

  GtkTreeModel *model;

  GtkTreeSelection *selection;

  GtkTreePath *last_event_path;

  gboolean hide_hidden_items;
  gboolean default_open_in_tab;
};


struct _BoardlistViewClass
{
  GtkTreeViewClass parent_class;

  void (*open_threadlist_view)(BoardlistView *view,
			       OchushaBulletinBoard *board,
			       gboolean in_tab, gboolean with_browser);

  void (*select_category)(BoardlistView *view, OchushaBoardCategory *category);
  void (*category_expanded)(BoardlistView *view,
			    OchushaBoardCategory *category);
  void (*category_collapsed)(BoardlistView *view,
			     OchushaBoardCategory *category);

  void (*toggle_hide_category)(BoardlistView *view,
			       OchushaBoardCategory *category);
  void (*toggle_hide_board)(BoardlistView *view,
			    OchushaBulletinBoard *board);
  gboolean (*kill_category)(BoardlistView *view,
			    OchushaBoardCategory *category);
  gboolean (*kill_board)(BoardlistView *view, OchushaBoardCategory *category,
			 OchushaBulletinBoard *board);
  void (*copy_board_url)(BoardlistView *view, OchushaBulletinBoard *board);
  void (*open_properties)(BoardlistView *view, GObject *object);

  gboolean (*advance_view)(BoardlistView *view);
  gboolean (*back_view)(BoardlistView *view);

  void (*do_on_selection)(BoardlistView *view, int what_to_do);
  void (*do_on_item_at_cursor)(BoardlistView *view,
			       BoardlistViewAction what_to_do);

  void (*uri_dropped)(BoardlistView *view,
		      OchushaBoardCategory *category,
		      OchushaBulletinBoard *board, const char *uri,
		      GtkTreeViewDropPosition pos);
};


GType boardlist_view_get_type(void) G_GNUC_CONST;
GtkWidget *boardlist_view_new(void);
void boardlist_view_open(BoardlistView *view, GSList *category_list,
			 gboolean really_hide_hidden_items);

void boardlist_view_add_category(BoardlistView *view,
				 OchushaBoardCategory *category,
				 OchushaBoardCategory *prev_category);
void boardlist_view_remove_category(BoardlistView *view,
				    OchushaBoardCategory *category);
void boardlist_view_add_board(BoardlistView *view,
			      OchushaBulletinBoard *board,
			      OchushaBoardCategory *category,
			      OchushaBulletinBoard *prev_board);
void boardlist_view_remove_board(BoardlistView *view,
				 OchushaBoardCategory *category,
				 OchushaBulletinBoard *board);

gboolean boardlist_view_get_default_open_in_tab(BoardlistView *view);
void boardlist_view_set_default_open_in_tab(BoardlistView *view,
					    gboolean default_open_in_tab);

void boardlist_view_expand_category(BoardlistView *view,
				    OchushaBoardCategory *category);
void boardlist_view_collapse_category(BoardlistView *view,
				      OchushaBoardCategory *category);

#ifdef __cplusplus
} // extern "C"
#endif

#endif	/* _BOARDLIST_VIEW_H_ */
