/*
 * Copyright (c) 2003-2004 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: bulletin_board_ui.h,v 1.28.2.2 2004/09/24 16:14:52 fuyu Exp $
 */

#ifndef _BULLETIN_BOARD_UI_H_
#define _BULLETIN_BOARD_UI_H_

#include "ochusha_async_buffer.h"

#include "ochusha.h"
#include "ochusha_ui.h"

#include "boardlist_view.h"
#include "icon_label.h"
#include "response_editor.h"
#include "threadlist_filter.h"
#include "threadlist_view.h"

#include <gtk/gtk.h>

#if HAVE_ONIG_ONIGPOSIX_H
# include <onig/onigposix.h>
#else
# include <onigposix.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif


typedef struct _BulletinBoardGUIInfo
{
  OchushaAsyncBuffer *threadlist_source_buffer;

  GtkWidget *properties_dialog;

  GtkWidget *new_thread_dialog;

  /* ᥤ򥹥ɽϤʸȤɽ뤫ɤĤΥǥե */
  ThreadViewMailtoMode show_mailto_mode;

  /* ξ֤ɽstatusbarΥåID */
  guint message_id;

  /* Ǹν񤭹߻̾ */
  gchar *last_name;

  /* Ǹν񤭹߻Υᥤ */
  gchar *last_mail;

  /* ܡ */
  gboolean board_local_a_bone;

  gboolean a_bone_by_name;
  gchar *a_bone_by_name_pattern;

  gboolean a_bone_by_id;
  gchar *a_bone_by_id_pattern;

  gboolean a_bone_by_content;
  gchar *a_bone_by_content_pattern;

  gboolean recover_mode;

  /* ե륿 */
  ThreadlistFilter filter;

  gboolean enable_filter;

  /* ¼VirtualBoard */
  gboolean use_static_threadlist;


  /* Ʊ */
  volatile gboolean is_busy;
  volatile gboolean redraw_required;
} BulletinBoardGUIInfo;


typedef struct _ThreadlistSearchQueryData
{
  gchar *key;
  int direction;
  ThreadlistView *last_view;

  regex_t regexp;
  gboolean regexp_available;

  gboolean enable_wrap;
  gboolean match_case;
  gboolean use_regexp;
} ThreadlistSearchQueryData;


void prepare_board_ui_initialization(OchushaApplication *application);
void initialize_board_ui(OchushaApplication *application);

BulletinBoardGUIInfo *ensure_bulletin_board_info(OchushaBulletinBoard *board,
					OchushaApplication *application);
GtkWidget *open_bulletin_board(OchushaApplication *application,
			       OchushaBulletinBoard *board,
			       IconLabel *tab_label);
void finalize_bulletin_board_view(OchushaBulletinBoard *board,
				  GtkWidget *widget);
void ochusha_open_thread(OchushaApplication *application,
			 OchushaBulletinBoard *board, const char *thread_id,
			 int res_num, gboolean in_tab);
void ochusha_open_kako_thread(OchushaApplication *application,
			      OchushaBulletinBoard *board, const char *url,
			      gboolean in_tab);

void write_response(OchushaApplication *application, int res_num);

void refresh_threadlist_view(OchushaApplication *application,
			     OchushaBulletinBoard *board);
void redraw_threadlist_view(OchushaApplication *application,
			    OchushaBulletinBoard *board);
OchushaAsyncBuffer *snatch_threadlist_source_buffer_for_board(
					OchushaBulletinBoard *board,
					OchushaApplication *application);
void update_threadlist_entry_style(OchushaApplication *application,
				   ThreadlistView *view,
				   OchushaBBSThread *thread);

void refresh_current_thread(OchushaApplication *application);
void start_thread_search(OchushaApplication *application);

void jump_to_bookmark_of_current_thread(OchushaApplication *application);
void go_to_the_last_response_of_current_thread(OchushaApplication *application);
void go_to_the_first_response_of_current_thread(OchushaApplication *application);

void bookmark_thread(OchushaBulletinBoard *board, OchushaBBSThread *thread,
		     OchushaApplication *application);
void bookmark_remove_thread(OchushaBulletinBoard *board,
			    OchushaBBSThread *thread,
			    OchushaApplication *application);

#if GTK_MINOR_VERSION > 2
void setup_aalist(OchushaApplication *application, ResponseEditor *editor);
#endif

#ifdef __cplusplus
} // extern "C"
#endif

#endif	/* _BULLETIN_BOARD_UI_H_ */
