/*
 * Copyright (c) 2003-2004 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: ochusha_ui.c,v 1.114.2.22 2005/06/06 19:00:13 fuyu Exp $
 */

#define _OCHUSHA_UI_C_
#include "config.h"

#include "ochusha_private.h"
#include "ochusha.h"
#include "ochusha_bbs_table.h"
#include "ochusha_network_broker.h"
#include "ochusha_utils_2ch.h"
#include "worker.h"

#include "ochusha_ui.h"
#include "bbs_thread_ui.h"
#include "boardlist_ui.h"
#include "bulletin_board_ui.h"
#include "download_ui.h"
#include "image_ui.h"

#include "icon_label.h"
#include "paned_notebook.h"
#include "regex_utils.h"
#include "session.h"
#include "virtual_board.h"
#include "xsmp.h"
#include "ugly_gtk2utils.h"

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixdata.h>

#include <pthread.h>

#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>

#include <zlib.h>

#include "missing_image_24.h"
#include "write_response_16.h"
#include "write_response_24.h"

#include "blue0.xpm"
#include "blue1.xpm"
#include "blue2.xpm"
#include "blue3.xpm"
#include "orange.xpm"
#include "black.xpm"
#include "red.xpm"

#include "ochusha48.h"

#undef NEW_YEAR_SPECIAL
#undef XMAS_SPECIAL

#include "ochusha-banner.h"

#ifdef NEW_YEAR_SPECIAL
#include "ochusha-banner1.h"
#endif

#ifdef XMAS_SPECIAL
#include "ochusha-banner2.h"
#endif


static void setup_ochusha_stock_items(void);
static void setup_tab_label_icons(void);
static void delete_event_cb(GtkWidget *widget, GdkEventAny *event,
			    gpointer user_data);
#if GTK_MINOR_VERSION <= 2
static void main_menu_callback(gpointer data, guint action, GtkWidget *widget);
#endif
static void ochusha_quit(OchushaApplication *application);

static void write_optional_prefs(gzFile file, gpointer user_data);
static void configure_application(OchushaApplication *application);
static void read_optional_prefs(GHashTable *pref_attrs, gpointer user_data);

static void access_started_cb(OchushaAsyncBuffer *buffer,
			      IconLabel *tab_label);
static void access_progressed_cb(OchushaAsyncBuffer *buffer,
				 int bytes_read, int bytes_total,
				 IconLabel *tab_label);
static void access_finished_cb(OchushaAsyncBuffer *buffer,
			       IconLabel *tab_label);
static void access_failed_cb(OchushaAsyncBuffer *buffer,
			     OchushaNetworkBrokerFailureReason reason_code,
			     const gchar *reason,
			     IconLabel *tab_label);
static void output_log_cb(OchushaNetworkBroker *broker, const gchar *message,
			  OchushaApplication *application);

#if ENABLE_DEBUG_CONSOLE
static void open_debug_console(OchushaApplication *application);
#endif
static void open_preference_dialog(OchushaApplication *application);
static void open_about_dialog(OchushaApplication *application);
static void about_dialog_response_cb(GtkWidget *dialog, int response_id,
				     gpointer unused);

static void initialize_open_url_dialog(OchushaApplication *application);
static void open_url_combo_entry_activate_cb(GtkWidget *entry,
					     GtkDialog *dialog);
static void open_url_dialog_response_cb(GtkWidget *dialog, int response_id,
					OchushaApplication *application);
static void popup_open_url_dialog(OchushaApplication *application);


#if GTK_MINOR_VERSION <= 2
static GtkItemFactory *main_menu_item_factory = NULL;
#else
static GtkUIManager *ui_manager = NULL;
#endif

static GdkPixbuf *tab_label_icons[7];
static GQuark tab_label_id;
static GQuark tab_label_icon_counter_id;
GQuark application_id;

static pthread_mutex_t main_ui_lock;

#define MAIN_UI_LOCK					\
  if (pthread_mutex_lock(&main_ui_lock) != 0)		\
    {							\
      fprintf(stderr, "Couldn't lock a mutex.\n");	\
      abort();						\
    }

#define MAIN_UI_UNLOCK					\
  if (pthread_mutex_unlock(&main_ui_lock) != 0)		\
    {							\
      fprintf(stderr, "Couldn't unlock a mutex.\n");	\
      abort();						\
    }

static GtkWidget *preference_dialog = NULL;
static GtkWidget *about_dialog = NULL;
static GdkPixbuf *banner_pixbuf = NULL;
#ifdef NEW_YEAR_SPECIAL
static GdkPixbuf *banner1_pixbuf = NULL;
#endif
#ifdef XMAS_SPECIAL
static GdkPixbuf *banner2_pixbuf = NULL;
#endif
static GtkWidget *open_url_dialog = NULL;
static GtkWidget *open_url_combo = NULL;
static GtkWidget *open_url_in_tab_button = NULL;
static GList *open_url_history = NULL;

#if ENABLE_DEBUG_CONSOLE
static gboolean enable_debug_outputs = FALSE;


static GtkWidget *debug_console = NULL;
static GtkTextBuffer *log_buffer = NULL;
static GtkTextIter log_tail_iter;
#endif




#ifdef HAVE_SIGNAL_H
static OchushaApplication *application_global = NULL;

static RETSIGTYPE
terminate_on_signal(int sig)
{
  OchushaApplication *application = application_global;
#if 0
  fprintf(application->log_file, "terminate_on_signal: sig=%d\n", sig);
  fflush(application->log_file);
#endif
  if (application == NULL)
    {
      fprintf(stderr, "Ochusha terminated on unexpected signal.\n");
      exit(0);
    }

  if (application->sm_client_id != NULL)
    {
      ochusha_session_discard_states(application, application->sm_client_id);
      application->session_subdir = NULL;
    }

  if (!application->states_saved)
    ochusha_session_save_states(application);

  gtk_main_quit();
}
#endif


#if GTK_MINOR_VERSION <= 2
enum
{
  MAIN_MENU_QUIT = 1,
  MAIN_MENU_OPEN_URL,
  MAIN_MENU_WORK_OFFLINE,
  MAIN_MENU_REFRESH_BOARDLIST,
  MAIN_MENU_CREATE_NEW_CATEGORY,
  MAIN_MENU_ADD_BOARD,
  MAIN_MENU_CREATE_NEW_VIRTUAL_BOARD,
  MAIN_MENU_HIDE_HIDDEN_BOARDLIST_ITEMS,
  MAIN_MENU_REFRESH_THREAD,
  MAIN_MENU_REFRESH_THREAD_ALL,
  MAIN_MENU_REFRESH_THREADLIST,
  MAIN_MENU_REFRESH_THREADLIST_ALL,
  MAIN_MENU_USE_BUILTIN_CALCULATE_WEIGHT,
  MAIN_MENU_SORT_THREADLIST_BY_WEIGHT,
  MAIN_MENU_HIDE_HIDDEN_THREADS,
  MAIN_MENU_DEBUG_CONSOLE,
  MAIN_MENU_OPTIONS,
  MAIN_MENU_SELECT_BOARDLIST_THREADLIST_FONTS,
  MAIN_MENU_SELECT_THREAD_FONTS,
  MAIN_MENU_ABOUT,
};


#else


static void
menu_open_url_cb(GtkAction *action, OchushaApplication *application)
{
  popup_open_url_dialog(application);
}


static void
menu_quit_cb(GtkAction *action, OchushaApplication *application)
{
  ochusha_quit(application);
}


static void
menu_refresh_boardlist_cb(GtkAction *action, OchushaApplication *application)
{
  refresh_boardlist(application);
}


static void
menu_create_new_category_cb(GtkAction *action, OchushaApplication *application)
{
  ochusha_create_new_category(application);
}


static void
menu_add_board_cb(GtkAction *action, OchushaApplication *application)
{
  ochusha_add_board(application, NULL);
}


static void
menu_create_new_virtual_board_cb(GtkAction *action,
				 OchushaApplication *application)
{
  ochusha_create_new_virtual_board(application);
}


static void
menu_refresh_current_thread_cb(GtkAction *action,
			       OchushaApplication *application)
{
  refresh_current_thread(application);
}


static void
menu_refresh_all_threads_cb(GtkAction *action, OchushaApplication *application)
{
#if 0
  refresh_current_thread_all(application);
#endif
}


static void
menu_refresh_current_threadlist_cb(GtkAction *action,
				   OchushaApplication *application)
{
  refresh_current_threadlist(application);
}


static void
menu_refresh_all_threadlists_cb(GtkAction *action,
				OchushaApplication *application)
{
  refresh_current_threadlist_all(application);
}


static void
menu_debug_console_cb(GtkAction *action, OchushaApplication *application)
{
#if ENABLE_DEBUG_CONSOLE
  open_debug_console(application);
#endif
}


static void
menu_options_cb(GtkAction *action, OchushaApplication *application)
{
  open_preference_dialog(application);
}


static void
menu_select_list_fonts_cb(GtkAction *action, OchushaApplication *application)
{
  select_tree_view_font(application);
}


static void
menu_select_thread_fonts_cb(GtkAction *action, OchushaApplication *application)
{
  select_thread_view_font(application);
}


static void
menu_about_cb(GtkAction *action, OchushaApplication *application)
{

  open_about_dialog(application);
}


static void
menu_work_offline_toggled_cb(GtkToggleAction *action,
			     OchushaApplication *application)
{
  application->config.offline
    = gtk_toggle_action_get_active(action);
}


static void
menu_hide_hidden_boardlist_items_toggled_cb(GtkToggleAction *action,
					    OchushaApplication *application)
{
  application->hide_hidden_boardlist_items
    = gtk_toggle_action_get_active(action);

  if (application->boardlist_view != NULL)
    redraw_boardlist(application);
}


static void
menu_use_builtin_calculate_weight_toggled_cb(GtkToggleAction *action,
					     OchushaApplication *application)
{
  application->use_builtin_calculate_weight
    = gtk_toggle_action_get_active(action);
}


static void
menu_sort_threadlist_by_weight_toggled_cb(GtkToggleAction *action,
					  OchushaApplication *application)
{
  application->sort_threadlist_by_weight
    = gtk_toggle_action_get_active(action);
}


static void
menu_hide_hidden_threads_toggled_cb(GtkToggleAction *action,
				    OchushaApplication *application)
{
  application->hide_hidden_threads
    = gtk_toggle_action_get_active(action);
}
#endif


void
ochusha_init_application(OchushaApplication *application)
{
#if GTK_MINOR_VERSION <= 2
  GtkItemFactoryEntry menu_items[] = {
    {
      _("/_File"),		/* menu path */
      NULL,			/* accelerator */
      NULL,			/* callback_func */
      0,			/* act */
      (char *)"<Branch>"	/* type */
    },
    {
      _("/File/_Open URL"),
      NULL,
      main_menu_callback,
      MAIN_MENU_OPEN_URL,
      NULL
    },
    {
      _("/File/-------"),
      NULL,
      NULL,
      0,
      (char *)"<Separator>"
    },
    {
      _("/File/_Work Offline"),
      NULL,
      main_menu_callback,
      MAIN_MENU_WORK_OFFLINE,
      (char *)"<CheckItem>",
    },
    {
      _("/File/_Quit"),
      (char *)"<control>Q",
      ochusha_quit,
      MAIN_MENU_QUIT,
      NULL
    },
    {
      _("/_Board"),
      NULL,
      NULL,
      0,
      (char *)"<Branch>"
    },
    {
      _("/Board/_Refresh Boardlist"),
      NULL,
      main_menu_callback,
      MAIN_MENU_REFRESH_BOARDLIST,
      NULL
    },

    {
      _("/Board/-------"),
      NULL,
      NULL,
      0,
      (char *)"<Separator>"
    },
#if 0	/* 褺ȥ */
    {
      _("/Board/_Create New Category"),
      NULL,
      main_menu_callback,
      MAIN_MENU_CREATE_NEW_CATEGORY,
      NULL,
    },
    {
      _("/Board/_Add Board"),
      NULL,
      main_menu_callback,
      MAIN_MENU_ADD_BOARD,
      NULL,
    },
#endif
    {
      _("/Board/Create New _Virtual Board"),
      NULL,
      main_menu_callback,
      MAIN_MENU_CREATE_NEW_VIRTUAL_BOARD,
      NULL,
    },
    {
      _("/Board/-------"),
      NULL,
      NULL,
      0,
      (char *)"<Separator>"
    },
    {
      _("/Board/Hide Hidden Boardlist Items"),
      NULL,
      main_menu_callback,
      MAIN_MENU_HIDE_HIDDEN_BOARDLIST_ITEMS,
      (char *)"<CheckItem>",
    },
    {
      _("/_Thread"),
      NULL,
      NULL,
      0,
      (char *)"<Branch>"
    },
    {
      _("/Thread/_Refresh Thread"),
      (char *)"F5",
      main_menu_callback,
      MAIN_MENU_REFRESH_THREAD,
      NULL
    },
    {
      _("/Thread/Refresh _All Threads"),
      NULL,
      main_menu_callback,
      MAIN_MENU_REFRESH_THREAD_ALL,
      NULL
    },
    {
      _("/Thread/-------"),
      NULL,
      NULL,
      0,
      (char *)"<Separator>"
    },
    {
      _("/Thread/Refresh _Threadlist"),
      NULL,
      main_menu_callback,
      MAIN_MENU_REFRESH_THREADLIST,
      NULL
    },
    {
      _("/Thread/Refresh All Threadlists"),
      NULL,
      main_menu_callback,
      MAIN_MENU_REFRESH_THREADLIST_ALL,
      NULL
    },
    {
      _("/Thread/-------"),
      NULL,
      NULL,
      0,
      (char *)"<Separator>"
    },
    {
      _("/Thread/Use Builtin Calculate Weight"),
      NULL,
      main_menu_callback,
      MAIN_MENU_USE_BUILTIN_CALCULATE_WEIGHT,
      (char *)"<CheckItem>",
    },
    {
      _("/Thread/_Sort Threadlist by Weight"),
      NULL,
      main_menu_callback,
      MAIN_MENU_SORT_THREADLIST_BY_WEIGHT,
      (char *)"<CheckItem>",
    },
    {
      _("/Thread/Hide Hidden Threads"),
      NULL,
      main_menu_callback,
      MAIN_MENU_HIDE_HIDDEN_THREADS,
      (char *)"<CheckItem>",
    },
    {
      _("/Tools"),
      NULL,
      NULL,
      0,
      (char *)"<Branch>"
    },
#if ENABLE_DEBUG_CONSOLE
    {
      _("/Tools/Debug Console"),
      NULL,
      main_menu_callback,
      MAIN_MENU_DEBUG_CONSOLE,
      NULL,
    },
    {
      _("/Tools/-------"),
      NULL,
      NULL,
      0,
      (char *)"<Separator>"
    },
#endif
    {
      _("/Tools/Options"),
      NULL,
      main_menu_callback,
      MAIN_MENU_OPTIONS,
      NULL,
    },
    {
      _("/Tools/Select Boardlist and Threadlist View Fonts"),
      NULL,
      main_menu_callback,
      MAIN_MENU_SELECT_BOARDLIST_THREADLIST_FONTS,
      NULL,
    },
    {
      _("/Tools/Select Thread View Fonts"),
      NULL,
      main_menu_callback,
      MAIN_MENU_SELECT_THREAD_FONTS,
      NULL,
    },
    {
      _("/_Help"),
      NULL,
      NULL,
      0,
      (char *)"<Branch>"
    },
    {
      _("/Help/_About"),
      NULL,
      main_menu_callback,
      MAIN_MENU_ABOUT,
      NULL
    }
  };

  int nmenu_items = sizeof(menu_items) / sizeof(menu_items[0]);
  GtkWidget *item;
#else
  static GtkActionEntry menubar_entries[] =
    {
      {
	"FileMenu", NULL,
	N_("_File"), NULL,
	NULL,
	NULL,
      },
      {
	"BoardMenu", NULL,
	N_("_Board"), NULL,
	NULL,
	NULL,
      },
      {
	"ThreadMenu", NULL,
	N_("_Thread"), NULL,
	NULL,
	NULL,
      },
      {
	"ToolsMenu", NULL,
	N_("Tools"), NULL,
	NULL,
	NULL,
      },
      {
	"HelpMenu", NULL,
	N_("_Help"), NULL,
	NULL,
	NULL,
      },
      {
	"OpenURL", GTK_STOCK_OPEN,
	N_("_Open URL"), NULL,
	N_("Open Specified URL"),
	(GCallback)menu_open_url_cb
      },
      {
	"Quit", GTK_STOCK_QUIT,
	N_("_Quit"), "<control>Q",
	N_("Quit Ochusha"),
	(GCallback)menu_quit_cb
      },
      {
	"RefreshBoardlist", GTK_STOCK_REFRESH,
	N_("_Refresh Boardlist"), NULL,
	N_("Refresh Boardlist"),
	(GCallback)menu_refresh_boardlist_cb
      },
      {
	"CreateNewCategory", GTK_STOCK_NEW,
	N_("_Create New Category"), NULL,
	N_("Create New Category"),
	(GCallback)menu_create_new_category_cb
      },
      {
	"AddBoard", GTK_STOCK_NEW,
	N_("_Add Board"), NULL,
	N_("Add Board"),
	(GCallback)menu_add_board_cb
      },
      {
	"CreateNewVirtualBoard", GTK_STOCK_NEW,
	N_("Create New _Virtual Board"), NULL,
	N_("Create New Virtual Board"),
	(GCallback)menu_create_new_virtual_board_cb
      },
      {
	"RefreshThread", GTK_STOCK_REFRESH,
	N_("_Refresh Thread"), "F5",
	N_("Refresh Current Thread"),
	(GCallback)menu_refresh_current_thread_cb
      },
      {
	"RefreshAllThreads", GTK_STOCK_REFRESH,
	N_("Refresh _All Threads"), NULL,
	N_("Refresh All Threads"),
	(GCallback)menu_refresh_all_threads_cb
      },
      {
	"RefreshThreadlist", GTK_STOCK_REFRESH,
	N_("Refresh _Threadlist"), NULL,
	N_("Refresh Current Threadlist"),
	(GCallback)menu_refresh_current_threadlist_cb
      },
      {
	"RefreshAllThreadlists", GTK_STOCK_REFRESH,
	N_("Refresh All Threadlists"), NULL,
	N_("Refresh All Threadlists"),
	(GCallback)menu_refresh_all_threadlists_cb
      },
      {
	"DebugConsole", NULL,
	N_("Debug Console"), NULL,
	N_("Open Debug Console"),
	(GCallback)menu_debug_console_cb
      },
      {
	"Options", GTK_STOCK_PREFERENCES,
	N_("Options"), NULL,
	N_("Open Preference Dialog"),
	(GCallback)menu_options_cb
      },
      {
	"SelectListViewFonts", GTK_STOCK_SELECT_FONT,
	N_("Select Boardlist and Threadlist View Fonts"), NULL,
	N_("Open Font Selection Dialog for Boardlist and Threadlist View"),
	(GCallback)menu_select_list_fonts_cb
      },
      {
	"SelectThreadViewFonts", GTK_STOCK_SELECT_FONT,
	N_("Select Thread View Fonts"), NULL,
	N_("Open Font Selection Dialog for Thread View"),
	(GCallback)menu_select_thread_fonts_cb
      },
      {
	"About", GTK_STOCK_HELP,
	N_("_About"), NULL,
	N_("Open About Dialog"),
	(GCallback)menu_about_cb
      }
    };

  static GtkToggleActionEntry menubar_toggle_entries[] =
    {
      {
	"WorkOffline", NULL,
	N_("_Work Offline"), NULL,
	N_("Work Offline"),
	(GCallback)menu_work_offline_toggled_cb,
	FALSE
      },
      {
	"HideHiddenBoardlistItems", NULL,
	N_("Hide Hidden Boardlist Items"), NULL,
	N_("Hide Hidden Boardlist Items"),
	(GCallback)menu_hide_hidden_boardlist_items_toggled_cb,
	FALSE
      },
      {
	"UseBuiltinCalculateWeight", NULL,
	N_("Use Builtin Calculate Weight"), NULL,
	N_("Use Builtin Calculate Weight"),
	(GCallback)menu_use_builtin_calculate_weight_toggled_cb,
	FALSE
      },
      {
	"SortThreadlistByWeight", GTK_STOCK_SORT_ASCENDING,
	N_("_Sort Threadlist by Weight"), NULL,
	N_("Sort Threadlist by Weight"),
	(GCallback)menu_sort_threadlist_by_weight_toggled_cb,
	FALSE
      },
      {
	"HideHiddenThreads", NULL,
	N_("Hide Hidden Threads"), NULL,
	N_("Hide Hidden Threads"),
	(GCallback)menu_hide_hidden_threads_toggled_cb,
	FALSE,
      }
    };

  static const char *menubar_ui =
    "<ui>"
    "  <menubar name=\"MainMenu\">"
    "    <menu action=\"FileMenu\">"
    "      <menuitem action=\"OpenURL\"/>"
    "      <separator/>"
    "      <menuitem action=\"WorkOffline\"/>"
    "      <menuitem action=\"Quit\"/>"
    "    </menu>"
    "    <menu action=\"BoardMenu\">"
    "      <menuitem action=\"RefreshBoardlist\"/>"
    "      <separator/>"
#if 0	/* 褺ȥ */
    "      <menuitem action=\"CreateNewCategory\"/>"
    "      <menuitem action=\"AddBoard\"/>"
#endif
    "      <menuitem action=\"CreateNewVirtualBoard\"/>"
    "      <separator/>"
    "      <menuitem action=\"HideHiddenBoardlistItems\"/>"
    "    </menu>"
    "    <menu action=\"ThreadMenu\">"
    "      <menuitem action=\"RefreshThread\"/>"
    "      <menuitem action=\"RefreshAllThreads\"/>"
    "      <separator/>"
    "      <menuitem action=\"RefreshThreadlist\"/>"
    "      <menuitem action=\"RefreshAllThreadlists\"/>"
    "      <separator/>"
    "      <menuitem action=\"UseBuiltinCalculateWeight\"/>"
    "      <menuitem action=\"SortThreadlistByWeight\"/>"
    "      <menuitem action=\"HideHiddenThreads\"/>"
    "    </menu>"
    "    <menu action=\"ToolsMenu\">"
    "      <menuitem action=\"DebugConsole\"/>"
    "      <separator/>"
    "      <menuitem action=\"Options\"/>"
    "      <menuitem action=\"SelectListViewFonts\"/>"
    "      <menuitem action=\"SelectThreadViewFonts\"/>"
    "    </menu>"
    "    <menu action=\"HelpMenu\">"
    "      <menuitem action=\"About\"/>"
    "    </menu>"
    "  </menubar>"
    "</ui>";

  GtkActionGroup *action_group;
  GError *error;
#endif

  GtkAccelGroup *accel_group;

  static GdkGeometry geometry = { 0, 0, 0, 0, 0, 0, 0, 0, 0.0, 0.0, 0 };

  if (pthread_mutex_init(&main_ui_lock, NULL) != 0)
    {
      fprintf(stderr, "Couldn't init a mutex.\n");
      abort();
    }

  if (application->sm_client_id != NULL)
    gdk_set_sm_client_id(application->sm_client_id);

  setup_ochusha_stock_items();
  setup_tab_label_icons();

  banner_pixbuf
    = gdk_pixbuf_new_from_inline(-1, ochusha_banner_inline, FALSE, NULL);
#ifdef NEW_YEAR_SPECIAL
  banner1_pixbuf
    = gdk_pixbuf_new_from_inline(-1, ochusha_banner1_inline, FALSE, NULL);
#endif
#ifdef XMAS_SPECIAL
  banner2_pixbuf
    = gdk_pixbuf_new_from_inline(-1, ochusha_banner2_inline, FALSE, NULL);
#endif

  configure_application(application);

  application->top_level = GTK_WINDOW(gtk_window_new(GTK_WINDOW_TOPLEVEL));
  g_signal_connect(G_OBJECT(application->top_level), "delete_event",
		   G_CALLBACK(delete_event_cb), application);
#if TRACE_MEMORY_USAGE
  g_signal_connect(G_OBJECT(application->top_level), "destroy",
		   G_CALLBACK(gtk_widget_destroyed), &application->top_level);
#endif

  gtk_window_set_title(application->top_level, _("Ochusha"));

  gtk_window_set_icon(application->top_level,
		      gdk_pixbuf_new_from_inline(-1, ochusha48_inline, FALSE,
						 NULL));
  gtk_widget_set_size_request(GTK_WIDGET(application->top_level),
			      application->width, application->height);
  gtk_widget_realize(GTK_WIDGET(application->top_level));
  if (application->x_pos >= 0 && application->y_pos >= 0)
    {
      gtk_window_move(application->top_level,
		      application->x_pos, application->y_pos);
      gtk_widget_queue_resize(GTK_WIDGET(application->top_level));
    }

  application->main_window = GTK_BOX(gtk_vbox_new(FALSE, 0));
  gtk_container_add(GTK_CONTAINER(application->top_level),
		    GTK_WIDGET(application->main_window));
#if TRACE_MEMORY_USAGE
  g_signal_connect(G_OBJECT(application->main_window), "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &application->main_window);
#endif
  gtk_window_set_geometry_hints(application->top_level,
				GTK_WIDGET(application->main_window),
				&geometry, GDK_HINT_MIN_SIZE);

  /* menu barŹ */
#if GTK_MINOR_VERSION <= 2
  accel_group = gtk_accel_group_new();
  main_menu_item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>",
						accel_group);
  gtk_item_factory_create_items(main_menu_item_factory,
				nmenu_items, menu_items, application);
  gtk_window_add_accel_group(application->top_level, accel_group);

  application->menu_bar = gtk_item_factory_get_widget(main_menu_item_factory,
						      "<main>");
  /* ̤ʵǽΥ˥塼򻦤 */
  item = gtk_item_factory_get_widget_by_action(main_menu_item_factory,
					       MAIN_MENU_REFRESH_THREAD_ALL);
  if (item != NULL)
    gtk_widget_set_sensitive(item, FALSE);

  /* ߤȿǤ롣 */
  gtk_check_menu_item_set_active(
		GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget_by_action(
				main_menu_item_factory,
				MAIN_MENU_WORK_OFFLINE)),
		application->config.offline);

  gtk_check_menu_item_set_active(
		GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget_by_action(
				main_menu_item_factory,
				MAIN_MENU_HIDE_HIDDEN_BOARDLIST_ITEMS)),
		application->hide_hidden_boardlist_items);

  gtk_check_menu_item_set_active(
		GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget_by_action(
				main_menu_item_factory,
				MAIN_MENU_USE_BUILTIN_CALCULATE_WEIGHT)),
		application->use_builtin_calculate_weight);

  gtk_check_menu_item_set_active(
		GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget_by_action(
				main_menu_item_factory,
				MAIN_MENU_SORT_THREADLIST_BY_WEIGHT)),
		application->sort_threadlist_by_weight);

  gtk_check_menu_item_set_active(
		GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget_by_action(
				main_menu_item_factory,
				MAIN_MENU_HIDE_HIDDEN_THREADS)),
		application->hide_hidden_threads);

#else
  action_group = gtk_action_group_new("MenubarActions");
  gtk_action_group_set_translation_domain(action_group, PACKAGE_NAME);
  gtk_action_group_add_actions(action_group,
			       menubar_entries,
			       G_N_ELEMENTS(menubar_entries),
			       application);
  gtk_action_group_add_toggle_actions(action_group,
				      menubar_toggle_entries,
				      G_N_ELEMENTS(menubar_toggle_entries),
				      application);
  ui_manager = gtk_ui_manager_new();
  gtk_ui_manager_insert_action_group(ui_manager, action_group, 0);

  accel_group = gtk_ui_manager_get_accel_group(ui_manager);
  gtk_window_add_accel_group(application->top_level, accel_group);

  error = NULL;
  if (!gtk_ui_manager_add_ui_from_string(ui_manager, menubar_ui, -1, &error))
    {
      g_message("building menus failed: %s\n", error->message);
      g_error_free(error);
      exit(EXIT_FAILURE);
    }

  application->menu_bar = gtk_ui_manager_get_widget(ui_manager, "/MainMenu");

  {
    /* ̤ʵǽΥ˥塼򻦤 */
    GtkAction *action = gtk_action_group_get_action(action_group,
						    "RefreshAllThreads");
    g_object_set((GObject *)action, "sensitive", FALSE, NULL);
  }

  /* ߤȿǤ */
  gtk_toggle_action_set_active(
	GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						"WorkOffline")),
	application->config.offline);
  gtk_toggle_action_set_active(
	GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						"HideHiddenBoardlistItems")),
	application->hide_hidden_boardlist_items);
			  
  gtk_toggle_action_set_active(
	GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						"UseBuiltinCalculateWeight")),
	application->use_builtin_calculate_weight);

  gtk_toggle_action_set_active(
	GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						"SortThreadlistByWeight")),
	application->sort_threadlist_by_weight);

  gtk_toggle_action_set_active(
	GTK_TOGGLE_ACTION(gtk_action_group_get_action(action_group,
						"HideHiddenThreads")),
	application->hide_hidden_threads);
#endif

#if TRACE_MEMORY_USAGE
  g_signal_connect(G_OBJECT(application->menu_bar), "destroy",
		   G_CALLBACK(gtk_widget_destroyed), &application->menu_bar);
#endif

  gtk_box_pack_start(application->main_window,
		     application->menu_bar, FALSE, TRUE, 0);
#if TRACE_MEMORY_USAGE
  g_signal_connect(G_OBJECT(application->menu_bar), "destroy",
		   G_CALLBACK(gtk_widget_destroyed), &application->menu_bar);
#endif

  /* ν */
  initialize_open_url_dialog(application);


  /* GUIν */
  prepare_boardlist_ui_initialization(application);
  prepare_board_ui_initialization(application);
  prepare_thread_ui_initialization(application);

  application->statusbar = GTK_STATUSBAR(gtk_statusbar_new());

  /* İν */
  initialize_boardlist(application);

  gtk_box_pack_start(application->main_window,
		     application->contents_window, TRUE, TRUE, 0);
#if TRACE_MEMORY_USAGE
  g_signal_connect(G_OBJECT(application->contents_window), "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &application->contents_window);
#endif

  initialize_board_ui(application);
  initialize_thread_ui(application);
  initialize_download_ui(application);
  initialize_image_ui(application);

  gtk_box_pack_end(application->main_window,
		   GTK_WIDGET(application->statusbar), FALSE, TRUE, 0);
#if TRACE_MEMORY_USAGE
  g_signal_connect(G_OBJECT(application->statusbar), "destroy",
		   G_CALLBACK(gtk_widget_destroyed),
		   &application->statusbar);
#endif

  application->clipboard = NULL;
  application->clipboard_primary = NULL;

#if ENABLE_DEBUG_CONSOLE
  log_buffer = gtk_text_buffer_new(NULL);
  gtk_text_buffer_get_start_iter(log_buffer, &log_tail_iter);
#endif

  gtk_widget_show_all(GTK_WIDGET(application->top_level));

#ifdef HAVE_SIGNAL_H
  application_global = application;
  if (signal(SIGTERM, terminate_on_signal) == SIG_ERR)
    {
      fprintf(stderr,
	      "sigaction failed: ochusha will continue to work anyway.\n");
    }
#endif
}


void
ochusha_clipboard_set_text(OchushaApplication *application, const gchar *text)
{
  MAIN_UI_LOCK
  {
    if (application->clipboard == NULL)
      {
	application->clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
	application->clipboard_primary
	  = gtk_clipboard_get(GDK_SELECTION_PRIMARY);
      }
#if DEBUG_GUI_MOST
    fprintf(stderr, "setting text on clipboard: \"%s\"\n", text);
#endif
    gtk_clipboard_set_text(application->clipboard_primary, text, -1);
    gtk_clipboard_set_text(application->clipboard, text, -1);
  }
  MAIN_UI_UNLOCK
}


#define COMMAND_LINE_BUFFER_SIZE	4096


static char *
validate_web_browser_template(const char *template)
{
  /* templateʸ%s⡹ĤʤȤݾڤʸ
   * ԡ֤
   */
  char *result = G_STRDUP(template);
  char *tmp_pos = result;
  gboolean percent_s_exist = FALSE;

  while (*tmp_pos != '\0')
    {
      if (*tmp_pos++ == '%')
	{
	  if (*tmp_pos == 's')
	    {
	      if (percent_s_exist)
		{
		  tmp_pos[-1] = ' ';
		  tmp_pos[0] = ' ';
		}
	      else
		percent_s_exist = TRUE;
	      tmp_pos++;
	    }
	}
    }

  if (strcmp(template, result) != 0)
    {
      fprintf(stderr, "multiple %%s isn't allowed\n");
      fprintf(stderr, "PRE validate: \"%s\"\n", template);
      fprintf(stderr, "POST validate: \"%s\"\n", result);
    }

  return result;
}


static gboolean
is_shell_escape_required(const char *command_line)
{
  gboolean quoted = FALSE;
  gboolean double_quoted = FALSE;
  const char *tmp_pos = command_line;

  while (*tmp_pos != '\0')
    {
      if (*tmp_pos == '#')
	{
	  if (!quoted && !double_quoted)
	    return TRUE;
	}
      else if (*tmp_pos == '"')
	{
	  if (!quoted)
	    double_quoted = !double_quoted;
	}
      else if (*tmp_pos == '\'')
	{
	  if (!double_quoted)
	    quoted = !quoted;
	}
      tmp_pos++;
    }
  return FALSE;
}


void
ochusha_open_url(OchushaApplication *application, const char *url,
		 gboolean in_tab, gboolean use_web_browser)
{
  OchushaBulletinBoard *board;
  char *board_url = NULL;
  int bbs_type = OCHUSHA_BBS_TYPE_UNKNOWN;
  char *thread_id = NULL;
  unsigned int from = 0;
  unsigned int to = 0;
  gboolean is_kako_html = FALSE;
  char *scheme;
  char command_line[COMMAND_LINE_BUFFER_SIZE];

  char *real_url = ochusha_utils_hack_known_special_url(url);

  if (real_url != NULL)
    url = real_url;

  if (!use_web_browser
      && ochusha_bbs_table_check_url(application->table, url,
				     &board, &board_url, &bbs_type,
				     &thread_id, &from, &to, &is_kako_html))
    {
#if 0
      if (board == NULL)
	{
	  /* TODO: ɲåɤߤʤΤư*/
	  fprintf(stderr, "ochusha_open_url()\n");
	  fprintf(stderr, "  url: %s\n", url);
	  fprintf(stderr, "  board_url: %s\n", board_url);
	  fprintf(stderr, "  bbs_type: %d\n", bbs_type);
	  fprintf(stderr, "  thread_id: %s\n", thread_id);
	  fprintf(stderr, "  from=%d, to=%d\n", from, to);
	  fprintf(stderr, "  is_kako_html=%d\n", is_kako_html);
	}
#endif

      if (board != NULL)
	{
	  /* ɽ */
	  if (is_kako_html)
	    ochusha_open_kako_thread(application, board, url, in_tab);
	  else if (thread_id != NULL)
	    ochusha_open_thread(application, board, thread_id, from, in_tab);
	  else
	    paned_notebook_open_item_view(PANED_NOTEBOOK(application->contents_window),
					  board, in_tab,
					  application->select_tab_when_opened);

	  if (board_url != NULL)
	    G_FREE(board_url);
	  if (thread_id != NULL)
	    G_FREE(thread_id);
	  if (real_url != NULL)
	    G_FREE(real_url);
	  return;
	}
    }

  if (board_url != NULL)
    G_FREE(board_url);
  if (thread_id != NULL)
    G_FREE(thread_id);

  scheme = ochusha_utils_url_extract_scheme(url);

  if (scheme == NULL)
    {
      if (real_url != NULL)
	G_FREE(real_url);
      return;
    }

  /* TODO: scheme̤˵ư볰ץѹǽˤ٤ */
  G_FREE(scheme);

  if (snprintf(command_line, COMMAND_LINE_BUFFER_SIZE,
	       application->web_browser_template, url)
      < COMMAND_LINE_BUFFER_SIZE)
    {
      GError *error = NULL;
      if (is_shell_escape_required(command_line))
	{
	  char *escaped_command_line
	    = ochusha_utils_shell_escape_text(command_line);
	  if (!g_spawn_command_line_async(escaped_command_line, &error))
	    {
	      /* TODO: 顼ɽ */
	      fprintf(stderr, "spawning error[%s]: %s (%d)\n",
		      escaped_command_line, error->message, error->code);
	      g_error_free(error);
	    }
	  G_FREE(escaped_command_line);
	}
      else
	{
	  if (!g_spawn_command_line_async(command_line, &error))
	    {
	      /* TODO: 顼ɽ */
	      fprintf(stderr, "spawning error[%s]: %s (%d)\n",
		      command_line, error->message, error->code);
	      g_error_free(error);
	    }
	}
    }

  if (real_url != NULL)
    G_FREE(real_url);

  return;
}


void
setup_for_tab_label_animation(OchushaAsyncBuffer *buffer, IconLabel *tab_label)
{
  g_return_if_fail(OCHUSHA_IS_ASYNC_BUFFER(buffer)
		   && IS_ICON_LABEL(tab_label));

  g_signal_connect_object(G_OBJECT(buffer), "access_started",
			  G_CALLBACK(access_started_cb), tab_label, 0);
  g_signal_connect_object(G_OBJECT(buffer), "access_progressed",
			  G_CALLBACK(access_progressed_cb), tab_label, 0);
  g_signal_connect_object(G_OBJECT(buffer), "access_finished",
			  G_CALLBACK(access_finished_cb), tab_label, 0);
  g_signal_connect_object(G_OBJECT(buffer), "access_failed",
			  G_CALLBACK(access_failed_cb), tab_label, 0);

#if 1	/* ɽ֤Ϲˤ褦 */
  icon_label_set_icon(tab_label, tab_label_icons[TAB_LABEL_ICON_DONE]);
#else
  if (buffer->fixed)
    icon_label_set_icon(tab_label, tab_label_icons[TAB_LABEL_ICON_DONE]);
#endif
}


#if TRACE_MEMORY_USAGE
static void
trace_free(gpointer pointer)
{
  G_FREE(pointer);
}
#define TRACE_FREE	trace_free
#else
#define TRACE_FREE	g_free
#endif


GtkWidget *
get_tab_label(const gchar *label_text)
{
  GtkWidget *label = icon_label_new(label_text,
				tab_label_icons[TAB_LABEL_ICON_PREPARATION]);
  g_object_set_qdata_full(G_OBJECT(label), tab_label_icon_counter_id,
			  G_NEW0(int, 1), (GDestroyNotify)TRACE_FREE);
  return label;
}


static void
add_stock_icon(GtkIconFactory *factory, const char *stock_id,
	       GdkPixbuf *pixbuf, GtkIconSize size)
{
  GtkIconSource *source = gtk_icon_source_new();
  GtkIconSet *set = gtk_icon_set_new();

#if GTK_MINOR_VERSION <= 2
  gtk_icon_source_set_size(source, size);
  gtk_icon_source_set_pixbuf(source, pixbuf);

  gtk_icon_source_set_size_wildcarded(source, TRUE);
#else
  gtk_icon_source_set_icon_name(source, stock_id);
#endif
  gtk_icon_set_add_source(set, source);

  gtk_icon_source_free(source);

  gtk_icon_factory_add(factory, stock_id, set);

  gtk_icon_set_unref(set);

#if GTK_MINOR_VERSION > 2
  {
    /* ɥۥåн */
    gint pixel_size;
    switch (size)
      {
      case GTK_ICON_SIZE_MENU:
	pixel_size = 16;
	break;
      case GTK_ICON_SIZE_SMALL_TOOLBAR:
	pixel_size = 18;
	break;
      case GTK_ICON_SIZE_LARGE_TOOLBAR:
	pixel_size = 24;
	break;
      case GTK_ICON_SIZE_BUTTON:
	pixel_size = 20;
	break;
      case GTK_ICON_SIZE_DND:
	pixel_size = 32;
	break;
      case GTK_ICON_SIZE_DIALOG:
	pixel_size = 48;
	break;
      default:
	pixel_size = 0;		/* Τǡ */
      }
    gtk_icon_theme_add_builtin_icon(stock_id, pixel_size, pixbuf);
  }
#endif

  g_object_unref(pixbuf);	/* ݤʤΤǤunrefƤޤ */
}


#if GTK_MINOR_VERSION > 2
static void
add_stock_icon2(GtkIconFactory *factory, const char *stock_id,
		GdkPixbuf *pixbuf1, gint size1,
		GdkPixbuf *pixbuf2, gint size2)
{
  GtkIconSource *source = gtk_icon_source_new();
  GtkIconSet *set = gtk_icon_set_new();

  gtk_icon_source_set_icon_name(source, stock_id);
  gtk_icon_set_add_source(set, source);

  gtk_icon_source_free(source);

  gtk_icon_factory_add(factory, stock_id, set);

  gtk_icon_set_unref(set);

  gtk_icon_theme_add_builtin_icon(stock_id, size1, pixbuf1);
  gtk_icon_theme_add_builtin_icon(stock_id, size2, pixbuf2);

  g_object_unref(pixbuf1);	/* ݤʤΤǤunrefƤޤ */
  g_object_unref(pixbuf2);	/* ݤʤΤǤunrefƤޤ */
}
#endif


static const GtkStockItem ochusha_stock_items[] =
{
  {
    (char *)OCHUSHA_STOCK_MISSING_IMAGE,
    (char *)N_("Missing Image"),
    0, 0,
    (char *)PACKAGE_NAME
  },
  {
    (char *)OCHUSHA_STOCK_WRITE_RESPONSE,
    (char *)N_("_Write"),
    0, 0,
    (char *)PACKAGE_NAME
  }
};


static void
setup_ochusha_stock_items(void)
{
  GtkIconFactory *factory = gtk_icon_factory_new();

  add_stock_icon(factory, OCHUSHA_STOCK_MISSING_IMAGE,
		 gdk_pixbuf_new_from_inline(-1, missing_image_24,
					    FALSE, NULL),
		 GTK_ICON_SIZE_LARGE_TOOLBAR);

#if GTK_MINOR_VERSION <= 2
  add_stock_icon(factory, OCHUSHA_STOCK_WRITE_RESPONSE,
		 gdk_pixbuf_new_from_inline(-1, write_response_24,
					    FALSE, NULL),
		 GTK_ICON_SIZE_LARGE_TOOLBAR);

  add_stock_icon(factory, OCHUSHA_STOCK_WRITE_RESPONSE,
		 gdk_pixbuf_new_from_inline(-1, write_response_16,
					    FALSE, NULL),
		 GTK_ICON_SIZE_MENU);
#else
  add_stock_icon2(factory, OCHUSHA_STOCK_WRITE_RESPONSE,
		  gdk_pixbuf_new_from_inline(-1, write_response_24,
					     FALSE, NULL),
		  24,
		  gdk_pixbuf_new_from_inline(-1, write_response_16,
					     FALSE, NULL),
		  16);
#endif

  add_stock_icon(factory, OCHUSHA_STOCK_NET_PROGRESS0,
		 gdk_pixbuf_new_from_xpm_data((const char **)blue0_xpm),
		 GTK_ICON_SIZE_MENU);

  add_stock_icon(factory, OCHUSHA_STOCK_NET_PROGRESS1,
		 gdk_pixbuf_new_from_xpm_data((const char **)blue1_xpm),
		 GTK_ICON_SIZE_MENU);

  add_stock_icon(factory, OCHUSHA_STOCK_NET_PROGRESS2,
		 gdk_pixbuf_new_from_xpm_data((const char **)blue2_xpm),
		 GTK_ICON_SIZE_MENU);

  add_stock_icon(factory, OCHUSHA_STOCK_NET_PROGRESS3,
		 gdk_pixbuf_new_from_xpm_data((const char **)blue3_xpm),
		 GTK_ICON_SIZE_MENU);

  add_stock_icon(factory, OCHUSHA_STOCK_NET_PREPARATION,
		 gdk_pixbuf_new_from_xpm_data((const char **)orange_xpm),
		 GTK_ICON_SIZE_MENU);

  add_stock_icon(factory, OCHUSHA_STOCK_NET_DONE,
		 gdk_pixbuf_new_from_xpm_data((const char **)black_xpm),
		 GTK_ICON_SIZE_MENU);

  add_stock_icon(factory, OCHUSHA_STOCK_NET_ERROR,
		 gdk_pixbuf_new_from_xpm_data((const char **)red_xpm),
		 GTK_ICON_SIZE_MENU);

  gtk_icon_factory_add_default(factory);

  g_object_unref(factory);

  gtk_stock_add_static(ochusha_stock_items, G_N_ELEMENTS(ochusha_stock_items));
}


static void
setup_tab_label_icons(void)
{
  GtkWidget *dummy = icon_label_new("", NULL);

  tab_label_icons[TAB_LABEL_ICON_PROGRESS0]
    = gtk_widget_render_icon(dummy, OCHUSHA_STOCK_NET_PROGRESS0,
			     GTK_ICON_SIZE_MENU, "ochusha::net-progress-0");

  tab_label_icons[TAB_LABEL_ICON_PROGRESS1]
    = gtk_widget_render_icon(dummy, OCHUSHA_STOCK_NET_PROGRESS1,
			     GTK_ICON_SIZE_MENU, "ochusha::net-progress-1");

  tab_label_icons[TAB_LABEL_ICON_PROGRESS2]
    = gtk_widget_render_icon(dummy, OCHUSHA_STOCK_NET_PROGRESS2,
			     GTK_ICON_SIZE_MENU, "ochusha::net-progress-2");

  tab_label_icons[TAB_LABEL_ICON_PROGRESS3]
    = gtk_widget_render_icon(dummy, OCHUSHA_STOCK_NET_PROGRESS3,
			     GTK_ICON_SIZE_MENU, "ochusha::net-progress-3");

  tab_label_icons[TAB_LABEL_ICON_PREPARATION]
    = gtk_widget_render_icon(dummy, OCHUSHA_STOCK_NET_PREPARATION,
			     GTK_ICON_SIZE_MENU, "ochusha::net-preparation");

  tab_label_icons[TAB_LABEL_ICON_DONE]
    = gtk_widget_render_icon(dummy, OCHUSHA_STOCK_NET_DONE,
			     GTK_ICON_SIZE_MENU, "ochusha::net-done");

  tab_label_icons[TAB_LABEL_ICON_ERROR]
    = gtk_widget_render_icon(dummy, OCHUSHA_STOCK_NET_ERROR,
			     GTK_ICON_SIZE_MENU, "ochusha::net-error");

  gtk_object_sink(GTK_OBJECT(dummy));

  tab_label_id = g_quark_from_static_string("OchushaUI::icon-label");
  tab_label_icon_counter_id
    = g_quark_from_static_string("OchushaUI::icon-counter");
  application_id = g_quark_from_static_string("OchushaUI::Application");
}


static void
delete_event_cb(GtkWidget *widget, GdkEventAny *event, gpointer user_data)
{
  ochusha_quit((OchushaApplication *)user_data);
}


static void
ochusha_quit(OchushaApplication *application)
{
#if ENABLE_DEBUG_CONSOLE
  enable_debug_outputs = FALSE;
#endif

  gdk_threads_leave();
  ochusha_async_buffer_suspend_all();
  cancel_pending_jobs();
  ochusha_async_buffer_terminate_all();
  gdk_threads_enter();

  if (application->sm_client_id != NULL)
    {
      ochusha_session_discard_states(application, application->sm_client_id);
      application->session_subdir = NULL;
    }

  ochusha_session_save_states(application);

  finalize_xsmp_handlers(application);

  gtk_main_quit();
}


void
ochusha_write_preferences(OchushaApplication *application)
{
  if (!ochusha_write_config_xml(&application->config,
				application->session_subdir,
				write_optional_prefs, application))
    fprintf(stderr, "Couldn't write %s/%s.\n",
	    application->session_subdir, OCHUSHA_CONFIG_XML);

  if (application->session_subdir != NULL
      && !ochusha_write_config_xml(&application->config, NULL,
				   write_optional_prefs, application))
    fprintf(stderr, "Couldn't write %s.\n", OCHUSHA_CONFIG_XML);
}


#if GTK_MINOR_VERSION <= 2
static void
main_menu_callback(gpointer data, guint action, GtkWidget *widget)
{
  OchushaApplication *application = (OchushaApplication *)data;

  switch (action)
    {
    case MAIN_MENU_OPEN_URL:
      popup_open_url_dialog(application);
      return;

    case MAIN_MENU_REFRESH_BOARDLIST:
      refresh_boardlist(application);
      return;

    case MAIN_MENU_CREATE_NEW_CATEGORY:
      ochusha_create_new_category(application);
      return;

    case MAIN_MENU_ADD_BOARD:
      ochusha_add_board(application, NULL);
      return;

    case MAIN_MENU_CREATE_NEW_VIRTUAL_BOARD:
      ochusha_create_new_virtual_board(application);
      return;

    case MAIN_MENU_WORK_OFFLINE:
    case MAIN_MENU_HIDE_HIDDEN_BOARDLIST_ITEMS:
    case MAIN_MENU_HIDE_HIDDEN_THREADS:
    case MAIN_MENU_USE_BUILTIN_CALCULATE_WEIGHT:
    case MAIN_MENU_SORT_THREADLIST_BY_WEIGHT:
      {
	GtkCheckMenuItem *item
	  = GTK_CHECK_MENU_ITEM(gtk_item_factory_get_widget_by_action(
							main_menu_item_factory,
							action));
	if (action == MAIN_MENU_WORK_OFFLINE)
	  {
	    application->config.offline
	      = gtk_check_menu_item_get_active(item);
	  }
	else if (action == MAIN_MENU_HIDE_HIDDEN_BOARDLIST_ITEMS)
	  {
	    application->hide_hidden_boardlist_items
	      = gtk_check_menu_item_get_active(item);
	    if (application->boardlist_view != NULL)
	      redraw_boardlist(application);
	  }
	else if (action == MAIN_MENU_HIDE_HIDDEN_THREADS)
	  application->hide_hidden_threads
	    = gtk_check_menu_item_get_active(item);
	else if (action == MAIN_MENU_USE_BUILTIN_CALCULATE_WEIGHT)
	  application->use_builtin_calculate_weight
	    = gtk_check_menu_item_get_active(item);
	else
	  application->sort_threadlist_by_weight
	    = gtk_check_menu_item_get_active(item);
	return;
      }

    case MAIN_MENU_REFRESH_THREAD:
      refresh_current_thread(application);
      return;

#if 0
    case MAIN_MENU_REFRESH_THREAD_ALL:
      refresh_current_thread_all(application);
      return;
#endif

    case MAIN_MENU_REFRESH_THREADLIST:
      refresh_current_threadlist(application);
      return;

    case MAIN_MENU_REFRESH_THREADLIST_ALL:
      refresh_current_threadlist_all(application);
      return;

#if ENABLE_DEBUG_CONSOLE
    case MAIN_MENU_DEBUG_CONSOLE:
      open_debug_console(application);
      return;
#endif

    case MAIN_MENU_OPTIONS:
      open_preference_dialog(application);
      return;

    case MAIN_MENU_SELECT_BOARDLIST_THREADLIST_FONTS:
      select_tree_view_font(application);
      return;

    case MAIN_MENU_SELECT_THREAD_FONTS:
      select_thread_view_font(application);
      return;

    case MAIN_MENU_ABOUT:
      open_about_dialog(application);
      return;
    }

  fprintf(stderr, "Not implemented yet.\n");
}
#endif


#define OUTPUT_CONFIG_ATTRIBUTE_INT(gzfile, config, attribute)	\
  gzprintf(gzfile,						\
	  "    <attribute name=\"" #attribute	"\">\n"		\
	  "      <int val=\"%d\"/>\n"				\
	  "    </attribute>\n", (config)->attribute)

#define OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(gzfile, config, attribute)	\
  gzprintf(gzfile,							\
	  "    <attribute name=\"" #attribute	"\">\n"			\
	  "      <boolean val=\"%s\"/>\n"				\
	  "    </attribute>\n", (config)->attribute ? "true" : "false" )

#define OUTPUT_CONFIG_ATTRIBUTE_STRING(gzfile, config, attribute)	\
  do									\
    {									\
      if ((config)->attribute != NULL)					\
	{								\
	  gchar *text = g_markup_escape_text((config)->attribute, -1);	\
	  gzprintf(gzfile,						\
		   "    <attribute name=\"" #attribute	"\">\n"		\
		   "      <string>%s</string>\n"			\
		   "    </attribute>\n", text);				\
	  g_free(text);							\
	}								\
    } while (0)


static void
write_optional_prefs(gzFile file, gpointer user_data)
{
  OchushaApplication *application = (OchushaApplication *)user_data;
  gchar *bookmark_text;

  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, config_version);

  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, width);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, height);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, boardlist_width);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, threadlist_height);

  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, recently_added_favorites);

  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, boardlist_pane_style);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, board_tab_shrinkable);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, board_tab_minimum_size);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  board_tab_enable_tooltips);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  board_tab_always_show);
  
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, threadlist_pane_style);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, thread_tab_shrinkable);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, thread_tab_minimum_size);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  thread_tab_enable_tooltips);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  thread_tab_always_show);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  show_threadlist_entry_when_thread_selected);
  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, threadlist_view_contents);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  use_builtin_calculate_weight);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  sort_threadlist_by_weight);

  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, thread_view_font_name);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  enable_transparent_a_bone);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, disable_popup_a_bone);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, hide_bookmark);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, stick_bookmark);

  if (application->bookmark_text != NULL
      && *application->bookmark_text != '\0')
    {
      bookmark_text
	= ochusha_utils_url_encode_string(application->bookmark_text);

      if (bookmark_text != NULL && *bookmark_text != '\0')
	{
	  gchar *text = g_markup_escape_text(bookmark_text, -1);
	  gzprintf(file,
		  "    <attribute name=\"bookmark_text\">\n"
		  "      <string>%s</string>\n"
		  "    </attribute>\n", text);
	  g_free(text);
	}
      if (bookmark_text != NULL)
	G_FREE(bookmark_text);
    }

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  show_threadlist_non_incrementally);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, default_open_in_tab);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, select_tab_when_opened);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, show_mailto_mode);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, enable_inline_image);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, inline_image_width);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, inline_image_height);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, enable_popup_image);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, popup_image_width);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, popup_image_height);

  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, capacity_of_image_cache);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application,
			      maximum_number_of_cache_files);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, a_bone_by_name);
  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, a_bone_by_name_pattern);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, a_bone_by_id);
  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, a_bone_by_id_pattern);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, a_bone_by_content);
  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application,
				 a_bone_by_content_pattern);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, enable_popup_title);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, popup_title_delay);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, enable_popup_response);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, popup_response_delay);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, popup_close_delay);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application,
			      maximum_number_of_popup_responses);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  analyze_informal_links);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, enable_popup_id);

  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, web_browser_template);

  if (application->user_default_filter.rule != NULL)
    OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application,
				   user_default_filter.rule);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application,
			      user_default_filter.ignore_threshold);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  hide_hidden_boardlist_items);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  expand_only_one_category);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, hide_hidden_threads);

  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, download_chunksize);
  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, image_chunksize);

  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, last_name);
  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, last_mail);

  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, last_download_directory);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, shutdown_succeeded);

  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, threadlist_search_direction);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  threadlist_search_enable_wrap);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  threadlist_search_match_case);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  threadlist_search_use_regexp);

  OUTPUT_CONFIG_ATTRIBUTE_INT(file, application, thread_search_direction);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  thread_search_enable_wrap);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  thread_search_match_case);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application,
				  thread_search_use_regexp);

  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, dolib_id);
  if (application->dolib_save_passwd)
    OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, dolib_passwd);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, dolib_save_passwd);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, dolib_login_at_startup);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, dolib_use_id_for_posting);

  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, be_id);
  if (application->be_save_passwd)
    OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, be_passwd);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, be_save_passwd);
  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, be_login_at_startup);

  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, aalist_filename);
  OUTPUT_CONFIG_ATTRIBUTE_STRING(file, application, aalist_char_encoding);

  OUTPUT_CONFIG_ATTRIBUTE_BOOLEAN(file, application, focus_follow_mouse);
}


static void
configure_application(OchushaApplication *application)
{
  gboolean succeeded;
  int version;

  ochusha_read_config_xml(&application->config, application->session_subdir,
			  read_optional_prefs, application);
  application->broker = ochusha_network_broker_new(&application->config);

  g_signal_connect(G_OBJECT(application->broker), "output_log",
		   G_CALLBACK(output_log_cb), application);
  if (application->dolib_login_at_startup
      && application->dolib_id != NULL
      && *application->dolib_id != '\0'
      && application->dolib_passwd != NULL
      && *application->dolib_passwd != '\0')
    {
      ochusha_utils_2ch_login(&application->config, application->broker,
			      application->dolib_id,
			      application->dolib_passwd);
    }

  if (application->be_login_at_startup
      && application->be_id != NULL && *application->be_id != '\0'
      && application->be_passwd != NULL && *application->be_passwd != '\0')
    {
      ochusha_utils_2ch_be_login(&application->config, application->broker,
				 application->be_id,
				 application->be_passwd);
    }

  /* Sets unspecified attributes */
  if (application->width == 0)
    {	/* ƵưȻפ */
      /* ɥν */
      application->x_pos = -1;
      application->y_pos = -1;
      application->width = OCHUSHA_DEFAULT_WIDTH;
      application->height = OCHUSHA_DEFAULT_HEIGHT;
      application->boardlist_width = OCHUSHA_DEFAULT_BOARDLIST_WIDTH;
      application->threadlist_height = OCHUSHA_DEFAULT_THREADLIST_HEIGHT;

      /* İϢ */
      application->boardlist_pane_style = HPANED_STYLE;
      application->board_tab_shrinkable = OCHUSHA_DEFAULT_BOARD_TAB_SHRINKABLE;
      application->board_tab_minimum_size
	= OCHUSHA_DEFAULT_BOARD_TAB_MINIMUM_SIZE;
      application->board_tab_enable_tooltips = FALSE;


      /* Ϣ */
      application->threadlist_pane_style = VPANED_STYLE;
      application->thread_tab_shrinkable
	= OCHUSHA_DEFAULT_THREAD_TAB_SHRINKABLE;
      application->thread_tab_minimum_size
	= OCHUSHA_DEFAULT_THREAD_TAB_MINIMUM_SIZE;
      application->thread_tab_enable_tooltips = TRUE;
      application->show_threadlist_entry_when_thread_selected = FALSE;


      /* ɽϢ */
      application->thread_view_font_name = NULL;
      application->enable_transparent_a_bone = FALSE;
      application->disable_popup_a_bone = FALSE;

      application->hide_bookmark = FALSE;
      application->stick_bookmark = FALSE;
      application->bookmark_text = G_STRDUP(_("==== My bookmark: I have read above. ===="));

      application->show_threadlist_non_incrementally = FALSE;

      application->default_open_in_tab = FALSE;
      application->select_tab_when_opened = FALSE;
      application->show_mailto_mode = OCHUSHA_DEFAULT_SHOW_MAILTO_MODE;
      application->a_bone_by_name = FALSE;
      application->a_bone_by_name_pattern = G_STRDUP("");
      application->a_bone_by_id = FALSE;
      application->a_bone_by_id_pattern = G_STRDUP("");
      application->a_bone_by_content = FALSE;
      application->a_bone_by_content_pattern = G_STRDUP("");


      /* 쥿Υݥåץå״Ϣ */
      application->enable_popup_title = OCHUSHA_DEFAULT_ENABLE_POPUP_TITLE;
      application->popup_title_delay = OCHUSHA_DEFAULT_POPUP_DELAY;


      /* 쥹Υݥåץå״Ϣ */
      application->enable_popup_response
	= OCHUSHA_DEFAULT_ENABLE_POPUP_RESPONSE;
      application->popup_response_delay = OCHUSHA_DEFAULT_POPUP_DELAY;
      application->popup_close_delay = OCHUSHA_DEFAULT_POPUP_CLOSE_DELAY;
      application->maximum_number_of_popup_responses
	= OCHUSHA_DEFAULT_MAXIMUM_NUMBER_OF_POPUP_RESPONSES;
      application->analyze_informal_links
	= OCHUSHA_DEFAULT_ANALYZE_INFORMAL_LINKS;
      application->enable_popup_id
	= OCHUSHA_DEFAULT_ENABLE_POPUP_ID;

      application->show_threadlist_entry_when_thread_selected
	= OCHUSHA_DEFAULT_SHOW_THREADLIST_ENTRY_WHEN_THREAD_SELECTED;
      application->hide_hidden_boardlist_items
	= OCHUSHA_DEFAULT_HIDE_HIDDEN_BOARDLIST_ITEMS;
      application->expand_only_one_category = FALSE;
      application->hide_hidden_threads
	= OCHUSHA_DEFAULT_HIDE_HIDDEN_THREADS;
      application->shutdown_succeeded = FALSE;

      application->aalist_char_encoding
	= G_STRDUP(OCHUSHA_DEFAULT_AALIST_CHAR_ENCODING);

      application->focus_follow_mouse = OCHUSHA_DEFAULT_FOCUS_FOLLOW_MOUSE;
    }

  if (application->threadlist_view_contents == NULL)
    application->threadlist_view_contents
      = G_STRDUP(OCHUSHA_DEFAULT_THREADLIST_VIEW_CONTENTS);

  if (application->board_tab_minimum_size == 0)
    application->board_tab_minimum_size
      = OCHUSHA_DEFAULT_BOARD_TAB_MINIMUM_SIZE;

  if (application->thread_tab_minimum_size == 0)
    application->thread_tab_minimum_size
      = OCHUSHA_DEFAULT_THREAD_TAB_MINIMUM_SIZE;

  if (application->popup_title_delay == 0)
    application->popup_title_delay = OCHUSHA_DEFAULT_POPUP_DELAY;

  if (application->web_browser_template == NULL)
    application->web_browser_template
      = validate_web_browser_template(OCHUSHA_DEFAULT_WEB_BROWSER_TEMPLATE);

  if (application->inline_image_width <= 0
      || application->inline_image_height <= 0)
    {
      application->inline_image_width = OCHUSHA_DEFAULT_INLINE_IMAGE_WIDTH;
      application->inline_image_height = OCHUSHA_DEFAULT_INLINE_IMAGE_HEIGHT;
    }

  if (application->popup_image_width <= 0
      || application->popup_image_height <= 0)
    {
      application->popup_image_width = OCHUSHA_DEFAULT_POPUP_IMAGE_WIDTH;
      application->popup_image_height = OCHUSHA_DEFAULT_POPUP_IMAGE_HEIGHT;
    }

  if (application->capacity_of_image_cache > 0
      && application->capacity_of_image_cache < OCHUSHA_DEFAULT_CAPACITY_OF_IMAGE_CACHE)
    application->capacity_of_image_cache
      = OCHUSHA_DEFAULT_CAPACITY_OF_IMAGE_CACHE;

  if (application->maximum_number_of_cache_files > 0
      && application->maximum_number_of_cache_files < OCHUSHA_DEFAULT_MAXIMUM_NUMBER_OF_CACHE_FILES)
    application->maximum_number_of_cache_files
      = OCHUSHA_DEFAULT_MAXIMUM_NUMBER_OF_CACHE_FILES;

  if (application->download_chunksize == 0)
    application->download_chunksize = 0x20000;

  if (application->image_chunksize == 0)
    application->image_chunksize = 0x20000;

  succeeded = application->shutdown_succeeded;
  version = application->config_version;
  application->config_version = OCHUSHA_CURRENT_CONFIG_VERSION;

  if (application->aalist_filename == NULL)
    application->aalist_filename = G_STRDUP(OCHUSHA_DEFAULT_AALIST_FILENAME);

  if (application->aalist_char_encoding == NULL)
    application->aalist_char_encoding
      = G_STRDUP(OCHUSHA_DEFAULT_AALIST_CHAR_ENCODING);

  application->shutdown_succeeded = FALSE;
  ochusha_write_preferences(application);
  application->shutdown_succeeded = succeeded;

  if (!succeeded || version != OCHUSHA_CURRENT_CONFIG_VERSION)
    application->recover_mode = TRUE;
  else
    application->recover_mode = FALSE;

  application->ice_connection_poll_id = 0;

  ochusha_session_restore_states(application);
}


static void
read_optional_prefs(GHashTable *pref_attrs, gpointer user_data)
{
  OchushaApplication *application = (OchushaApplication *)user_data;
  gchar *tmp_string;

  application->config_version
    = ochusha_utils_get_attribute_int(pref_attrs, "config_version");
  application->width = ochusha_utils_get_attribute_int(pref_attrs, "width");

  application->height = ochusha_utils_get_attribute_int(pref_attrs, "height");

  application->boardlist_width
    = ochusha_utils_get_attribute_int(pref_attrs, "boardlist_width");

  application->threadlist_height
    = ochusha_utils_get_attribute_int(pref_attrs, "threadlist_height");

  if (application->recently_added_favorites != NULL)
    {
      char *tmp_id = application->recently_added_favorites;
      application->recently_added_favorites
	= ochusha_utils_get_attribute_string(pref_attrs,
					     "recently_added_favorites");
      if (application->recently_added_favorites == NULL)
	application->recently_added_favorites = tmp_id;
      else
	G_FREE(tmp_id);
    }
  else
    application->recently_added_favorites
      = ochusha_utils_get_attribute_string(pref_attrs,
					   "recently_added_favorites");

  application->download_chunksize
    = ochusha_utils_get_attribute_int(pref_attrs, "download_chunksize");
  application->image_chunksize
    = ochusha_utils_get_attribute_int(pref_attrs, "image_chunksize");

  if (g_hash_table_lookup(pref_attrs, "boardlist_pane_style") != NULL)
    application->boardlist_pane_style
      = ochusha_utils_get_attribute_int(pref_attrs, "boardlist_pane_style");
  else
    application->boardlist_pane_style = HPANED_STYLE;

  if (g_hash_table_lookup(pref_attrs, "board_tab_shrinkable") != NULL)
    application->board_tab_shrinkable
      = ochusha_utils_get_attribute_boolean(pref_attrs,
					    "board_tab_shrinkable");
  else
    application->board_tab_shrinkable = OCHUSHA_DEFAULT_BOARD_TAB_SHRINKABLE;

  if (g_hash_table_lookup(pref_attrs, "board_tab_minimum_size") != NULL)
    application->board_tab_minimum_size
      = ochusha_utils_get_attribute_int(pref_attrs, "board_tab_minimum_size");
  else
    application->board_tab_minimum_size
      = OCHUSHA_DEFAULT_BOARD_TAB_MINIMUM_SIZE;

  if (g_hash_table_lookup(pref_attrs, "board_tab_enable_tooltips") != NULL)
    application->board_tab_enable_tooltips
      = ochusha_utils_get_attribute_boolean(pref_attrs,
					    "board_tab_enable_tooltips");
  else
    application->board_tab_enable_tooltips
      = OCHUSHA_DEFAULT_BOARD_TAB_ENABLE_TOOLTIPS;

  application->board_tab_always_show
    = ochusha_utils_get_attribute_boolean(pref_attrs, "board_tab_always_show");

  if (g_hash_table_lookup(pref_attrs, "threadlist_pane_style") != NULL)
    application->threadlist_pane_style
      = ochusha_utils_get_attribute_int(pref_attrs, "threadlist_pane_style");
  else
    application->threadlist_pane_style = VPANED_STYLE;

  if (g_hash_table_lookup(pref_attrs, "thread_tab_shrinkable") != NULL)
    application->thread_tab_shrinkable
      = ochusha_utils_get_attribute_boolean(pref_attrs,
					    "thread_tab_shrinkable");
  else
    application->thread_tab_shrinkable = OCHUSHA_DEFAULT_THREAD_TAB_SHRINKABLE;

  if (g_hash_table_lookup(pref_attrs, "thread_tab_minimum_size") != NULL)
    application->thread_tab_minimum_size
      = ochusha_utils_get_attribute_int(pref_attrs, "thread_tab_minimum_size");
  else
    application->thread_tab_minimum_size
      = OCHUSHA_DEFAULT_THREAD_TAB_MINIMUM_SIZE;

  if (g_hash_table_lookup(pref_attrs, "thread_tab_enable_tooltips") != NULL)
    application->thread_tab_enable_tooltips
      = ochusha_utils_get_attribute_boolean(pref_attrs,
					    "thread_tab_enable_tooltips");
  else
    application->thread_tab_enable_tooltips
      = OCHUSHA_DEFAULT_THREAD_TAB_ENABLE_TOOLTIPS;

  application->thread_tab_always_show
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "thread_tab_always_show");

  if (g_hash_table_lookup(pref_attrs,
			  "show_threadlist_entry_when_thread_selected")
      != NULL)
    application->show_threadlist_entry_when_thread_selected
      = ochusha_utils_get_attribute_boolean(pref_attrs,
				"show_threadlist_entry_when_thread_selected");
  else
    application->show_threadlist_entry_when_thread_selected
      = OCHUSHA_DEFAULT_SHOW_THREADLIST_ENTRY_WHEN_THREAD_SELECTED;

  if (g_hash_table_lookup(pref_attrs, "threadlist_view_contents") != NULL)
    application->threadlist_view_contents
      = ochusha_utils_get_attribute_string(pref_attrs,
					   "threadlist_view_contents");
  else
    application->threadlist_view_contents
      = G_STRDUP(OCHUSHA_DEFAULT_THREADLIST_VIEW_CONTENTS);

  application->use_builtin_calculate_weight
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "use_builtin_calculate_weight");
  application->sort_threadlist_by_weight
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "sort_threadlist_by_weight");

  application->show_threadlist_non_incrementally
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "show_threadlist_non_incrementally");

  application->default_open_in_tab
    = ochusha_utils_get_attribute_boolean(pref_attrs, "default_open_in_tab");

  application->select_tab_when_opened
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "select_tab_when_opened");
  application->thread_view_font_name
    = ochusha_utils_get_attribute_string(pref_attrs, "thread_view_font_name");

  application->enable_transparent_a_bone
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "enable_transparent_a_bone");
  application->disable_popup_a_bone
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "disable_popup_a_bone");

  application->hide_bookmark
    = ochusha_utils_get_attribute_boolean(pref_attrs, "hide_bookmark");
  application->stick_bookmark
    = ochusha_utils_get_attribute_boolean(pref_attrs, "stick_bookmark");
  application->bookmark_text = NULL;
  tmp_string = ochusha_utils_get_attribute_string(pref_attrs, "bookmark_text");
  if (tmp_string != NULL)
    {
      application->bookmark_text
	= ochusha_utils_url_decode_string(tmp_string);
      G_FREE(tmp_string);
    }

  if (application->bookmark_text != NULL
      && *application->bookmark_text == '\0')
    {
      G_FREE(application->bookmark_text);
      application->bookmark_text = NULL;
    }

  if (application->bookmark_text == NULL)
    application->bookmark_text
      = G_STRDUP(_("==== My bookmark: I have read above. ===="));

  application->show_mailto_mode
    = (ThreadViewMailtoMode)ochusha_utils_get_attribute_int(pref_attrs,
							   "show_mailto_mode");
  if (application->show_mailto_mode < THREAD_VIEW_MAILTO_MODE_DEFAULT
      || application->show_mailto_mode > THREAD_VIEW_MAILTO_MODE_HIDE)
    application->show_mailto_mode = THREAD_VIEW_MAILTO_MODE_DEFAULT;

  application->enable_inline_image
    = ochusha_utils_get_attribute_boolean(pref_attrs, "enable_inline_image");
  application->inline_image_width
    = ochusha_utils_get_attribute_int(pref_attrs, "inline_image_width");
  application->inline_image_height
    = ochusha_utils_get_attribute_int(pref_attrs, "inline_image_height");

  application->enable_popup_image
    = ochusha_utils_get_attribute_boolean(pref_attrs, "enable_popup_image");
  application->popup_image_width
    = ochusha_utils_get_attribute_int(pref_attrs, "popup_image_width");
  application->popup_image_height
    = ochusha_utils_get_attribute_int(pref_attrs, "popup_image_height");

  application->capacity_of_image_cache
    = ochusha_utils_get_attribute_int(pref_attrs, "capacity_of_image_cache");
  application->maximum_number_of_cache_files
    = ochusha_utils_get_attribute_int(pref_attrs,
				      "maximum_number_of_cache_files");

  application->a_bone_by_name
    = ochusha_utils_get_attribute_boolean(pref_attrs, "a_bone_by_name");
  application->a_bone_by_name_pattern
    = ochusha_utils_get_attribute_string(pref_attrs, "a_bone_by_name_pattern");
  if (application->a_bone_by_name_pattern == NULL)
    application->a_bone_by_name_pattern = G_STRDUP("");

  application->a_bone_by_id
    = ochusha_utils_get_attribute_boolean(pref_attrs, "a_bone_by_id");
  application->a_bone_by_id_pattern
    = ochusha_utils_get_attribute_string(pref_attrs, "a_bone_by_id_pattern");
  if (application->a_bone_by_id_pattern == NULL)
    application->a_bone_by_id_pattern = G_STRDUP("");

  application->a_bone_by_content
    = ochusha_utils_get_attribute_boolean(pref_attrs, "a_bone_by_content");
  application->a_bone_by_content_pattern
    = ochusha_utils_get_attribute_string(pref_attrs,
					 "a_bone_by_content_pattern");
  if (application->a_bone_by_content_pattern == NULL)
    application->a_bone_by_content_pattern = G_STRDUP("");

  if (g_hash_table_lookup(pref_attrs, "enable_popup_title") != NULL)
    application->enable_popup_title
      = ochusha_utils_get_attribute_boolean(pref_attrs, "enable_popup_title");
  else
    application->enable_popup_title = FALSE;

  application->popup_title_delay
    = ochusha_utils_get_attribute_int(pref_attrs, "popup_title_delay");

  if (g_hash_table_lookup(pref_attrs, "enable_popup_response") != NULL)
    application->enable_popup_response
      = ochusha_utils_get_attribute_boolean(pref_attrs,
					    "enable_popup_response");
  else
    application->enable_popup_response = TRUE;

  if (g_hash_table_lookup(pref_attrs, "popup_response_delay") != NULL)
    application->popup_response_delay
      = ochusha_utils_get_attribute_int(pref_attrs, "popup_response_delay");
  else
    application->popup_response_delay = OCHUSHA_DEFAULT_POPUP_DELAY;

  if (g_hash_table_lookup(pref_attrs, "popup_close_delay") != NULL)
    application->popup_close_delay
      = ochusha_utils_get_attribute_int(pref_attrs, "popup_close_delay");
  else
    application->popup_close_delay = OCHUSHA_DEFAULT_POPUP_CLOSE_DELAY;

  if (g_hash_table_lookup(pref_attrs, "maximum_number_of_popup_responses")
      != NULL)
    application->maximum_number_of_popup_responses
      = ochusha_utils_get_attribute_int(pref_attrs,
					"maximum_number_of_popup_responses");
  else
    application->maximum_number_of_popup_responses
      = OCHUSHA_DEFAULT_MAXIMUM_NUMBER_OF_POPUP_RESPONSES;

  if (g_hash_table_lookup(pref_attrs, "analyze_informal_links") != NULL)
    application->analyze_informal_links
      = ochusha_utils_get_attribute_boolean(pref_attrs,
					    "analyze_informal_links");
  else
    application->analyze_informal_links
      = OCHUSHA_DEFAULT_ANALYZE_INFORMAL_LINKS;

  if (g_hash_table_lookup(pref_attrs, "enable_popup_id") != NULL)
    application->enable_popup_id
      = ochusha_utils_get_attribute_boolean(pref_attrs,
					    "enable_popup_id");
  else
    application->enable_popup_id = OCHUSHA_DEFAULT_ENABLE_POPUP_ID;

  tmp_string = g_hash_table_lookup(pref_attrs, "web_browser_template");
  if (tmp_string == NULL)
    tmp_string = OCHUSHA_DEFAULT_WEB_BROWSER_TEMPLATE;
  application->web_browser_template
    = validate_web_browser_template(tmp_string);

  application->user_default_filter.rule
    = ochusha_utils_get_attribute_string(pref_attrs,
					 "user_default_filter.rule");

  application->user_default_filter.ignore_threshold
    = ochusha_utils_get_attribute_int(pref_attrs,
				      "user_default_filter.ignore_threshold");

  application->hide_hidden_boardlist_items
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "hide_hidden_boardlist_items");
  application->expand_only_one_category
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "expand_only_one_category");
  application->hide_hidden_threads
    = ochusha_utils_get_attribute_boolean(pref_attrs, "hide_hidden_threads");

  application->last_name
    = ochusha_utils_get_attribute_string(pref_attrs, "last_name");
  application->last_mail
    = ochusha_utils_get_attribute_string(pref_attrs, "last_mail");
  application->last_download_directory
    = ochusha_utils_get_attribute_string(pref_attrs,
					 "last_download_directory");
  application->shutdown_succeeded
    = ochusha_utils_get_attribute_boolean(pref_attrs, "shutdown_succeeded");

  application->threadlist_search_direction
    = ochusha_utils_get_attribute_int(pref_attrs,
				      "threadlist_search_direction");
  application->threadlist_search_enable_wrap
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "threadlist_search_enable_wrap");
  application->threadlist_search_match_case
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "threadlist_search_match_case");
  application->threadlist_search_use_regexp
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "threadlist_search_use_regexp");

  application->thread_search_direction
    = ochusha_utils_get_attribute_int(pref_attrs,
				      "thread_search_direction");
  application->thread_search_enable_wrap
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "thread_search_enable_wrap");
  application->thread_search_match_case
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "thread_search_match_case");
  application->thread_search_use_regexp
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "thread_search_use_regexp");

  application->dolib_id
    = ochusha_utils_get_attribute_string(pref_attrs, "dolib_id");
  application->dolib_passwd
    = ochusha_utils_get_attribute_string(pref_attrs, "dolib_passwd");

  application->dolib_save_passwd
    = ochusha_utils_get_attribute_boolean(pref_attrs, "dolib_save_passwd");
  application->dolib_login_at_startup
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "dolib_login_at_startup");
  application->dolib_use_id_for_posting
    = ochusha_utils_get_attribute_boolean(pref_attrs,
					  "dolib_use_id_for_posting");

  application->be_id
    = ochusha_utils_get_attribute_string(pref_attrs, "be_id");
  application->be_passwd
    = ochusha_utils_get_attribute_string(pref_attrs, "be_passwd");

  application->be_save_passwd
    = ochusha_utils_get_attribute_boolean(pref_attrs, "be_save_passwd");
  application->be_login_at_startup
    = ochusha_utils_get_attribute_boolean(pref_attrs, "be_login_at_startup");

  application->aalist_filename
    = ochusha_utils_get_attribute_string(pref_attrs, "aalist_filename");
  application->aalist_char_encoding
    = ochusha_utils_get_attribute_string(pref_attrs, "aalist_char_encoding");

  application->focus_follow_mouse
    = ochusha_utils_get_attribute_boolean(pref_attrs, "focus_follow_mouse");
}


static void
access_started_cb(OchushaAsyncBuffer *buffer, IconLabel *tab_label)
{
  int *counter;

#if DEBUG_GUI
  fprintf(stderr, "access_started_cb\n");
#endif
  g_return_if_fail(IS_ICON_LABEL(tab_label));

  counter = g_object_get_qdata(G_OBJECT(tab_label), tab_label_icon_counter_id);
  *counter = 0;

  gdk_threads_enter();
  icon_label_set_icon(tab_label, tab_label_icons[TAB_LABEL_ICON_PREPARATION]);
  gdk_threads_leave();

  return;
}


static int progress_icon_number[7] =
{
  TAB_LABEL_ICON_PROGRESS0,
  TAB_LABEL_ICON_PROGRESS1,
  TAB_LABEL_ICON_PROGRESS2,
  TAB_LABEL_ICON_PROGRESS3,
  TAB_LABEL_ICON_PROGRESS2,
  TAB_LABEL_ICON_PROGRESS1,
  TAB_LABEL_ICON_PROGRESS0,
};


static void
access_progressed_cb(OchushaAsyncBuffer *buffer,
		     int bytes_read, int bytes_total,
		     IconLabel *tab_label)
{
  int *counter;
#if DEBUG_GUI_MOST
  fprintf(stderr, "access_progressed_cb\n");
#endif

  g_return_if_fail(IS_ICON_LABEL(tab_label));

  counter = g_object_get_qdata(G_OBJECT(tab_label), tab_label_icon_counter_id);
  *counter = (*counter + 1) % 7;

  if (*counter < 0 || *counter > 6)	/* ǰΤad-hoc˥å */
    *counter = 0;

  gdk_threads_enter();
  icon_label_set_icon(tab_label,
		      tab_label_icons[progress_icon_number[*counter]]);
  gdk_threads_leave();

  return;
}


static void
access_finished_cb(OchushaAsyncBuffer *buffer, IconLabel *tab_label)
{
#if DEBUG_GUI
  fprintf(stderr, "access_completed_cb\n");
#endif
  g_return_if_fail(IS_ICON_LABEL(tab_label));

  gdk_threads_enter();
  icon_label_set_icon(tab_label, tab_label_icons[TAB_LABEL_ICON_DONE]);
  gdk_threads_leave();

  return;
}


static void
access_failed_cb(OchushaAsyncBuffer *buffer,
		 OchushaNetworkBrokerFailureReason reason_code,
		 const gchar *reason, IconLabel *tab_label)
{
#if DEBUG_GUI
  gchar *native_reason = g_locale_from_utf8(reason, -1, NULL, NULL, NULL);
  fprintf(stderr,
	  "access_failed_cb(): buffer=%p, reason_code=%d, reason=\"%s\"\n",
	  buffer, reason_code, native_reason);
  g_free(native_reason);
#endif
  g_return_if_fail(IS_ICON_LABEL(tab_label));

  gdk_threads_enter();
  icon_label_set_icon(tab_label, tab_label_icons[TAB_LABEL_ICON_ERROR]);
  gdk_threads_leave();

  return;
}


static void
output_log_cb(OchushaNetworkBroker *broker, const gchar *message,
	      OchushaApplication *application)
{
#if 0
  fprintf(stderr, "LOG: %s", message);
#endif

#if ENABLE_DEBUG_CONSOLE
  if (enable_debug_outputs)
    {
      g_return_if_fail(GTK_IS_TEXT_BUFFER(log_buffer));

      gdk_threads_enter();
      gtk_text_buffer_insert(log_buffer, &log_tail_iter, message, -1);
      gdk_threads_leave();
    }
#endif
}


static void
preference_dialog_response_cb(GtkWidget *widget, int response_id,
			      OchushaApplication *application)
{
  switch (response_id)
    {
    case GTK_RESPONSE_APPLY:
    case GTK_RESPONSE_OK:
      {
	PanedNotebook *boardlist_pane
	  = PANED_NOTEBOOK(application->contents_window);
	GtkWidget *current_board_pane
	  = paned_notebook_get_current_item_view(boardlist_pane);
	gpointer tmp_data;

	tmp_data = g_object_get_data(G_OBJECT(widget), "web_browser_template");
	if (tmp_data != NULL)
	  {
	    const gchar *template = gtk_entry_get_text(GTK_ENTRY(tmp_data));
	    if (template != NULL && *template != '\0')
	      {
		G_FREE(application->web_browser_template);
		application->web_browser_template
		  = validate_web_browser_template(template);
	      }
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "focus_follow_mouse");
	if (tmp_data != NULL)
	  {
	    application->focus_follow_mouse
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "expand_only_one_category");
	if (tmp_data != NULL)
	  {
	    application->expand_only_one_category
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "board_tab_enable_tooltips");
	if (tmp_data != NULL)
	  {
	    application->board_tab_enable_tooltips
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	    paned_notebook_set_tooltips(boardlist_pane,
				application->board_tab_enable_tooltips);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "board_tab_always_show");
	if (tmp_data != NULL)
	  {
	    gboolean always_show
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	    application->board_tab_always_show = always_show;
	    paned_notebook_set_tab_policy(boardlist_pane,
					  always_show
					  ? GTK_POLICY_ALWAYS
					  : GTK_POLICY_AUTOMATIC);
	  }

#if 0	/* ɤ */
	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "boardlist_pane_style");
	if (tmp_data != NULL)
	  {
	    application->boardlist_pane_style
	      = (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data))
		 ? VPANED_STYLE : HPANED_STYLE);
	    paned_notebook_set_paned_style(boardlist_pane,
					   application->boardlist_pane_style);
	  }
#endif

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "threadlist_pane_style");
	if (tmp_data != NULL)
	  {
	    application->threadlist_pane_style
	      = (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data))
		 ? VPANED_STYLE : HPANED_STYLE);
	    if (current_board_pane != NULL)
	      paned_notebook_set_paned_style(PANED_NOTEBOOK(current_board_pane), application->threadlist_pane_style);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "threadlist_chunksize");
	if (tmp_data != NULL)
	  {
	    application->config.threadlist_chunksize
	      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "thread_chunksize");
	if (tmp_data != NULL)
	  {
	    application->config.thread_chunksize
	      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "download_chunksize");
	if (tmp_data != NULL)
	  {
	    application->download_chunksize
	      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "image_chunksize");
	if (tmp_data != NULL)
	  {
	    application->image_chunksize
	      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "enable_proxy");
	if (tmp_data != NULL)
	  {
	    application->config.enable_proxy
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "enable_proxy_only_for_posting");
	if (tmp_data != NULL)
	  {
	    application->config.enable_proxy_only_for_posting
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "proxy_url");
	if (tmp_data != NULL)
	  {
	    const char *url;
	    if (application->config.proxy_url != NULL)
	      G_FREE(application->config.proxy_url);
	    url = gtk_entry_get_text(GTK_ENTRY(tmp_data));
	    if (*url == '\0')
	      application->config.proxy_url = NULL;
	    else
	      application->config.proxy_url = G_STRDUP(url);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "proxy_user");
	if (tmp_data != NULL)
	  {
	    const gchar *name;
	    if (application->config.proxy_user != NULL)
	      G_FREE(application->config.proxy_user);
	    name = gtk_entry_get_text(GTK_ENTRY(tmp_data));
	    if (*name == '\0')
	      application->config.proxy_user = NULL;
	    else
	      application->config.proxy_user = G_STRDUP(name);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "proxy_password");
	if (tmp_data != NULL)
	  {
	    const gchar *password;
	    if (application->config.proxy_password != NULL)
	      G_FREE(application->config.proxy_password);
	    password = gtk_entry_get_text(GTK_ENTRY(tmp_data));
	    if (*password == '\0')
	      application->config.proxy_password = NULL;
	    else
	      application->config.proxy_password = G_STRDUP(password);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "enable_popup_title");
	if (tmp_data != NULL)
	  {
	    application->enable_popup_title
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "thread_tab_enable_tooltips");
	if (tmp_data != NULL)
	  {
	    application->thread_tab_enable_tooltips
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	    if (current_board_pane != NULL)
	      paned_notebook_set_tooltips(PANED_NOTEBOOK(current_board_pane),
				application->thread_tab_enable_tooltips);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "thread_tab_always_show");
	if (tmp_data != NULL)
	  {
	    gboolean always_show
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	    application->thread_tab_always_show = always_show;
	    if (current_board_pane != NULL)
	      paned_notebook_set_tab_policy(PANED_NOTEBOOK(current_board_pane),
					    always_show
					    ? GTK_POLICY_ALWAYS
					    : GTK_POLICY_AUTOMATIC);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "show_threadlist_entry");
	if (tmp_data != NULL)
	  {
	    application->show_threadlist_entry_when_thread_selected
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "filter_setting");

	if (tmp_data != NULL)
	  {
	    const ThreadlistFilter *filter;
	    filter = threadlist_filter_setting_get_filter(THREADLIST_FILTER_SETTING(tmp_data));
	    G_FREE(application->user_default_filter.rule);
	    application->user_default_filter.rule = G_STRDUP(filter->rule);
	    application->user_default_filter.ignore_threshold
	      = filter->ignore_threshold;
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "enable_popup_response");
	if (tmp_data != NULL)
	  {
	    application->enable_popup_response
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "popup_response_delay");
	if (tmp_data != NULL)
	  {
	    application->popup_response_delay
	      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "popup_close_delay");
	if (tmp_data != NULL)
	  {
	    application->popup_close_delay
	      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "enable_transparent_a_bone");
	if (tmp_data != NULL)
	  {
	    application->enable_transparent_a_bone
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "disable_popup_a_bone");
	if (tmp_data != NULL)
	  {
	    application->disable_popup_a_bone
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "show_mailto_literally");
	if (tmp_data != NULL)
	  {
	    application->show_mailto_mode
	      = (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data))
		 ? THREAD_VIEW_MAILTO_MODE_SHOW
		 : THREAD_VIEW_MAILTO_MODE_HIDE);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "enable_inline_image");
	if (tmp_data != NULL)
	  {
	    application->enable_inline_image
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "inline_image_width");
	if (tmp_data != NULL)
	  {
	    application->inline_image_width
	      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "inline_image_height");
	if (tmp_data != NULL)
	  {
	    application->inline_image_height
	      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "enable_popup_image");
	if (tmp_data != NULL)
	  {
	    application->enable_popup_image
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "popup_image_width");
	if (tmp_data != NULL)
	  {
	    application->popup_image_width
	      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "popup_image_height");
	if (tmp_data != NULL)
	  {
	    application->popup_image_height
	      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "maximum_number_of_cache_files");
	if (tmp_data != NULL)
	  {
	    GtkSpinButton *spin = GTK_SPIN_BUTTON(tmp_data);
	    int num_files = gtk_spin_button_get_value_as_int(spin);
	    if (num_files != 0
		&& num_files < OCHUSHA_DEFAULT_MAXIMUM_NUMBER_OF_CACHE_FILES)
	      {
		num_files = OCHUSHA_DEFAULT_MAXIMUM_NUMBER_OF_CACHE_FILES;
		gtk_spin_button_set_value(spin, num_files);
	      }
	    application->maximum_number_of_cache_files = num_files;
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "capacity_of_image_cache");
	if (tmp_data != NULL)
	  {
	    GtkSpinButton *spin = GTK_SPIN_BUTTON(tmp_data);
	    off_t capacity
	      = gtk_spin_button_get_value_as_int(spin);
	    if (capacity != 0
		&& capacity < OCHUSHA_DEFAULT_CAPACITY_OF_IMAGE_CACHE)
	      {
		capacity = OCHUSHA_DEFAULT_CAPACITY_OF_IMAGE_CACHE;
		gtk_spin_button_set_value(spin, capacity);
	      }
	    application->capacity_of_image_cache = capacity;
	  }

	ochusha_image_cache_ensure_limits(application);

	tmp_data = g_object_get_data(G_OBJECT(widget), "a_bone_by_name");
	if (tmp_data != NULL)
	  {
	    application->a_bone_by_name
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "a_bone_by_name_pattern");
	if (tmp_data != NULL)
	  {
	    if (application->a_bone_by_name_pattern != NULL)
	      G_FREE(application->a_bone_by_name_pattern);
	    application->a_bone_by_name_pattern
	      = regex_editor_get_regex_text(REGEX_EDITOR(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "a_bone_by_id");
	if (tmp_data != NULL)
	  {
	    application->a_bone_by_id
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "a_bone_by_id_pattern");
	if (tmp_data != NULL)
	  {
	    if (application->a_bone_by_id_pattern != NULL)
	      G_FREE(application->a_bone_by_id_pattern);
	    application->a_bone_by_id_pattern
	      = regex_editor_get_regex_text(REGEX_EDITOR(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "a_bone_by_content");
	if (tmp_data != NULL)
	  {
	    application->a_bone_by_content
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "a_bone_by_content_pattern");
	if (tmp_data != NULL)
	  {
	    if (application->a_bone_by_content_pattern != NULL)
	      G_FREE(application->a_bone_by_content_pattern);
	    application->a_bone_by_content_pattern
	      = regex_editor_get_regex_text(REGEX_EDITOR(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "show_threadlist_non_incrementally");
	if (tmp_data != NULL)
	  {
	    application->show_threadlist_non_incrementally
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "default_open_in_tab");
	if (tmp_data != NULL)
	  {
	    application->default_open_in_tab
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	    boardlist_view_set_default_open_in_tab(
				BOARDLIST_VIEW(application->boardlist_view),
				application->default_open_in_tab);
	    if (current_board_pane != NULL)
	      {
		GtkWidget *tmp_widget
		  = paned_notebook_get_selector(PANED_NOTEBOOK(current_board_pane));
		if (tmp_widget != NULL)
		  {
		    tmp_widget = gtk_bin_get_child(GTK_BIN(tmp_widget));
		    threadlist_view_set_default_open_in_tab(
					THREADLIST_VIEW(tmp_widget),
					application->default_open_in_tab);
		  }
	      }
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "select_tab_when_opened");
	if (tmp_data != NULL)
	  application->select_tab_when_opened
	    = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));

	tmp_data = g_object_get_data(G_OBJECT(widget), "hide_bookmark");
	if (tmp_data != NULL)
	  {
	    application->hide_bookmark
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "stick_bookmark");
	if (tmp_data != NULL)
	  {
	    application->stick_bookmark
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "bookmark_text");
	if (tmp_data != NULL)
	  {
	    GtkTextBuffer *text_buffer = GTK_TEXT_BUFFER(tmp_data);
	    GtkTextIter start_iter;
	    GtkTextIter end_iter;
	    gchar *new_text;
	    gtk_text_buffer_get_start_iter(text_buffer, &start_iter);
	    gtk_text_buffer_get_end_iter(text_buffer, &end_iter);
	    new_text = gtk_text_buffer_get_text(text_buffer, &start_iter,
						&end_iter, FALSE);
	    if (*new_text == '\0')
	      {
		G_FREE(new_text);	/* ǥեȤΤ᤹ */
		new_text = G_STRDUP(_("==== My bookmark: I have read above. ===="));
		if (response_id == GTK_RESPONSE_APPLY)
		  gtk_text_buffer_insert(text_buffer, &start_iter,
					 new_text, -1);
	      }

	    if (application->bookmark_text != NULL)
	      G_FREE(application->bookmark_text);
	    application->bookmark_text = new_text;
	  }

#if ENABLE_LOGIN_2CH_URL_ENTRY
	tmp_data = g_object_get_data(G_OBJECT(widget), "login_2ch_url");
	if (tmp_data != NULL)
	  {
	    const gchar *url;
	    if (application->config.login_2ch_url != NULL)
	      G_FREE(application->config.login_2ch_url);
	    url = gtk_entry_get_text(GTK_ENTRY(tmp_data));
	    if (*url == '\0')
	      application->config.login_2ch_url
		= G_STRDUP(OCHUSHA_DOLIB_URL);
	    else
	      application->config.login_2ch_url = G_STRDUP(url);
	  }
#endif

	tmp_data = g_object_get_data(G_OBJECT(widget), "dolib_id");
	if (tmp_data != NULL)
	  {
	    const gchar *id;
	    if (application->dolib_id != NULL)
	      G_FREE(application->dolib_id);
	    id = gtk_entry_get_text(GTK_ENTRY(tmp_data));
	    if (*id == '\0')
	      application->dolib_id = NULL;
	    else
	      application->dolib_id = G_STRDUP(id);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "dolib_passwd");
	if (tmp_data != NULL)
	  {
	    const gchar *passwd;
	    if (application->dolib_passwd != NULL)
	      G_FREE(application->dolib_passwd);
	    passwd = gtk_entry_get_text(GTK_ENTRY(tmp_data));
	    if (*passwd == '\0')
	      application->dolib_passwd = NULL;
	    else
	      application->dolib_passwd = G_STRDUP(passwd);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "dolib_save_passwd");
	if (tmp_data != NULL)
	  {
	    application->dolib_save_passwd
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "dolib_login_at_startup");
	if (tmp_data != NULL)
	  {
	    application->dolib_login_at_startup
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "dolib_use_id_for_posting");
	if (tmp_data != NULL)
	  {
	    application->dolib_use_id_for_posting
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "be_id");
	if (tmp_data != NULL)
	  {
	    const gchar *id;
	    if (application->be_id != NULL)
	      G_FREE(application->be_id);
	    id = gtk_entry_get_text(GTK_ENTRY(tmp_data));
	    if (*id == '\0')
	      application->be_id = NULL;
	    else
	      application->be_id = G_STRDUP(id);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "be_passwd");
	if (tmp_data != NULL)
	  {
	    const gchar *passwd;
	    if (application->be_passwd != NULL)
	      G_FREE(application->be_passwd);
	    passwd = gtk_entry_get_text(GTK_ENTRY(tmp_data));
	    if (*passwd == '\0')
	      application->be_passwd = NULL;
	    else
	      application->be_passwd = G_STRDUP(passwd);
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget), "be_save_passwd");
	if (tmp_data != NULL)
	  {
	    application->be_save_passwd
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "be_login_at_startup");
	if (tmp_data != NULL)
	  {
	    application->be_login_at_startup
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	tmp_data = g_object_get_data(G_OBJECT(widget),
				     "use_2ch_be_id_for_2ch");
	if (tmp_data != NULL)
	  {
	    application->config.use_2ch_be_id_for_2ch
	      = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(tmp_data));
	  }

	break;
      }

    case GTK_RESPONSE_CANCEL:
      break;
    }

  if (response_id == GTK_RESPONSE_APPLY)
    {
      return;
    }

  gtk_widget_hide(widget);
  gtk_widget_unrealize(widget);
  gtk_widget_destroy(widget);
}


static GtkWidget *
build_ochusha_setting_frame(OchushaApplication *application,
			    GtkWidget *dialog)
{
  GtkWidget *ochusha_frame = gtk_frame_new(_("Settings of Ochusha"));
  GtkWidget *outer_vbox = gtk_vbox_new(FALSE, 0);
  GtkWidget *hbox;
  GtkWidget *tmp_widget;

  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("External Web Browser:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);

  tmp_widget = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(tmp_widget), application->web_browser_template);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, TRUE, TRUE, 5);
  g_object_set_data(G_OBJECT(dialog), "web_browser_template",
		    tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(outer_vbox), hbox, FALSE, FALSE, 5);

  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("(%s will be replaced with URI)"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 100);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(outer_vbox), hbox, FALSE, FALSE, 0);

  gtk_widget_show(outer_vbox);
  gtk_container_add(GTK_CONTAINER(ochusha_frame), outer_vbox);

  tmp_widget = gtk_check_button_new_with_label(_("Focus Follows Mouse."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->focus_follow_mouse);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "focus_follow_mouse", tmp_widget);

  tmp_widget = gtk_check_button_new_with_label(_("Open in Tab by Default."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->default_open_in_tab);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "default_open_in_tab", tmp_widget);

  tmp_widget = gtk_check_button_new_with_label(_("Select the Tab When a New Tab is Opened."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->select_tab_when_opened);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "select_tab_when_opened", tmp_widget);

  tmp_widget = gtk_check_button_new_with_label(_("Expand Only One Category at Once."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->expand_only_one_category);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "expand_only_one_category",
		    tmp_widget);

  tmp_widget = gtk_check_button_new_with_label(_("Enable Tooltips of Board Tab Label."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->board_tab_enable_tooltips);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "board_tab_enable_tooltips",
		    tmp_widget);

  tmp_widget = gtk_check_button_new_with_label(_("Always Show Board Tab."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->board_tab_always_show);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "board_tab_always_show", tmp_widget);

#if 0
  tmp_widget = gtk_check_button_new_with_label(_("Place Boardlist and Threadlist Vertically."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->boardlist_pane_style
			       == VPANED_STYLE);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "boardlist_pane_style",
		    tmp_widget);
#endif

  tmp_widget = gtk_check_button_new_with_label(_("Place Threadlist and Thread Vertically."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->threadlist_pane_style
			       == VPANED_STYLE);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "threadlist_pane_style",
		    tmp_widget);

  return ochusha_frame;
}


static void
restore_default_cb(ThreadlistFilterSetting *setting,
		   OchushaApplication *application)
{
  threadlist_filter_setting_set_filter(setting,
				       &application->user_default_filter);

  setting->filter.ignore_threshold
    = application->user_default_filter.ignore_threshold;
}


static GtkWidget *
build_threadlist_view_setting_frame(OchushaApplication *application,
				    GtkWidget *dialog)
{
  GtkWidget *threadlist_frame
    = gtk_frame_new(_("Settings of Threadlist View"));
  GtkWidget *filter_frame
    = gtk_frame_new(_("Default Settings of Threadlist's Filter"));
  GtkWidget *outer_vbox = gtk_vbox_new(FALSE, 0);
  GtkWidget *vbox;
  GtkWidget *tmp_widget;

  gtk_widget_show(outer_vbox);
  gtk_container_add(GTK_CONTAINER(threadlist_frame), outer_vbox);

  tmp_widget = gtk_check_button_new_with_label(_("Show Threadlist Non Incrementally."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->show_threadlist_non_incrementally);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "show_threadlist_non_incrementally",
		    tmp_widget);

  tmp_widget = gtk_check_button_new_with_label(_("Enable Popup of Thread Title."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->enable_popup_title);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "enable_popup_title",
		    tmp_widget);

  tmp_widget = gtk_check_button_new_with_label(_("Enable Tooltips of Thread Tab Label."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->thread_tab_enable_tooltips);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "thread_tab_enable_tooltips",
		    tmp_widget);

  tmp_widget = gtk_check_button_new_with_label(_("Always Show Thread Tab."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->thread_tab_always_show);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "thread_tab_always_show", tmp_widget);

  tmp_widget = gtk_check_button_new_with_label(_("Show Threadlist Entry for Selected Thread."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->show_threadlist_entry_when_thread_selected);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "show_threadlist_entry",
		    tmp_widget);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(filter_frame), vbox);

  tmp_widget = threadlist_filter_setting_new_with_initial_setting(&application->user_default_filter);
  g_signal_connect(G_OBJECT(tmp_widget), "restore_default",
		   G_CALLBACK(restore_default_cb), application);

  g_object_set_data(G_OBJECT(dialog), "filter_setting", tmp_widget);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(vbox), tmp_widget, FALSE, FALSE, 0);
  gtk_widget_show(filter_frame);

  gtk_box_pack_start(GTK_BOX(outer_vbox), filter_frame, FALSE, FALSE, 5);

  return threadlist_frame;
}


static GtkWidget *
build_thread_view_setting_frame(OchushaApplication *application,
				GtkWidget *dialog)
{
  GtkWidget *outer_frame = gtk_frame_new(_("Settings of Thread View"));
  GtkWidget *outer_vbox = gtk_vbox_new(FALSE, 0);
  GtkWidget *thread_frame;
  GtkWidget *tmp_widget;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkObject *adjustment;

  gtk_widget_show(outer_vbox);
  gtk_container_add(GTK_CONTAINER(outer_frame), outer_vbox);

  tmp_widget = gtk_check_button_new_with_label(_("Enable Popup of Response."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->enable_popup_response);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "enable_popup_response",
		    tmp_widget);

  hbox = gtk_hbox_new(FALSE, 0);
  tmp_widget = gtk_label_new(_("Delay to Show Response Popup:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 10);

  adjustment = gtk_adjustment_new(application->popup_response_delay,
				  0.0, 10000.0, 100.0, 1000.0, 1000.0);
  tmp_widget = gtk_spin_button_new((GtkAdjustment *)adjustment, 100.0, 0);
  gtk_widget_show(tmp_widget);
  g_object_set_data(G_OBJECT(dialog), "popup_response_delay",
		    tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);

  tmp_widget = gtk_label_new(_("(ms)"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(outer_vbox), hbox, FALSE, FALSE, 0);
  
  hbox = gtk_hbox_new(FALSE, 0);
  tmp_widget = gtk_label_new(_("Delay to Close Response Popup:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 10);

  adjustment = gtk_adjustment_new(application->popup_close_delay,
				  0.0, 10000.0, 100.0, 1000.0, 1000.0);
  tmp_widget = gtk_spin_button_new((GtkAdjustment *)adjustment, 100.0, 0);
  gtk_widget_show(tmp_widget);
  g_object_set_data(G_OBJECT(dialog), "popup_close_delay",
		    tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);

  tmp_widget = gtk_label_new(_("(ms)"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(outer_vbox), hbox, FALSE, FALSE, 0);


  tmp_widget
    = gtk_check_button_new_with_label(_("Enable Transparent A Bone."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->enable_transparent_a_bone);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "enable_transparent_a_bone",
		    tmp_widget);

  tmp_widget = gtk_check_button_new_with_label(_("Disable Popup of Response Made A Bone."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->disable_popup_a_bone);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "disable_popup_a_bone", tmp_widget);


  thread_frame = gtk_frame_new(_("Default Settings of Thread Views"));

  vbox = gtk_vbox_new(FALSE, 0);

  gtk_widget_show(vbox);
  gtk_container_add(GTK_CONTAINER(thread_frame), vbox);

  tmp_widget = gtk_check_button_new_with_label(_("Show `mailto' Field Literally."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->show_mailto_mode
			       == THREAD_VIEW_MAILTO_MODE_SHOW);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "show_mailto_literally",
		    tmp_widget);


  tmp_widget = gtk_check_button_new_with_label(_("A Bone by Name Field."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->a_bone_by_name);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "a_bone_by_name", tmp_widget);

  tmp_widget = regex_editor_new(application->a_bone_by_name_pattern);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(vbox), tmp_widget, TRUE, TRUE, 0);
  g_object_set_data(G_OBJECT(dialog), "a_bone_by_name_pattern", tmp_widget);


  tmp_widget = gtk_check_button_new_with_label(_("A Bone by ID."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->a_bone_by_id);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "a_bone_by_id", tmp_widget);

  tmp_widget = regex_editor_new(application->a_bone_by_id_pattern);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(vbox), tmp_widget, TRUE, TRUE, 0);
  g_object_set_data(G_OBJECT(dialog), "a_bone_by_id_pattern", tmp_widget);


  tmp_widget = gtk_check_button_new_with_label(_("A Bone by Content."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->a_bone_by_content);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "a_bone_by_content", tmp_widget);

  tmp_widget = regex_editor_new(application->a_bone_by_content_pattern);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(vbox), tmp_widget, TRUE, TRUE, 0);
  g_object_set_data(G_OBJECT(dialog), "a_bone_by_content_pattern",
		    tmp_widget);


  gtk_widget_show(thread_frame);
  gtk_box_pack_start(GTK_BOX(outer_vbox), thread_frame, TRUE, TRUE, 5);

  return outer_frame;
}


static GtkWidget *
build_image_view_setting_frame(OchushaApplication *application,
				GtkWidget *dialog)
{
  GtkWidget *outer_frame = gtk_frame_new(_("Settings of Image View"));
  GtkWidget *outer_vbox = gtk_vbox_new(FALSE, 0);
  GtkWidget *inner_frame;
  GtkWidget *inner_vbox;
  GtkWidget *tmp_widget;
  GtkWidget *hbox;
  GtkObject *adjustment;

  gtk_widget_show(outer_vbox);
  gtk_container_add(GTK_CONTAINER(outer_frame), outer_vbox);

  tmp_widget = gtk_check_button_new_with_label(_("Enable Inline Preview of Images."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->enable_inline_image);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "enable_inline_image", tmp_widget);

  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Maximum Inline Image Size:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 10);

  tmp_widget = gtk_label_new(_("Width:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);

  adjustment = gtk_adjustment_new(application->inline_image_width,
				  1, 1024, 1, 32, 0);
  tmp_widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 16.0, 0);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "inline_image_width", tmp_widget);

  tmp_widget = gtk_label_new(_("x Height:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);

  adjustment = gtk_adjustment_new(application->inline_image_height,
				  1, 768, 1, 32, 0);
  tmp_widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 16.0, 0);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "inline_image_height", tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(outer_vbox), hbox, FALSE, FALSE, 0);


  tmp_widget
    = gtk_check_button_new_with_label(_("Enable Popup Preview of Images."));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->enable_popup_image);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "enable_popup_image", tmp_widget);


  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Maximum Popup Image Size:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 10);

  tmp_widget = gtk_label_new(_("Width:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);


  adjustment = gtk_adjustment_new(application->popup_image_width,
				  1, 2048, 1, 32, 0);
  tmp_widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 16.0, 0);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "popup_image_width", tmp_widget);

  tmp_widget = gtk_label_new(_("x Height:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);

  adjustment = gtk_adjustment_new(application->popup_image_height,
				  1, 1536, 1, 32, 0);
  tmp_widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 16.0, 0);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "popup_image_height", tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(outer_vbox), hbox, FALSE, FALSE, 0);

  inner_frame = gtk_frame_new(_("Settings of Image Cache"));
  gtk_widget_show(inner_frame);
  gtk_box_pack_end(GTK_BOX(outer_vbox), inner_frame, TRUE, TRUE, 10);
  
  inner_vbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(inner_vbox);
  gtk_container_add(GTK_CONTAINER(inner_frame), inner_vbox);



  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Maximum Number of Image Cache Files:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 10);

  adjustment = gtk_adjustment_new(application->maximum_number_of_cache_files,
				  0, 0x100000,
				  OCHUSHA_DEFAULT_MAXIMUM_NUMBER_OF_CACHE_FILES, 1024, 0);
  tmp_widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 16.0, 0);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "maximum_number_of_cache_files",
		    tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(inner_vbox), hbox, FALSE, FALSE, 0);


  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Capacity of Image Cache:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 10);

  adjustment = gtk_adjustment_new(application->capacity_of_image_cache,
				  0, 0x100000,
				  OCHUSHA_DEFAULT_CAPACITY_OF_IMAGE_CACHE,
				  1024, 0);
  tmp_widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 16.0, 0);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "capacity_of_image_cache",
		    tmp_widget);

  tmp_widget = gtk_label_new(_("(MB)"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(inner_vbox), hbox, FALSE, FALSE, 0);


  return outer_frame;
}


static GtkWidget *
build_bookmark_setting_frame(OchushaApplication *application,
			     GtkWidget *dialog)
{
  GtkWidget *outer_frame = gtk_frame_new(_("Settings of My Bookmark"));
  GtkWidget *outer_vbox = gtk_vbox_new(FALSE, 0);
  GtkWidget *tmp_widget;
  GtkWidget *tmp_container;
  GtkTextBuffer *text_buffer;
  GtkTextIter iter;

  gtk_widget_show(outer_vbox);
  gtk_container_add(GTK_CONTAINER(outer_frame), outer_vbox);


  tmp_widget = gtk_check_button_new_with_label(_("Do NOT Show Bookmarks."));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "hide_bookmark", tmp_widget);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->hide_bookmark);

  tmp_widget = gtk_radio_button_new_with_label(NULL, _("Do NOT Move Bookmarks Automatically."));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "stick_bookmark", tmp_widget);

  tmp_widget
    = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(tmp_widget),
						  _("Move Bookmarks to Lead Unread Responses before Loading and Refreshment of Threads."));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_widget, FALSE, FALSE, 0);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       !application->stick_bookmark);


  tmp_container = gtk_hbox_new(FALSE, 0);
  tmp_widget = gtk_label_new(_("My Bookmark"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(tmp_container), tmp_widget, FALSE, FALSE, 5);
  gtk_widget_show(tmp_container);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_container, FALSE, FALSE, 5);


  tmp_container = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(tmp_container),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  text_buffer = gtk_text_buffer_new(NULL);
  gtk_text_buffer_get_start_iter(text_buffer, &iter);
  gtk_text_buffer_insert(text_buffer, &iter, application->bookmark_text, -1);
  g_object_set_data(G_OBJECT(dialog), "bookmark_text", text_buffer);

  tmp_widget = gtk_text_view_new_with_buffer(text_buffer);
  gtk_widget_show(tmp_widget);
  gtk_container_add(GTK_CONTAINER(tmp_container), tmp_widget);
  gtk_widget_show(tmp_container);
  gtk_box_pack_start(GTK_BOX(outer_vbox), tmp_container, TRUE, TRUE, 0);

  return outer_frame;
}


static GtkWidget *
build_network_setting_frame(OchushaApplication *application,
			    GtkWidget *dialog)
{
  GtkWidget *network_frame = gtk_frame_new(_("Settings of Network"));
  GtkWidget *outer_vbox = gtk_vbox_new(FALSE, 0);
  GtkWidget *proxy_frame;
  GtkWidget *proxy_vbox;
  GtkWidget *hbox;
  GtkWidget *tmp_widget;
  GtkObject *adjustment;

  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Chunksize for Downloading of Threadlist:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 10);

  adjustment = gtk_adjustment_new(application->config.threadlist_chunksize,
				  1024, 0x40000, 1024, 0x10000, 0);
  tmp_widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 16.0, 0);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "threadlist_chunksize", tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(outer_vbox), hbox, FALSE, FALSE, 0);


  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Chunksize for Downloading of Thread(DAT):"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 10);

  adjustment = gtk_adjustment_new(application->config.thread_chunksize,
				  1024, 0x40000, 1024, 0x10000, 0);
  tmp_widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 16.0, 0);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "thread_chunksize", tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(outer_vbox), hbox, FALSE, FALSE, 0);

  gtk_widget_show(outer_vbox);
  gtk_container_add(GTK_CONTAINER(network_frame), outer_vbox);


  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Chunksize for Downloading of Files:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 10);

  adjustment = gtk_adjustment_new(application->download_chunksize,
				  1024, 0x40000, 1024, 0x10000, 0);
  tmp_widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 16.0, 0);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "download_chunksize", tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(outer_vbox), hbox, FALSE, FALSE, 0);


  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Chunksize for Downloading of Images:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 10);

  adjustment = gtk_adjustment_new(application->image_chunksize,
				  1024, 0x40000, 1024, 0x10000, 0);
  tmp_widget = gtk_spin_button_new(GTK_ADJUSTMENT(adjustment), 16.0, 0);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "image_chunksize", tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(outer_vbox), hbox, FALSE, FALSE, 0);



  proxy_frame = gtk_frame_new(_("Proxy Settings"));
  proxy_vbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(proxy_vbox);
  gtk_container_add(GTK_CONTAINER(proxy_frame), proxy_vbox);

  hbox = gtk_hbox_new(FALSE, 0);
  tmp_widget = gtk_label_new(_("WARNING: Use of PROXY never tested.  If you know whether this works well or not, please report that."));
  gtk_label_set_justify(GTK_LABEL(tmp_widget), GTK_JUSTIFY_LEFT);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(proxy_vbox), hbox, FALSE, FALSE, 5);

  hbox = gtk_hbox_new(FALSE, 0);
  tmp_widget = gtk_label_new(_("WARNING: If specified PROXY server isn't alive, Ochusha may hang up."));
  gtk_label_set_justify(GTK_LABEL(tmp_widget), GTK_JUSTIFY_LEFT);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(proxy_vbox), hbox, FALSE, FALSE, 5);

  hbox = gtk_hbox_new(FALSE, 0);
  tmp_widget = gtk_label_new(_("WARNING: Sorry, I cannot recommend to use PROXY at this time."));
  gtk_label_set_justify(GTK_LABEL(tmp_widget), GTK_JUSTIFY_LEFT);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);
  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(proxy_vbox), hbox, FALSE, FALSE, 5);

  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_radio_button_new_with_label(NULL, _("Don't Use Proxy"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);

  tmp_widget
    = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(tmp_widget),
						  _("Use Proxy"));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->config.enable_proxy);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "enable_proxy", tmp_widget);

  tmp_widget
    = gtk_radio_button_new_with_label_from_widget(GTK_RADIO_BUTTON(tmp_widget),
						  _("Use Proxy Only for Posting"));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->config.enable_proxy_only_for_posting);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 0);
  g_object_set_data(G_OBJECT(dialog), "enable_proxy_only_for_posting",
		    tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(proxy_vbox), hbox, FALSE, FALSE, 0);



  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Proxy URL:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);

  tmp_widget = gtk_entry_new();
  if (application->config.proxy_url != NULL)
    gtk_entry_set_text(GTK_ENTRY(tmp_widget), application->config.proxy_url);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, TRUE, TRUE, 5);
  g_object_set_data(G_OBJECT(dialog), "proxy_url", tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(proxy_vbox), hbox, FALSE, FALSE, 0);


  hbox = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Proxy User:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);

  tmp_widget = gtk_entry_new();
  if (application->config.proxy_user != NULL)
    gtk_entry_set_text(GTK_ENTRY(tmp_widget), application->config.proxy_user);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, TRUE, TRUE, 5);
  g_object_set_data(G_OBJECT(dialog), "proxy_user", tmp_widget);

  tmp_widget = gtk_label_new(_("Proxy Password:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, FALSE, FALSE, 5);

  tmp_widget = gtk_entry_new();
  if (application->config.proxy_password != NULL)
    gtk_entry_set_text(GTK_ENTRY(tmp_widget),
		       application->config.proxy_password);
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(hbox), tmp_widget, TRUE, TRUE, 5);
  g_object_set_data(G_OBJECT(dialog), "proxy_password", tmp_widget);

  gtk_widget_show(hbox);
  gtk_box_pack_start(GTK_BOX(proxy_vbox), hbox, FALSE, FALSE, 5);


  gtk_widget_show(proxy_frame);
  gtk_box_pack_end(GTK_BOX(outer_vbox), proxy_frame, FALSE, FALSE, 5);

  return network_frame;
}


static void
be_login_button_toggled_cb(GtkToggleButton *button,
			   OchushaApplication *application)
{
  gpointer tmp_data;
  const gchar *id;
  const gchar *passwd;

  if (!gtk_toggle_button_get_active(button))
    return;

  tmp_data = g_object_get_data(G_OBJECT(button), "be_id_entry");
  if (tmp_data != NULL)
    {
      id = gtk_entry_get_text(GTK_ENTRY(tmp_data));
      if (*id == '\0')
	id = NULL;
    }
  else
    id = NULL;

  tmp_data = g_object_get_data(G_OBJECT(button), "be_passwd_entry");
  if (tmp_data != NULL)
    {
      passwd = gtk_entry_get_text(GTK_ENTRY(tmp_data));
      if (*passwd == '\0')
	passwd = NULL;
    }
  else
    passwd = NULL;

  if (id != NULL && passwd != NULL)
    {
      gdk_threads_leave();
      ochusha_utils_2ch_be_login(&application->config, application->broker,
				 id, passwd);
      gdk_threads_enter();
    }

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
			       application->config.session_id_2ch_be != NULL);
  gtk_widget_set_sensitive(GTK_WIDGET(button),
			   application->config.session_id_2ch_be == NULL);

  tmp_data = g_object_get_data(G_OBJECT(button), "be_logout_button");
  if (tmp_data != NULL)
    {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_data),
				   application->config.session_id_2ch_be == NULL);
      gtk_widget_set_sensitive(GTK_WIDGET(tmp_data),
			       application->config.session_id_2ch_be != NULL);
    }
}


static void
be_logout_button_toggled_cb(GtkToggleButton *button,
			 OchushaApplication *application)
{
  gpointer tmp_data;

  if (!gtk_toggle_button_get_active(button))
    return;

  ochusha_utils_2ch_be_logout(&application->config);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
			       application->config.session_id_2ch_be == NULL);
  gtk_widget_set_sensitive(GTK_WIDGET(button),
			   application->config.session_id_2ch_be != NULL);

  tmp_data = g_object_get_data(G_OBJECT(button), "be_login_button");
  if (tmp_data != NULL)
    {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_data),
				   application->config.session_id_2ch_be != NULL);
      gtk_widget_set_sensitive(GTK_WIDGET(tmp_data),
			       application->config.session_id_2ch_be == NULL);
    }
}


static void
login_button_toggled_cb(GtkToggleButton *button,
			OchushaApplication *application)
{
  gpointer tmp_data;
  const gchar *id;
  const gchar *passwd;

  if (!gtk_toggle_button_get_active(button))
    return;

  tmp_data = g_object_get_data(G_OBJECT(button), "id_entry");
  if (tmp_data != NULL)
    {
      id = gtk_entry_get_text(GTK_ENTRY(tmp_data));
      if (*id == '\0')
	id = NULL;
    }
  else
    id = NULL;

  tmp_data = g_object_get_data(G_OBJECT(button), "passwd_entry");
  if (tmp_data != NULL)
    {
      passwd = gtk_entry_get_text(GTK_ENTRY(tmp_data));
      if (*passwd == '\0')
	passwd = NULL;
    }
  else
    passwd = NULL;

  if (id != NULL && passwd != NULL)
    {
      gdk_threads_leave();
      ochusha_utils_2ch_login(&application->config, application->broker,
			      id, passwd);
      gdk_threads_enter();
    }

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
			       application->config.login_2ch);
  gtk_widget_set_sensitive(GTK_WIDGET(button), !application->config.login_2ch);

  tmp_data = g_object_get_data(G_OBJECT(button), "logout_button");
  if (tmp_data != NULL)
    {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_data),
				   !application->config.login_2ch);
      gtk_widget_set_sensitive(GTK_WIDGET(tmp_data),
			       application->config.login_2ch);
    }
}


static void
logout_button_toggled_cb(GtkToggleButton *button,
			 OchushaApplication *application)
{
  gpointer tmp_data;

  if (!gtk_toggle_button_get_active(button))
    return;

  ochusha_utils_2ch_logout(&application->config);

  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button),
			       !application->config.login_2ch);
  gtk_widget_set_sensitive(GTK_WIDGET(button), application->config.login_2ch);

  tmp_data = g_object_get_data(G_OBJECT(button), "login_button");
  if (tmp_data != NULL)
    {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_data),
				   application->config.login_2ch);
      gtk_widget_set_sensitive(GTK_WIDGET(tmp_data),
			       !application->config.login_2ch);
    }
}


static GtkWidget *
build_dolib_setting_frame(OchushaApplication *application, GtkWidget *dialog)
{
  GtkWidget *outer_vbox = gtk_vbox_new(FALSE, 0);
  GtkWidget *inner_frame;
  GtkWidget *inner_vbox;
  GtkWidget *tmp_widget;
  GtkWidget *tmp_container;
#if ENABLE_LOGIN_2CH_URL_ENTRY
  GtkWidget *url_entry;
#endif
  GtkWidget *id_entry;
  GtkWidget *passwd_entry;
  GtkWidget *login_button;
  GtkWidget *logout_button;


  /* 2ch Be */
  inner_frame = gtk_frame_new(_("Settings of 2ch (Be) ID"));
  gtk_widget_show(inner_frame);
  gtk_container_add(GTK_CONTAINER(outer_vbox), inner_frame);

  inner_vbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(inner_vbox);
  gtk_container_add(GTK_CONTAINER(inner_frame), inner_vbox);


  tmp_container = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("User ID:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(tmp_container), tmp_widget, FALSE, FALSE, 5);

  id_entry = gtk_entry_new();
  if (application->be_id != NULL)
    gtk_entry_set_text(GTK_ENTRY(id_entry), application->be_id);
  gtk_widget_show(id_entry);
  gtk_box_pack_start(GTK_BOX(tmp_container), id_entry, TRUE, TRUE, 5);
  g_object_set_data(G_OBJECT(dialog), "be_id", id_entry);

  gtk_widget_show(tmp_container);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_container, FALSE, FALSE, 5);


  tmp_container = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Password:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(tmp_container), tmp_widget, FALSE, FALSE, 5);

  passwd_entry = gtk_entry_new();
  if (application->be_passwd != NULL)
    gtk_entry_set_text(GTK_ENTRY(passwd_entry), application->be_passwd);
  gtk_entry_set_visibility(GTK_ENTRY(passwd_entry), FALSE);
  gtk_widget_show(passwd_entry);
  gtk_box_pack_start(GTK_BOX(tmp_container), passwd_entry, TRUE, TRUE, 5);
  g_object_set_data(G_OBJECT(dialog), "be_passwd", passwd_entry);

  gtk_widget_show(tmp_container);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_container, FALSE, FALSE, 5);


  tmp_widget = gtk_check_button_new_with_label(_("Remember Password."));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "be_save_passwd", tmp_widget);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->be_save_passwd);


  tmp_widget = gtk_check_button_new_with_label(_("Login at Startup."));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "be_login_at_startup", tmp_widget);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->be_login_at_startup);


  tmp_widget = gtk_check_button_new_with_label(_("Use Be ID to post non-Be 2ch BBSs as well."));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "use_2ch_be_id_for_2ch", tmp_widget);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->config.use_2ch_be_id_for_2ch);


  tmp_container = gtk_hbox_new(FALSE, 0);

  login_button = gtk_toggle_button_new_with_label(_("Login Now"));
  gtk_widget_show(login_button);
  gtk_box_pack_start(GTK_BOX(tmp_container), login_button, FALSE, FALSE, 5);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(login_button),
			       application->config.session_id_2ch_be != NULL);
  gtk_widget_set_sensitive(login_button,
			   application->config.session_id_2ch_be == NULL);
  g_signal_connect(G_OBJECT(login_button), "toggled",
		   G_CALLBACK(be_login_button_toggled_cb), application);

  logout_button = gtk_toggle_button_new_with_label(_("Logout Now"));
  gtk_widget_show(logout_button);
  gtk_box_pack_start(GTK_BOX(tmp_container), logout_button, FALSE, FALSE, 5);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(logout_button),
			       application->config.session_id_2ch_be == NULL);
  gtk_widget_set_sensitive(logout_button,
			   application->config.session_id_2ch_be != NULL);
  g_signal_connect(G_OBJECT(logout_button), "toggled",
		   G_CALLBACK(be_logout_button_toggled_cb), application);

  g_object_set_data(G_OBJECT(login_button), "be_logout_button", logout_button);
  g_object_set_data(G_OBJECT(login_button), "be_id_entry", id_entry);
  g_object_set_data(G_OBJECT(login_button), "be_passwd_entry", passwd_entry);
  g_object_set_data(G_OBJECT(logout_button), "be_login_button", login_button);

  gtk_widget_show(tmp_container);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_container, FALSE, FALSE, 5);


  /*  */
  inner_frame = gtk_frame_new(_("Settings of 2ch (Oyster) ID"));
  gtk_widget_show(inner_frame);
  gtk_container_add(GTK_CONTAINER(outer_vbox), inner_frame);

  inner_vbox = gtk_vbox_new(FALSE, 0);
  gtk_widget_show(inner_vbox);
  gtk_container_add(GTK_CONTAINER(inner_frame), inner_vbox);

#if ENABLE_LOGIN_2CH_URL_ENTRY
  tmp_container = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Auth. URL:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(tmp_container), tmp_widget, FALSE, FALSE, 5);

  url_entry = gtk_entry_new();
  if (application->config.login_2ch_url != NULL)
    gtk_entry_set_text(GTK_ENTRY(url_entry),
		       application->config.login_2ch_url);
  else
    gtk_entry_set_text(GTK_ENTRY(url_entry), OCHUSHA_DOLIB_URL);
  gtk_widget_show(url_entry);
  gtk_box_pack_start(GTK_BOX(tmp_container), url_entry, TRUE, TRUE, 5);
  g_object_set_data(G_OBJECT(dialog), "login_2ch_url", url_entry);

  gtk_widget_show(tmp_container);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_container, FALSE, FALSE, 5);
#endif

  tmp_container = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("User ID:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(tmp_container), tmp_widget, FALSE, FALSE, 5);

  id_entry = gtk_entry_new();
  if (application->dolib_id != NULL)
    gtk_entry_set_text(GTK_ENTRY(id_entry), application->dolib_id);
  gtk_widget_show(id_entry);
  gtk_box_pack_start(GTK_BOX(tmp_container), id_entry, TRUE, TRUE, 5);
  g_object_set_data(G_OBJECT(dialog), "dolib_id", id_entry);

  gtk_widget_show(tmp_container);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_container, FALSE, FALSE, 5);


  tmp_container = gtk_hbox_new(FALSE, 0);

  tmp_widget = gtk_label_new(_("Password:"));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(tmp_container), tmp_widget, FALSE, FALSE, 5);

  passwd_entry = gtk_entry_new();
  if (application->dolib_passwd != NULL)
    gtk_entry_set_text(GTK_ENTRY(passwd_entry), application->dolib_passwd);
  gtk_entry_set_visibility(GTK_ENTRY(passwd_entry), FALSE);
  gtk_widget_show(passwd_entry);
  gtk_box_pack_start(GTK_BOX(tmp_container), passwd_entry, TRUE, TRUE, 5);
  g_object_set_data(G_OBJECT(dialog), "dolib_passwd", passwd_entry);

  gtk_widget_show(tmp_container);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_container, FALSE, FALSE, 5);


  tmp_widget = gtk_check_button_new_with_label(_("Remember Password."));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "dolib_save_passwd", tmp_widget);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->dolib_save_passwd);


  tmp_widget = gtk_check_button_new_with_label(_("Login at Startup."));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "dolib_login_at_startup", tmp_widget);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->dolib_login_at_startup);


  tmp_widget = gtk_check_button_new_with_label(_("Use ID for Posting."));
  gtk_widget_show(tmp_widget);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_widget, FALSE, FALSE, 5);
  g_object_set_data(G_OBJECT(dialog), "dolib_use_id_for_posting", tmp_widget);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp_widget),
			       application->dolib_use_id_for_posting);


  tmp_container = gtk_hbox_new(FALSE, 0);

  login_button = gtk_toggle_button_new_with_label(_("Login Now"));
  gtk_widget_show(login_button);
  gtk_box_pack_start(GTK_BOX(tmp_container), login_button, FALSE, FALSE, 5);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(login_button),
			       application->config.login_2ch);
  gtk_widget_set_sensitive(login_button, !application->config.login_2ch);
  g_signal_connect(G_OBJECT(login_button), "toggled",
		   G_CALLBACK(login_button_toggled_cb), application);

  logout_button = gtk_toggle_button_new_with_label(_("Logout Now"));
  gtk_widget_show(logout_button);
  gtk_box_pack_start(GTK_BOX(tmp_container), logout_button, FALSE, FALSE, 5);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(logout_button),
			       !application->config.login_2ch);
  gtk_widget_set_sensitive(logout_button, application->config.login_2ch);
  g_signal_connect(G_OBJECT(logout_button), "toggled",
		   G_CALLBACK(logout_button_toggled_cb), application);

  g_object_set_data(G_OBJECT(login_button), "logout_button", logout_button);
  g_object_set_data(G_OBJECT(login_button), "id_entry", id_entry);
  g_object_set_data(G_OBJECT(login_button), "passwd_entry", passwd_entry);
  g_object_set_data(G_OBJECT(logout_button), "login_button", login_button);

  gtk_widget_show(tmp_container);
  gtk_box_pack_start(GTK_BOX(inner_vbox), tmp_container, FALSE, FALSE, 5);


  return outer_vbox;
}


#if ENABLE_DEBUG_CONSOLE
static void
debug_console_response_cb(GtkWidget *widget, int response_id, gpointer unused)
{
  if (widget == NULL)
    return;

  switch (response_id)
    {
    case GTK_RESPONSE_REJECT:
      {
	GtkTextIter start_iter;
	GtkTextIter end_iter;
	g_return_if_fail(GTK_IS_TEXT_BUFFER(log_buffer));

	gtk_text_buffer_get_start_iter(log_buffer, &start_iter);
	gtk_text_buffer_get_end_iter(log_buffer, &end_iter);
	gtk_text_buffer_delete(log_buffer, &start_iter, &end_iter);
	gtk_text_buffer_get_start_iter(log_buffer, &log_tail_iter);
      }
      break;

    case GTK_RESPONSE_OK:
    default:
      gtk_widget_hide(widget);
    }
}


static void
display_debug_output_toggled_cb(GtkToggleButton *button,
				OchushaApplication *application)
{
  enable_debug_outputs = gtk_toggle_button_get_active(button);
}


static void
open_debug_console(OchushaApplication *application)
{
  GtkWidget *dialog;
  GtkWidget *scroll;
  GtkWidget *text_view;
  GtkWidget *button;

  if (debug_console != NULL)
    {
      gtk_widget_show_all(debug_console);
      return;
    }

  dialog = gtk_dialog_new_with_buttons(_("Debug Outputs"),
				       NULL,
				       GTK_DIALOG_DESTROY_WITH_PARENT,
				       GTK_STOCK_CLEAR, GTK_RESPONSE_REJECT,
				       GTK_STOCK_OK, GTK_RESPONSE_OK,
				       NULL);
  g_signal_connect(G_OBJECT(dialog), "response",
		   G_CALLBACK(debug_console_response_cb), application);
  g_signal_connect(G_OBJECT(dialog), "destroy",
		   G_CALLBACK(gtk_widget_destroyed), &debug_console);
  gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);

  scroll = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
				 GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  text_view = gtk_text_view_new_with_buffer(log_buffer);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(text_view), FALSE);
  gtk_widget_show(text_view);
  gtk_container_add(GTK_CONTAINER(scroll), text_view);
  gtk_widget_show(scroll);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), scroll,
		     TRUE, TRUE, 0);

  button = gtk_check_button_new_with_label(_("Display Debug Outputs in This Console."));
  gtk_widget_show(button);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), button,
		     FALSE, FALSE, 0);
  g_signal_connect(G_OBJECT(button), "toggled",
		   G_CALLBACK(display_debug_output_toggled_cb), application);

  gtk_widget_show(dialog);
  debug_console = dialog;
}
#endif


static void
open_preference_dialog(OchushaApplication *application)
{
  MAIN_UI_LOCK
  {
    if (preference_dialog == NULL)
      {
	GtkWidget *notebook;
	GtkWidget *tmp_widget;
	preference_dialog
	  = gtk_dialog_new_with_buttons(_("Preferences of Ochusha"),
					application->top_level,
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_APPLY, GTK_RESPONSE_APPLY,
					GTK_STOCK_OK, GTK_RESPONSE_OK,
					NULL);
	gtk_dialog_set_default_response(GTK_DIALOG(preference_dialog),
					GTK_RESPONSE_OK);
	g_signal_connect(G_OBJECT(preference_dialog), "response",
			 G_CALLBACK(preference_dialog_response_cb),
			 application);
	g_signal_connect(G_OBJECT(preference_dialog), "destroy",
			 G_CALLBACK(gtk_widget_destroyed),
			 &preference_dialog);

	notebook = gtk_notebook_new();
	gtk_widget_show(notebook);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(preference_dialog)->vbox),
			  notebook);

	tmp_widget = build_ochusha_setting_frame(application,
						 preference_dialog);
	gtk_widget_show(tmp_widget);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tmp_widget,
				 gtk_label_new(_("General")));

	tmp_widget = build_threadlist_view_setting_frame(application,
							 preference_dialog);
	gtk_widget_show(tmp_widget);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tmp_widget,
				 gtk_label_new(_("Threadlist View")));

	tmp_widget = build_thread_view_setting_frame(application,
						     preference_dialog);
	gtk_widget_show(tmp_widget);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tmp_widget,
				 gtk_label_new(_("Thread View")));

	tmp_widget = build_image_view_setting_frame(application,
						    preference_dialog);
	gtk_widget_show(tmp_widget);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tmp_widget,
				 gtk_label_new(_("Image View")));

	tmp_widget = build_bookmark_setting_frame(application,
						  preference_dialog);
	gtk_widget_show(tmp_widget);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tmp_widget,
				 gtk_label_new(_("My Bookmark")));

	tmp_widget = build_network_setting_frame(application,
						 preference_dialog);
	gtk_widget_show(tmp_widget);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tmp_widget,
				 gtk_label_new(_("Network")));

	tmp_widget = build_dolib_setting_frame(application,
					       preference_dialog);
	gtk_widget_show(tmp_widget);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tmp_widget,
				 gtk_label_new(_("2ch ID")));

	gtk_widget_show(preference_dialog);
      }
  }
  MAIN_UI_UNLOCK;
}


static void
open_about_dialog(OchushaApplication *application)
{
  MAIN_UI_LOCK
  {
    if (about_dialog == NULL)
      {
	gchar text_buffer[1024];
	GtkWidget *label;
	GdkPixbuf *banner = NULL;

	about_dialog = gtk_dialog_new_with_buttons(_("About Ochusha"),
					application->top_level,
					GTK_DIALOG_DESTROY_WITH_PARENT,
					GTK_STOCK_OK, GTK_RESPONSE_OK,
					NULL);
	gtk_dialog_set_default_response(GTK_DIALOG(about_dialog),
					GTK_RESPONSE_OK);
	g_signal_connect(G_OBJECT(about_dialog), "response",
			 G_CALLBACK(about_dialog_response_cb),
			 NULL);
	g_signal_connect(G_OBJECT(about_dialog), "destroy",
			 G_CALLBACK(gtk_widget_destroyed),
			 &about_dialog);

	snprintf(text_buffer, 1024, "%s %s %s",
		 _("Ochusha"), _("Version"), PACKAGE_VERSION);
	label = gtk_label_new(text_buffer);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(about_dialog)->vbox),
			  label);

#ifdef HAVE_LOCALTIME_R
	{
	  struct tm tm;
	  time_t utc = time(NULL);
	  localtime_r(&utc, &tm);
	  banner = NULL;

#ifdef NEW_YEAR_SPECIAL
	  if (tm.tm_year == 104 && tm.tm_yday < 3)
	    banner = banner1_pixbuf;
#endif
#ifdef XMAS_SPECIAL
	  if (tm.tm_yday >= 357 && tm.tm_yday <= 358)
	    banner = banner2_pixbuf;
#endif
	  if (banner == NULL)
	    banner = banner_pixbuf;
	}
#else
	banner = banner_pixbuf;
#endif
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(about_dialog)->vbox),
			  gtk_image_new_from_pixbuf(banner));

	label = gtk_label_new("Copyright(c) 2003-2004 The Ochusha Project");
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(about_dialog)->vbox),
			  label);

	gtk_widget_show_all(about_dialog);
      }
  }
  MAIN_UI_UNLOCK;
}


static void
about_dialog_response_cb(GtkWidget *widget, int response_id, gpointer unused)
{
  if (widget == NULL)
    return;

  gtk_widget_hide(widget);
  gtk_widget_unrealize(widget);
  gtk_widget_destroy(widget);
}


#if GTK_MINOR_VERSION > 2
static void
append_url_func(const char *url, GtkComboBox *combo_box)
{
  gtk_combo_box_append_text(combo_box, url);
}
#endif


static void
initialize_open_url_dialog(OchushaApplication *application)
{
  GtkWidget *tmp_hbox;
  open_url_dialog = gtk_message_dialog_new(application->top_level,
					   GTK_DIALOG_DESTROY_WITH_PARENT,
					   GTK_MESSAGE_QUESTION,
					   GTK_BUTTONS_OK_CANCEL,
					   _("Input the URL of the document you want to open."));
  gtk_dialog_set_default_response(GTK_DIALOG(open_url_dialog),
				  GTK_RESPONSE_OK);
  g_signal_connect(G_OBJECT(open_url_dialog), "response",
		   G_CALLBACK(open_url_dialog_response_cb),
		   application);
  g_signal_connect(G_OBJECT(open_url_dialog), "destroy",
		   G_CALLBACK(gtk_widget_destroyed), &open_url_dialog);

#if GTK_MINOR_VERSION <= 2
  open_url_combo = gtk_combo_new();
  gtk_entry_set_width_chars(GTK_ENTRY(GTK_COMBO(open_url_combo)->entry), 48);
  if (open_url_history != NULL)
    gtk_combo_set_popdown_strings(GTK_COMBO(open_url_combo),
				  open_url_history);
  gtk_combo_disable_activate(GTK_COMBO(open_url_combo));
  g_signal_connect(G_OBJECT(GTK_COMBO(open_url_combo)->entry), "activate",
		   G_CALLBACK(open_url_combo_entry_activate_cb),
		   open_url_dialog);
#else
  open_url_combo = gtk_combo_box_entry_new_text();
  gtk_entry_set_width_chars(GTK_ENTRY(GTK_BIN(open_url_combo)->child), 48);
  if (open_url_history != NULL)
    {
      g_list_foreach(open_url_history, (GFunc)append_url_func, open_url_combo);
    }
  g_signal_connect(G_OBJECT(GTK_BIN(open_url_combo)->child), "activate",
		   G_CALLBACK(open_url_combo_entry_activate_cb),
		   open_url_dialog);
#endif
  tmp_hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(tmp_hbox), gtk_label_new(_("URL: ")),
		     FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX(tmp_hbox), open_url_combo, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(open_url_dialog)->vbox),
		     tmp_hbox, FALSE, FALSE, 0);
  open_url_in_tab_button = gtk_check_button_new_with_label(_("Open in Tab"));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(open_url_dialog)->vbox),
		     open_url_in_tab_button, FALSE, FALSE, 0);
}


static void
open_url_combo_entry_activate_cb(GtkWidget *entry, GtkDialog *dialog)
{
  gtk_dialog_response(dialog, GTK_RESPONSE_OK);
}


static void
open_url_dialog_response_cb(GtkWidget *dialog, int response_id,
			    OchushaApplication *application)
{
  if (response_id == GTK_RESPONSE_OK)
    {
#if GTK_MINOR_VERSION <= 2
      const char *url
	= gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(open_url_combo)->entry));
#else
      const char *url
	= gtk_entry_get_text(GTK_ENTRY(GTK_BIN(open_url_combo)->child));
#endif
      gboolean in_tab = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(open_url_in_tab_button));
      ochusha_open_url(application, url, in_tab, FALSE);
    }

  gtk_widget_hide_all(open_url_dialog);
}


static void
popup_open_url_dialog(OchushaApplication *application)
{
  if (open_url_dialog == NULL)
    initialize_open_url_dialog(application);
  gtk_widget_show_all(open_url_dialog);
}


void
ochusha_create_new_category(OchushaApplication *application)
{
}


void
ochusha_add_board(OchushaApplication *application, const char *url)
{
}


void
ochusha_create_new_virtual_board(OchushaApplication *application)
{
  OchushaBulletinBoard *board;
  char id_buffer[PATH_MAX];
  gchar name_buffer[1024];
  guint number = g_slist_length(application->favorites_list);

  snprintf(id_buffer, PATH_MAX, "http:///favorites-%d", number);
  snprintf(name_buffer, 1024, _("Favorites-%d"), number);
  board = virtual_board_new(name_buffer, id_buffer, application);
  ochusha_board_category_add_board(application->my_folder, board);
  application->favorites_list
    = g_slist_append(application->favorites_list, board);
  boardlist_view_add_board(BOARDLIST_VIEW(application->boardlist_view),
			   board, application->my_folder, NULL);
}
