/*
 * Copyright (c) 2003-2004 The Ochusha Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: ochusha_async_buffer.h,v 1.14.2.2 2004/06/11 12:00:48 fuyu Exp $
 */

#ifndef _OCHUSHA_ASYNCBUF_H_
#define _OCHUSHA_ASYNCBUF_H_

#include "monitor.h"

#include <glib-object.h>
#include <glib.h>

#ifdef __cplusplus
extern "C" {
#endif


#define DEBUG_ASYNC_BUFFER	0
#define DEBUG_ASYNC_BUFFER_MOST	0


#define OCHUSHA_TYPE_ASYNC_BUFFER		(ochusha_async_buffer_get_type())
#define OCHUSHA_ASYNC_BUFFER(obj)		(G_TYPE_CHECK_INSTANCE_CAST((obj), OCHUSHA_TYPE_ASYNC_BUFFER, OchushaAsyncBuffer))
#define OCHUSHA_ASYNC_BUFFER_CLASS(klass)	(G_TYPE_CHECK_CLASS_CAST((klass), OCHUSHA_TYPE_ASYNC_BUFFER, OchushaAsyncBufferClass))
#define OCHUSHA_IS_ASYNC_BUFFER(obj)		(G_TYPE_CHECK_INSTANCE_TYPE((obj), OCHUSHA_TYPE_ASYNC_BUFFER))
#define OCHUSHA_IS_ASYNC_BUFFER_CLASS(klass)	(G_TYPE_CHECK_CLASS_TYPE((klass), OCHUSHA_TYPE_ASYNC_BUFFER))
#define OCHUSHA_ASYNC_BUFFER_GET_CLASS(obj)	(G_TYPE_INSTANCE_GET_CLASS((obj), OCHUSHA_TYPE_ASYNC_BUFFER, OchushaAsyncBufferClass))


typedef struct _OchushaAsyncBuffer OchushaAsyncBuffer;
typedef struct _OchushaAsyncBufferClass OchushaAsyncBufferClass;

typedef void DestructFunc(OchushaAsyncBuffer *buffer);

typedef enum
{
  OCHUSHA_ASYNC_BUFFER_OK,
  OCHUSHA_ASYNC_BUFFER_SUSPENDED,
  OCHUSHA_ASYNC_BUFFER_TERMINATED,
} OchushaAsyncBufferStatus;


struct _OchushaAsyncBuffer
{
  GObject parent_object;

  volatile gboolean fixed;
  char volatile *buffer;
  volatile size_t length;		/* bufferͭʥǡĹ */

  /* ʲprivate */
  size_t buffer_length;
  volatile int number_of_active_users;
  volatile int number_of_suspended_users;
  volatile OchushaAsyncBufferStatus state;

  Monitor *monitor;
  DestructFunc *destructor;

  volatile unsigned int number_of_lock_waiters;
};


struct _OchushaAsyncBufferClass
{
  GObjectClass parent_class;

  /* signal handlers */
  void (*wakeup_now)(OchushaAsyncBuffer *buffer);

  /* ΥʥOchushaAsyncBufferΥʥȤΤϡ
   * ݤʤΤǲ
   */
  void (*access_started)(OchushaAsyncBuffer *buffer);
  void (*access_progressed)(OchushaAsyncBuffer *buffer,
			    int bytes_read, int bytes_total);
  void (*access_finished)(OchushaAsyncBuffer *buffer);
  void (*access_failed)(OchushaAsyncBuffer *buffer, int reason_code,
			const gchar *reason);
};


GType ochusha_async_buffer_get_type(void) G_GNUC_CONST;

OchushaAsyncBuffer *ochusha_async_buffer_new(char *buffer, size_t length,
					     DestructFunc *destructor);
OchushaAsyncBuffer *ochusha_async_buffer_new_with_file(int fd);
OchushaAsyncBuffer *ochusha_async_buffer_new_with_file_mmap(int fd);
gboolean ochusha_async_buffer_reset(OchushaAsyncBuffer *buffer);

gboolean ochusha_async_buffer_active_ref(OchushaAsyncBuffer *buffer);
void ochusha_async_buffer_active_unref(OchushaAsyncBuffer *buffer);

void ochusha_async_buffer_suspend_all(void);
void ochusha_async_buffer_resume_all(void);
void ochusha_async_buffer_terminate_all(void);

void ochusha_async_buffer_suspend(OchushaAsyncBuffer *buffer);
void ochusha_async_buffer_resume(OchushaAsyncBuffer *buffer);
void ochusha_async_buffer_terminate(OchushaAsyncBuffer *buffer);


gboolean ochusha_async_buffer_fix(OchushaAsyncBuffer *buffer);
gboolean ochusha_async_buffer_update_length(OchushaAsyncBuffer *buffer,
					    size_t length);
gboolean ochusha_async_buffer_resize(OchushaAsyncBuffer *buffer,
				     size_t length);
gboolean ochusha_async_buffer_ensure_free_space(OchushaAsyncBuffer *buffer,
						size_t length);

gboolean ochusha_async_buffer_append_data(OchushaAsyncBuffer *buffer,
					  const char *data, size_t length);

gboolean ochusha_async_buffer_read_file(OchushaAsyncBuffer *buffer, int fd);

void ochusha_async_buffer_lock(OchushaAsyncBuffer *buffer);
void ochusha_async_buffer_unlock(OchushaAsyncBuffer *buffer);

gboolean ochusha_async_buffer_check_active(OchushaAsyncBuffer *buffer);
gboolean ochusha_async_buffer_signal(OchushaAsyncBuffer *buffer);
gboolean ochusha_async_buffer_broadcast(OchushaAsyncBuffer *buffer);
gboolean ochusha_async_buffer_wait(OchushaAsyncBuffer *buffer);
gboolean ochusha_async_buffer_is_busy(OchushaAsyncBuffer *buffer);

void ochusha_async_buffer_free_when_finished(OchushaAsyncBuffer *buffer);

void ochusha_async_buffer_emit_access_started(OchushaAsyncBuffer *buffer);
void ochusha_async_buffer_emit_access_progressed(OchushaAsyncBuffer *buffer,
						 int bytes_read,
						 int bytes_total);
void ochusha_async_buffer_emit_access_finished(OchushaAsyncBuffer *buffer);
void ochusha_async_buffer_emit_access_failed(OchushaAsyncBuffer *buffer,
					     int reason_code,
					     const gchar *reason);


#ifdef __cplusplus
} // extern "C"
#endif

#endif	/* _OCHUSHA_ASYNCBUF_H_ */
